/* Dynamicfunction
 * c = Dynamicfunction(a,b);
 * Adds offset argument a to each element of double array b and
 * returns the modified array c.
*/
#include <cmath>
#include <array>
#include "mat.inl"
#include "mex.hpp"
#include "mexAdapter.hpp"


mat<6,1> Gen3_VG(mat<6,1> q);

using namespace matlab::data;
using matlab::mex::ArgumentList;
class MexFunction : public matlab::mex::Function {
public:
    void operator()(ArgumentList outputs, ArgumentList inputs) {
        //checkArguments(outputs, inputs);
        ArrayFactory factory;

       TypedArray<double> q(inputs[0]);
        
        auto VG = Gen3_VG({q[0],q[1],q[2],q[3],q[4],q[5]});
        auto dims = ArrayDimensions(2);
        dims[0] = 6;
        dims[1] = 1;

        auto out = factory.createArray<double>(dims);
         
        for(int i  =  0 ; i < dims[0]; i++) {
            for(int j = 0; j < dims[1]; j++){
                out[i][j] = VG[i][j];
            }            
        }
        outputs[0] = out;
    }
    void checkArguments(ArgumentList outputs, ArgumentList inputs) {
        // Get pointer to engine
        std::shared_ptr<matlab::engine::MATLABEngine> matlabPtr = getEngine();
        // Get array factory
        ArrayFactory factory;
        // Check offset argument: First input must be scalar double
        if (inputs[0].getType() != ArrayType::DOUBLE ||
            inputs[0].getType() == ArrayType::COMPLEX_DOUBLE ||
            inputs[0].getNumberOfElements() != 1)
        {
            matlabPtr->feval(u"error",
                0,
                std::vector<Array>({ factory.createScalar("First input must be scalar double") }));
        }
        // Check array argument: Second input must be double array
        if (inputs[1].getType() != ArrayType::DOUBLE ||
            inputs[1].getType() == ArrayType::COMPLEX_DOUBLE)
        {
            matlabPtr->feval(u"error",
                0,
                std::vector<Array>({ factory.createScalar("Input must be double array") }));
        }
        // Check number of outputs
        if (outputs.size() > 1) {
            matlabPtr->feval(u"error",
                0,
                std::vector<Array>({ factory.createScalar("Only one output is returned") }));
        }
    }
};


mat<6,1> Gen3_VG(mat<6,1> q){

    // Precalculate sin(q) and cos(q)
        double s1 = sin(q[0][0]);
        double s2 = sin(q[1][0]);
        double s3 = sin(q[2][0]);
        double s4 = sin(q[3][0]);
        double s5 = sin(q[4][0]);
        double s6 = sin(q[5][0]);
        double c1 = cos(q[0][0]);
        double c2 = cos(q[1][0]);
        double c3 = cos(q[2][0]);
        double c4 = cos(q[3][0]);
        double c5 = cos(q[4][0]);
        double c6 = cos(q[5][0]);
    
/* Actualizado 11/10/23  KINOVA GEN 3 LITE GRAVITY VECTOR*/

    mat<6,1> g = {};

    g[0][0] = 0;
    g[1][0] = 0.7885*c5*(c2*s3 - 1.0*c3*s2) - 7.683*s2 - 0.34637*c2 - 0.17681*c2*c3 - 0.055877*s5*(c2*s3 - 1.0*c3*s2) + 3.6129*c2*s3 - 3.6129*c3*s2 - 0.17681*s2*s3 - 0.019697*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 0.019659*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.41174*c4*(c2*c3 + s2*s3) + 0.051715*s4*(c2*c3 + s2*s3) - 0.019659*c4*c6*(c2*c3 + s2*s3) + 0.055877*c5*s4*(c2*c3 + s2*s3) + 0.019697*c4*s6*(c2*c3 + s2*s3) + 0.7885*s4*s5*(c2*c3 + s2*s3);
    g[2][0] = 0.17681*c2*c3 - 0.7885*c5*(c2*s3 - 1.0*c3*s2) + 0.055877*s5*(c2*s3 - 1.0*c3*s2) - 3.6129*c2*s3 + 3.6129*c3*s2 + 0.17681*s2*s3 + 0.019697*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.019659*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 0.41174*c4*(c2*c3 + s2*s3) - 0.051715*s4*(c2*c3 + s2*s3) + 0.019659*c4*c6*(c2*c3 + s2*s3) - 0.055877*c5*s4*(c2*c3 + s2*s3) - 0.019697*c4*s6*(c2*c3 + s2*s3) - 0.7885*s4*s5*(c2*c3 + s2*s3);
    g[3][0] = 0.41174*s4*(c2*s3 - 1.0*c3*s2) - 0.051715*c4*(c2*s3 - 1.0*c3*s2) - 0.055877*c4*c5*(c2*s3 - 1.0*c3*s2) - 0.7885*c4*s5*(c2*s3 - 1.0*c3*s2) - 0.019659*c6*s4*(c2*s3 - 1.0*c3*s2) + 0.019697*s4*s6*(c2*s3 - 1.0*c3*s2) - 0.019697*c4*c5*c6*(c2*s3 - 1.0*c3*s2) - 0.019659*c4*c5*s6*(c2*s3 - 1.0*c3*s2);
    g[4][0] = 0.055877*s4*s5*(c2*s3 - 1.0*c3*s2) - 0.019659*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.055877*c5*(c2*c3 + s2*s3) - 0.7885*s5*(c2*c3 + s2*s3) - 0.7885*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.019697*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2));
    g[5][0] = 0.019697*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.019659*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.019697*c4*c6*(c2*s3 - 1.0*c3*s2) - 0.019659*c4*s6*(c2*s3 - 1.0*c3*s2);

    return g;
}


 