/* MyMEXFunction
 * c = MyMEXFunction(a,b);
 * Adds offset argument a to each element of double array b and
 * returns the modified array c.
*/

/*  Gen3 Lite Analytical Jacobian (Roll-Pitch-Yaw)
    Luis Tabarez
    Oct-13-2024
*/ 

#include <cmath>
#include <array>
#include <vector>
#include "mex.hpp"
#include "mexAdapter.hpp"


#define d1 0.2433
#define a2 0.2800
#define d2 0.03
#define d3 0.02
#define d4 0.2450
#define d5 0.0570
#define d6 0.2350



using namespace matlab::data;
using matlab::mex::ArgumentList;


class MexFunction : public matlab::mex::Function {
public:
    void operator()(ArgumentList outputs, ArgumentList inputs) {
        //checkArguments(outputs, inputs);
        ArrayFactory factory;

        TypedArray<double> q(inputs[0]);

        double c_12, s_12, c_gamma, c_beta, s_gamma, s_beta, s_gg, c_gg, sinq[6], cosq[6];

        for (int i = 0; i < 6; i++)
        {
            cosq[i]=cos(q[i][0]);
            sinq[i]=sin(q[i][0]);
        }

        c_12=cos(q[1][0] - q[2][0]);
        s_12=sin(q[1][0] - q[2][0]);
        c_gamma=cos(q[8][0]);
        c_beta=cos(q[7][0]);
        s_gamma=sin(q[8][0]);
        s_beta=sin(q[7][0]);
        
        s_gg=pow(s_gamma,2);
        c_gg=pow(c_gamma,2);

        auto dims = ArrayDimensions(2);
        dims[0] = 6;
        dims[1] = 6;

        auto out = factory.createArray<double>(dims);

         // Row 1
        out[0][0]=d6*(sinq[4]*(cosq[0]*cosq[3] - c_12*sinq[0]*sinq[3]) + s_12*cosq[4]*sinq[0]) - d5*(cosq[0]*sinq[3] + c_12*cosq[3]*sinq[0]) + d2*cosq[0] - d3*cosq[0] + a2*sinq[0]*sinq[1] + d4*s_12*sinq[0];
        out[0][1]=-cosq[0]*(a2*cosq[1] + d6*(c_12*cosq[4] + s_12*sinq[3]*sinq[4]) + d4*c_12 + d5*s_12*cosq[3]);
        out[0][2]=cosq[0]*(d6*(c_12*cosq[4] + s_12*sinq[3]*sinq[4]) + d4*c_12 + d5*s_12*cosq[3]);
        out[0][3]=d6*c_12*cosq[0]*cosq[3]*sinq[4] - d6*sinq[0]*sinq[3]*sinq[4] - d5*c_12*cosq[0]*sinq[3] - d5*cosq[3]*sinq[0];
        out[0][4]=d6*(s_12*cosq[0]*sinq[4] + cosq[3]*cosq[4]*sinq[0] + c_12*cosq[0]*cosq[4]*sinq[3]);
        out[0][5]=0;
        // Row 2
        out[1][0]=d6*(sinq[4]*(cosq[3]*sinq[0] + c_12*cosq[0]*sinq[3]) - s_12*cosq[0]*cosq[4]) - d5*(sinq[0]*sinq[3] - c_12*cosq[0]*cosq[3]) + d2*sinq[0] - d3*sinq[0] - a2*cosq[0]*sinq[1] - d4*s_12*cosq[0];
        out[1][1]=-sinq[0]*(a2*cosq[1] + d6*(c_12*cosq[4] + s_12*sinq[3]*sinq[4]) + d4*c_12 + d5*s_12*cosq[3]);
        out[1][2]=sinq[0]*(d6*(c_12*cosq[4] + s_12*sinq[3]*sinq[4]) + d4*c_12 + d5*s_12*cosq[3]);
        out[1][3]=d5*cosq[0]*cosq[3] + d6*cosq[0]*sinq[3]*sinq[4] - d5*c_12*sinq[0]*sinq[3] + d6*c_12*cosq[3]*sinq[0]*sinq[4];
        out[1][4]=d6*(s_12*sinq[0]*sinq[4] - cosq[0]*cosq[3]*cosq[4] + c_12*cosq[4]*sinq[0]*sinq[3]);
        out[1][5]=0;
        // Row 3
        out[2][0]=0;
        out[2][1]=d4*cosq[1]*sinq[2] - a2*sinq[1] - d4*cosq[2]*sinq[1] + d5*cosq[1]*cosq[2]*cosq[3] + d6*cosq[1]*cosq[4]*sinq[2] - d6*cosq[2]*cosq[4]*sinq[1] + d5*cosq[3]*sinq[1]*sinq[2] + d6*cosq[1]*cosq[2]*sinq[3]*sinq[4] + d6*sinq[1]*sinq[2]*sinq[3]*sinq[4];
        out[2][2]=d4*s_12 - d5*c_12*cosq[3] + d6*s_12*cosq[4] - d6*c_12*sinq[3]*sinq[4];
        out[2][3]=-s_12*(d5*sinq[3] - d6*cosq[3]*sinq[4]);
        out[2][4]=d6*s_12*cosq[4]*sinq[3] - d6*c_12*sinq[4];
        out[2][5]=0;
        // Row 4
        out[3][0]=0;
        out[3][1]=(c_gamma*sinq[0] - s_gamma*cosq[0])/(c_beta*(c_gg + s_gg));
        out[3][2]=-(c_gamma*sinq[0] - s_gamma*cosq[0])/(c_beta*(c_gg + s_gg));
        out[3][3]=-(s_12*(c_gamma*cosq[0] + s_gamma*sinq[0]))/(c_beta*(c_gg + s_gg));
        out[3][4]=(s_gamma*(cosq[0]*sinq[3] + c_12*cosq[3]*sinq[0]))/(c_beta*(c_gg + s_gg)) - (c_gamma*(sinq[0]*sinq[3] - c_12*cosq[0]*cosq[3]))/(c_beta*(c_gg + s_gg));
        out[3][5]=(c_gamma*(sinq[4]*(cosq[3]*sinq[0] + c_12*cosq[0]*sinq[3]) - s_12*cosq[0]*cosq[4]))/(c_beta*(c_gg + s_gg)) - (s_gamma*(sinq[4]*(cosq[0]*cosq[3] - c_12*sinq[0]*sinq[3]) + s_12*cosq[4]*sinq[0]))/(c_beta*(c_gg + s_gg));
        // Row 5
        out[4][0]=0;
        out[4][1]=-(c_gamma*cosq[0] + s_gamma*sinq[0])/(c_gg + s_gg);
        out[4][2]=(c_gamma*cosq[0] + s_gamma*sinq[0])/(c_gg + s_gg);
        out[4][3]=-(s_12*(c_gamma*sinq[0] - s_gamma*cosq[0]))/(c_gg + s_gg);
        out[4][4]=(c_gamma*(cosq[0]*sinq[3] + c_12*cosq[3]*sinq[0]))/(c_gg + s_gg) + (s_gamma*(sinq[0]*sinq[3] - c_12*cosq[0]*cosq[3]))/(c_gg + s_gg);
        out[4][5]=- (c_gamma*(sinq[4]*(cosq[0]*cosq[3] - c_12*sinq[0]*sinq[3]) + s_12*cosq[4]*sinq[0]))/(c_gg + s_gg) - (s_gamma*(sinq[4]*(cosq[3]*sinq[0] + c_12*cosq[0]*sinq[3]) - s_12*cosq[0]*cosq[4]))/(c_gg + s_gg);
        // Row 6
        out[5][0]=1;
        out[5][1]=(s_beta*(c_gamma*sinq[0] - s_gamma*cosq[0]))/(c_beta*(c_gg + s_gg));
        out[5][2]=-(s_beta*(c_gamma*sinq[0] - s_gamma*cosq[0]))/(c_beta*(c_gg + s_gg));
        out[5][3]=c_12 - (c_gamma*s_beta*s_12*cosq[0])/(c_beta*(c_gg + s_gg)) - (s_beta*s_gamma*s_12*sinq[0])/(c_beta*(c_gg + s_gg));
        out[5][4]=s_12*cosq[3] - (c_gamma*s_beta*(sinq[0]*sinq[3] - c_12*cosq[0]*cosq[3]))/(c_beta*(c_gg + s_gg)) + (s_beta*s_gamma*(cosq[0]*sinq[3] + c_12*cosq[3]*sinq[0]))/(c_beta*(c_gg + s_gg));
        out[5][5]=c_12*cosq[4] + s_12*sinq[3]*sinq[4] + (c_gamma*s_beta*(sinq[4]*(cosq[3]*sinq[0] + c_12*cosq[0]*sinq[3]) - s_12*cosq[0]*cosq[4]))/(c_beta*(c_gg + s_gg)) - (s_beta*s_gamma*(sinq[4]*(cosq[0]*cosq[3] - c_12*sinq[0]*sinq[3]) + s_12*cosq[4]*sinq[0]))/(c_beta*(c_gg + s_gg));

        outputs[0] = out;
    }

    void checkArguments(ArgumentList outputs, ArgumentList inputs) {
        // Get pointer to engine
        std::shared_ptr<matlab::engine::MATLABEngine> matlabPtr = getEngine();

        // Get array factory
        ArrayFactory factory;

        // Check offset argument: First input must be scalar double
        if (inputs[0].getType() != ArrayType::DOUBLE ||
            inputs[0].getType() == ArrayType::COMPLEX_DOUBLE ||
            inputs[0].getNumberOfElements() != 1)
        {
            matlabPtr->feval(u"error",
                0,
                std::vector<Array>({ factory.createScalar("First input must be scalar double") }));
        }

        // Check array argument: Second input must be double array
        if (inputs[1].getType() != ArrayType::DOUBLE ||
            inputs[1].getType() == ArrayType::COMPLEX_DOUBLE)
        {
            matlabPtr->feval(u"error",
                0,
                std::vector<Array>({ factory.createScalar("Input must be double array") }));
        }
        // Check number of outputs
        if (outputs.size() > 1) {
            matlabPtr->feval(u"error",
                0,
                std::vector<Array>({ factory.createScalar("Only one output is returned") }));
        }
    }
};

