/* MyMEXFunction
 * c = MyMEXFunction(a,b);
 * Adds offset argument a to each element of double array b and
 * returns the modified array c.
*/

/*  Gen3 Lite Direct kinematics
    Luis Tabarez
    Oct-13-2024
*/ 

#include <cmath>
#include <array>
#include <vector>
#include "mex.hpp"
#include "mexAdapter.hpp"


#define d1 0.2433
#define a2 0.2800
#define d2 0.03
#define d3 0.02
#define d4 0.2450
#define d5 0.0570
#define d6 0.2350



using namespace matlab::data;
using matlab::mex::ArgumentList;


class MexFunction : public matlab::mex::Function {
public:
    void operator()(ArgumentList outputs, ArgumentList inputs) {
        //checkArguments(outputs, inputs);
        ArrayFactory factory;

        TypedArray<double> q(inputs[0]);

        double sinq[6], cosq[6];

        for (int i = 0; i < 6; i++)
        {
            cosq[i]=cos(q[i][0]);
            sinq[i]=sin(q[i][0]);
        }

        double c_12=cos(q[1][0] - q[2][0]);
        double s_12=sin(q[1][0] - q[2][0]);
        
        auto dims = ArrayDimensions(2);
        dims[0] = 4;
        dims[1] = 4;

        auto out = factory.createArray<double>(dims);

         // Row 1
        out[0][0]= -sinq[5]*(cosq[4]*(cosq[3]*sinq[0] + c_12*cosq[0]*sinq[3]) + s_12*cosq[0]*sinq[4]) - cosq[5]*(sinq[0]*sinq[3] - c_12*cosq[0]*cosq[3]);
        out[0][1]=sinq[5]*(sinq[0]*sinq[3] - c_12*cosq[0]*cosq[3]) - cosq[5]*(cosq[4]*(cosq[3]*sinq[0] + c_12*cosq[0]*sinq[3]) + s_12*cosq[0]*sinq[4]);
        out[0][2]=sinq[4]*(cosq[3]*sinq[0] + c_12*cosq[0]*sinq[3]) - s_12*cosq[0]*cosq[4];
        out[0][3]=d6*(sinq[4]*(cosq[3]*sinq[0] + c_12*cosq[0]*sinq[3]) - s_12*cosq[0]*cosq[4]) - d5*(sinq[0]*sinq[3] - c_12*cosq[0]*cosq[3]) + d2*sinq[0] - d3*sinq[0] - a2*cosq[0]*sinq[1] - d4*s_12*cosq[0];
        // Row 2
        out[1][0]=sinq[5]*(cosq[4]*(cosq[0]*cosq[3] - c_12*sinq[0]*sinq[3]) - s_12*sinq[0]*sinq[4]) + cosq[5]*(cosq[0]*sinq[3] + c_12*cosq[3]*sinq[0]);
        out[1][1]=cosq[5]*(cosq[4]*(cosq[0]*cosq[3] - c_12*sinq[0]*sinq[3]) - s_12*sinq[0]*sinq[4]) - sinq[5]*(cosq[0]*sinq[3] + c_12*cosq[3]*sinq[0]);
        out[1][2]=-sinq[4]*(cosq[0]*cosq[3] - c_12*sinq[0]*sinq[3]) - s_12*cosq[4]*sinq[0];
        out[1][3]=d5*(cosq[0]*sinq[3] + c_12*cosq[3]*sinq[0]) - d6*(sinq[4]*(cosq[0]*cosq[3] - c_12*sinq[0]*sinq[3]) + s_12*cosq[4]*sinq[0]) - d2*cosq[0] + d3*cosq[0] - a2*sinq[0]*sinq[1] - d4*s_12*sinq[0];
        // Row 3
        out[2][0]=sinq[5]*(c_12*sinq[4] - s_12*cosq[4]*sinq[3]) + s_12*cosq[3]*cosq[5];
        out[2][1]=cosq[5]*(c_12*sinq[4] - s_12*cosq[4]*sinq[3]) - s_12*cosq[3]*sinq[5];
        out[2][2]=c_12*cosq[4] + s_12*sinq[3]*sinq[4];
        out[2][3]=d1 + a2*cosq[1] + d6*(c_12*cosq[4] + s_12*sinq[3]*sinq[4]) + d4*c_12 + d5*s_12*cosq[3];
        // Row 4
        out[3][0]=0;
        out[3][1]=0;
        out[3][2]=0;
        out[3][3]=1;

        outputs[0] = out;
    }

    void checkArguments(ArgumentList outputs, ArgumentList inputs) {
        // Get pointer to engine
        std::shared_ptr<matlab::engine::MATLABEngine> matlabPtr = getEngine();

        // Get array factory
        ArrayFactory factory;

        // Check offset argument: First input must be scalar double
        if (inputs[0].getType() != ArrayType::DOUBLE ||
            inputs[0].getType() == ArrayType::COMPLEX_DOUBLE ||
            inputs[0].getNumberOfElements() != 1)
        {
            matlabPtr->feval(u"error",
                0,
                std::vector<Array>({ factory.createScalar("First input must be scalar double") }));
        }

        // Check array argument: Second input must be double array
        if (inputs[1].getType() != ArrayType::DOUBLE ||
            inputs[1].getType() == ArrayType::COMPLEX_DOUBLE)
        {
            matlabPtr->feval(u"error",
                0,
                std::vector<Array>({ factory.createScalar("Input must be double array") }));
        }
        // Check number of outputs
        if (outputs.size() > 1) {
            matlabPtr->feval(u"error",
                0,
                std::vector<Array>({ factory.createScalar("Only one output is returned") }));
        }
    }
};

