/* Dynamicfunction
 * c = Dynamicfunction(a,b);
 * Adds offset argument a to each element of double array b and
 * returns the modified array c.
*/
#include <cmath>
#include <array>
#include "mat.inl"
#include "mex.hpp"
#include "mexAdapter.hpp"
//#include "MatlabDataArray.hpp"
//#include "ArrayFactory.hpp"


mat<6,6> Gen3_MC(mat<6,1> q, mat<6,1> qp);

using namespace matlab::data;
using matlab::mex::ArgumentList;
class MexFunction : public matlab::mex::Function {
public:
    void operator()(ArgumentList outputs, ArgumentList inputs) {
        //checkArguments(outputs, inputs);
        ArrayFactory factory;

        TypedArray<double> q(inputs[0]);
        TypedArray<double> qp(inputs[1]);

        auto MC = Gen3_MC({q[0],q[1],q[2],q[3],q[4],q[5]}, {qp[0],qp[1],qp[2],qp[3],qp[4],qp[5]});

        auto dims = ArrayDimensions(2);
        dims[0] = 6;
        dims[1] = 6;

        auto out = factory.createArray<double>(dims);
        for(int i = 0 ; i < dims[0]; i++) {
            for(int j = 0; j < dims[1]; j++){
                out[i][j] = MC[i][j];
            } 
        }
        outputs[0] = out;
    }
    void checkArguments(ArgumentList outputs, ArgumentList inputs) {
        // Get pointer to engine
        std::shared_ptr<matlab::engine::MATLABEngine> matlabPtr = getEngine();
        // Get array factory
        ArrayFactory factory;
        // Check offset argument: First input must be scalar double
        if (inputs[0].getType() != ArrayType::DOUBLE ||
            inputs[0].getType() == ArrayType::COMPLEX_DOUBLE ||
            inputs[0].getNumberOfElements() != 1)
        {
            matlabPtr->feval(u"error",
            0,
            std::vector<Array>({ factory.createScalar("First input must be scalar double") }));
        }
        // Check array argument: Second input must be double array
        if (inputs[1].getType() != ArrayType::DOUBLE ||
            inputs[1].getType() == ArrayType::COMPLEX_DOUBLE)
        {
            matlabPtr->feval(u"error",
            0,
            std::vector<Array>({ factory.createScalar("Input must be double array") }));
        }
        // Check number of outputs
        if (outputs.size() > 1) {
            matlabPtr->feval(u"error",
            0,
            std::vector<Array>({ factory.createScalar("Only one output is returned") }));
        }
    }
};


mat<6,6> Gen3_MC(mat<6,1> q, mat<6,1> qp){
 mat<6,6> C = {};

 // Precalculate sin(q) and cos(q)
    double s1 = sin(q[0][0]);
    double s2 = sin(q[1][0]);
    double s3 = sin(q[2][0]);
    double s4 = sin(q[3][0]);
    double s5 = sin(q[4][0]);
    double s6 = sin(q[5][0]);
    double s2_10_3 = sin(q[1][0] - 1.0 * q[2][0]);              //sin(q2 - 1.0*q3)
    double s4_5 = sin(q[3][0] + q[4][0]);                       //sin(q4 + q5)
    double s4_10_5 = sin(q[3][0] - 1.0 * q[4][0]);              //sin(q4 - 1.0*q5)
    double s3_10_2_12 = sin(q[2][0] - 1.0 * q[1][0] + 1.2635);  //cos(q3 - 1.0*q2 + 1.2635)

    double c1 = cos(q[0][0]);
    double c2 = cos(q[1][0]);
    double c3 = cos(q[2][0]);
    double c4 = cos(q[3][0]);
    double c5 = cos(q[4][0]);
    double c6 = cos(q[5][0]); 
    double c2_10_3 = cos(q[1][0] - 1.0 * q[2][0]);              //cos(q2 - 1.0*q3)
    double c4_5 = cos(q[3][0] + q[4][0]);                       //cos(q4 + q5)
    double c4_10_5 = cos(q[3][0] - 1.0 * q[4][0]);              //cos(q4 - 1.0*q5)
    double c3_10_2_12 = cos(q[2][0] - 1.0 * q[1][0] + 1.2635);  //cos(q3 - 1.0*q2 + 1.2635)

    double qp1 = qp[0][0];
    double qp2 = qp[1][0];
    double qp3 = qp[2][0];
    double qp4 = qp[3][0];
    double qp5 = qp[4][0];
    double qp6 = qp[5][0];

 /* C[0] = std::array<double,6>{};  13/10/23 KINOVA GEN 3 LITE CORIOLIS MATRIX */
 
 C[0] = std::array<double,6>{ - 1.0*qp6*(0.5*(0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.01*s1 - 0.28*c1*s2 - 0.057*s1*s4 + 0.16636*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*c4*(c1*c2*c3 + c1*s2*s3) - 0.0099495*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) - 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.0099495*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.0099304*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099304*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099495*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) - 0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.0099495*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.0099304*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099495*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*(0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.01*c1 - 0.057*c1*s4 + 0.28*s1*s2 + 0.0099304*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099495*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.16636*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*c4*(s1*s2*s3 + c2*c3*s1) - 0.16636*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*c2*s1*s3 + 0.245*c3*s1*s2) + 0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + qp2*(0.5*(0.16735*c1*c2 + 0.056793*c1*c2*c3 + 0.018024*c1*c2*s3 - 0.018024*c1*c3*s2 + 0.056793*c1*s2*s3)*(0.28*c1*s2 - 0.037355*s1 + 0.030156*c1*c2*c3 - 0.095022*c1*c2*s3 + 0.095022*c1*c3*s2 + 0.030156*c1*s2*s3) - 0.5*(0.16267*c2*s1 - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.14234*s1*s2*s3 + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.047228*c1*s4 - 0.01*c1 - 0.28*s1*s2 + 0.0098041*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.080565*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.047228*c2_10_3*c4*s1 - 0.080565*s2_10_3*c5*s1 + 0.0098041*s2_10_3*s1*s5 + 0.245*c2*s1*s3 - 0.245*c3*s1*s2) + 0.5*(0.28*c1*c2 + 0.095022*c1*c2*c3 + 0.030156*c1*c2*s3 - 0.030156*c1*c3*s2 + 0.095022*c1*s2*s3)*(0.16735*c1*s2 - 0.022327*s1 + 0.018024*c1*c2*c3 - 0.056793*c1*c2*s3 + 0.056793*c1*c3*s2 + 0.018024*c1*s2*s3) + 0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) - 0.5*(0.28*c2*s1 - 0.0098041*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.080565*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.245*s1*s2*s3 + 0.080565*c2_10_3*c5*s1 + 0.047228*s2_10_3*c4*s1 - 0.0098041*c2_10_3*s1*s5 + 0.245*c2*c3*s1)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(0.14754*c2*s1 + 0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1)*(0.01*c1 + 0.010004*c1*c4 - 0.0057515*c1*s4 + 0.28*s1*s2 - 0.0057515*c2_10_3*c4*s1 - 0.010004*c2_10_3*s1*s4 - 0.22719*c2*s1*s3 + 0.22719*c3*s1*s2) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 0.5*(c2*c3 + s2*s3)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + 0.5*(0.28*c2*s1 + 0.22719*s1*s2*s3 + 0.0057515*s2_10_3*c4*s1 + 0.010004*s2_10_3*s1*s4 + 0.22719*c2*c3*s1)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) + 0.5*(c2*s3 - 1.0*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 0.5*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(c2*c3 + s2*s3)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) - 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 0.5*(c2*s3 - 1.0*c3*s2)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3) + 0.5*(c2*c3 + s2*s3)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) - 0.5*(0.16267*c1*c2 - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.046806*c2_10_3*c1*c5 + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 - 0.047228*s1*s4 + 0.0098041*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.080565*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.047228*c2_10_3*c1*c4 - 0.080565*s2_10_3*c1*c5 + 0.0098041*s2_10_3*c1*s5 + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*c2*(1.0e-6*c2 - 0.0010285*s2) + 0.5*(0.28*c2*s1 + 0.16636*c5*(s1*s2*s3 + c2*c3*s1) - 0.057*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0099495*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.0099304*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.245*s1*s2*s3 + 0.0099304*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0099495*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.16636*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.245*c2*c3*s1)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(0.16735*c2*s1 + 0.056793*s1*s2*s3 + 0.056793*c2*c3*s1 + 0.018024*c2*s1*s3 - 0.018024*c3*s1*s2)*(0.037355*c1 + 0.28*s1*s2 + 0.030156*s1*s2*s3 + 0.030156*c2*c3*s1 - 0.095022*c2*s1*s3 + 0.095022*c3*s1*s2) + 0.5*(0.24912*c1*c2 - 0.035308*c1*s2)*(0.029983*c1*c2 - 0.075303*s1 + 0.21155*c1*s2) - 0.5*s2*(0.0010285*c2 + 1.0e-6*s2) - 0.5*(0.28*c1*c2 + 0.16636*c5*(c1*c2*c3 + c1*s2*s3) - 0.0099495*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.057*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0099304*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0099495*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.16636*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.245*c1*c2*c3 + 0.245*c1*s2*s3)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*(0.28*c2*s1 + 0.095022*s1*s2*s3 + 0.095022*c2*c3*s1 + 0.030156*c2*s1*s3 - 0.030156*c3*s1*s2)*(0.022327*c1 + 0.16735*s1*s2 + 0.018024*s1*s2*s3 + 0.018024*c2*c3*s1 - 0.056793*c2*s1*s3 + 0.056793*c3*s1*s2) - 0.5*(0.28*c1*c2 - 0.0098041*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.080565*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.080565*c2_10_3*c1*c5 + 0.047228*s2_10_3*c1*c4 - 0.0098041*c2_10_3*c1*s5 + 0.245*c1*c2*c3 + 0.245*c1*s2*s3)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 0.5*c2*(1.0e-6*c2 + 0.011493*s2) + 0.5*(0.21155*c1*c2 - 0.029983*c1*s2)*(0.035308*c1*c2 - 0.088677*s1 + 0.24912*c1*s2) - 0.5*(c2*s3 - 1.0*c3*s2)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) + 0.5*(c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1)*(0.01*c1 - 0.057*c1*s4 + 0.28*s1*s2 + 0.0099304*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099495*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.16636*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*c4*(s1*s2*s3 + c2*c3*s1) - 0.16636*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*c2*s1*s3 + 0.245*c3*s1*s2) + 0.5*s2*(0.011493*c2 - 1.0e-6*s2) - 0.5*(0.14754*c1*c2 + 0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 + 0.010004*c4*s1 - 0.0057515*s1*s4 + 0.0057515*c2_10_3*c1*c4 + 0.010004*c2_10_3*c1*s4 + 0.22719*c1*c2*s3 - 0.22719*c1*c3*s2) - 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 0.5*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 - 0.057*s1*s4 + 0.16636*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*c4*(c1*c2*c3 + c1*s2*s3) - 0.0099495*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*(0.24912*c2*s1 - 0.035308*s1*s2)*(0.075303*c1 + 0.029983*c2*s1 + 0.21155*s1*s2) - 0.5*(0.28*c1*c2 + 0.0057515*s2_10_3*c1*c4 + 0.010004*s2_10_3*c1*s4 + 0.22719*c1*c2*c3 + 0.22719*c1*s2*s3)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 0.5*(s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.21155*c2*s1 - 0.029983*s1*s2)*(0.088677*c1 + 0.035308*c2*s1 + 0.24912*s1*s2) - 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + 0.5*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) - 0.5*s4*(c2*s3 - 1.0*c3*s2)*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - 0.5*c4*(c2*c3 + s2*s3)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - 0.5*s4*(c2*c3 + s2*s3)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + qp4*(0.5*(7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c2*c3 + s2*s3) - 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(s4*s6*(c2*s3 - 1.0*c3*s2) - 1.0*c4*c5*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.057*c1*c4 + 0.16636*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099304*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0099495*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*s4*(s1*s2*s3 + c2*c3*s1) + 0.0099495*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.0099304*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(0.047228*c1*c4 + 0.0098041*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.080565*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.047228*c2_10_3*s1*s4)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(c6*s4*(c2*s3 - 1.0*c3*s2) + c4*c5*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.01*c1 - 0.057*c1*s4 + 0.28*s1*s2 + 0.0099304*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099495*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.16636*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*c4*(s1*s2*s3 + c2*c3*s1) - 0.16636*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*c2*s1*s3 + 0.245*c3*s1*s2) + 0.5*(0.027438*c1*c4 + 0.0056959*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.046806*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.027438*c2_10_3*s1*s4)*(0.047228*c1*s4 - 0.01*c1 - 0.28*s1*s2 + 0.0098041*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.080565*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.047228*c2_10_3*c4*s1 - 0.080565*s2_10_3*c5*s1 + 0.0098041*s2_10_3*s1*s5 + 0.245*c2*s1*s3 - 0.245*c3*s1*s2) - 0.5*(0.057*c4*s1 - 0.0099304*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099495*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*s4*(c1*c2*c3 + c1*s2*s3) + 0.0099495*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*(0.0030306*c4*s1 + 0.0052716*s1*s4 - 0.0052716*c2_10_3*c1*c4 + 0.0030306*c2_10_3*c1*s4)*(0.01*s1 - 0.28*c1*s2 + 0.010004*c4*s1 - 0.0057515*s1*s4 + 0.0057515*c2_10_3*c1*c4 + 0.010004*c2_10_3*c1*s4 + 0.22719*c1*c2*s3 - 0.22719*c1*c3*s2) - 0.5*(0.0057515*c4*s1 + 0.010004*s1*s4 - 0.010004*c2_10_3*c1*c4 + 0.0057515*c2_10_3*c1*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) - 0.5*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.01*s1 - 0.28*c1*s2 - 0.057*s1*s4 + 0.16636*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*c4*(c1*c2*c3 + c1*s2*s3) - 0.0099495*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) - 0.5*(0.0030306*c1*c4 + 0.0052716*c1*s4 + 0.0052716*c2_10_3*c4*s1 - 0.0030306*c2_10_3*s1*s4)*(0.01*c1 + 0.010004*c1*c4 - 0.0057515*c1*s4 + 0.28*s1*s2 - 0.0057515*c2_10_3*c4*s1 - 0.010004*c2_10_3*s1*s4 - 0.22719*c2*s1*s3 + 0.22719*c3*s1*s2) - 0.5*(0.0057515*c1*c4 + 0.010004*c1*s4 + 0.010004*c2_10_3*c4*s1 - 0.0057515*c2_10_3*s1*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) + 0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.047228*c4*s1 - 0.0098041*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.080565*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.047228*c2_10_3*c1*s4)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.027438*c4*s1 - 0.0056959*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.046806*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.027438*c2_10_3*c1*s4)*(0.01*s1 - 0.28*c1*s2 - 0.047228*s1*s4 + 0.0098041*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.080565*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.047228*c2_10_3*c1*c4 - 0.080565*s2_10_3*c1*c5 + 0.0098041*s2_10_3*c1*s5 + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 0.5*s4*(c2*s3 - 1.0*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 0.5*c4*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2))*(c2*s3 - 1.0*c3*s2) + 0.5*c4*(c2*s3 - 1.0*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + 0.5*s4*(0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c2*s3 - 1.0*c3*s2) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*s4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*c4*s5*(c2*s3 - 1.0*c3*s2)*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*c4*c5*(c2*s3 - 1.0*c3*s2)*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*c4*s5*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2))) + qp5*(0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.0099495*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.16636*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.16636*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*(0.080565*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0098041*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0098041*s2_10_3*c5*s1 + 0.080565*s2_10_3*s1*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.046806*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0056959*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0056959*s2_10_3*c5*s1 + 0.046806*s2_10_3*s1*s5)*(0.047228*c1*s4 - 0.01*c1 - 0.28*s1*s2 + 0.0098041*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.080565*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.047228*c2_10_3*c4*s1 - 0.080565*s2_10_3*c5*s1 + 0.0098041*s2_10_3*s1*s5 + 0.245*c2*s1*s3 - 0.245*c3*s1*s2) + 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.01*s1 - 0.28*c1*s2 - 0.057*s1*s4 + 0.16636*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*c4*(c1*c2*c3 + c1*s2*s3) - 0.0099495*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*(0.080565*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0098041*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0098041*s2_10_3*c1*c5 + 0.080565*s2_10_3*c1*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*(0.0099495*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.16636*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.16636*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(0.046806*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0056959*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0056959*s2_10_3*c1*c5 + 0.046806*s2_10_3*c1*s5)*(0.01*s1 - 0.28*c1*s2 - 0.047228*s1*s4 + 0.0098041*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.080565*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.047228*c2_10_3*c1*c4 - 0.080565*s2_10_3*c1*c5 + 0.0098041*s2_10_3*c1*s5 + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.01*c1 - 0.057*c1*s4 + 0.28*s1*s2 + 0.0099304*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099495*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.16636*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*c4*(s1*s2*s3 + c2*c3*s1) - 0.16636*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*c2*s1*s3 + 0.245*c3*s1*s2) + 0.5*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 1.0*qp3*(0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + 0.5*(0.056793*s1*s2*s3 + 0.056793*c2*c3*s1 + 0.018024*c2*s1*s3 - 0.018024*c3*s1*s2)*(0.037355*c1 + 0.28*s1*s2 + 0.030156*s1*s2*s3 + 0.030156*c2*c3*s1 - 0.095022*c2*s1*s3 + 0.095022*c3*s1*s2) - 0.5*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.01*c1 - 0.057*c1*s4 + 0.28*s1*s2 + 0.0099304*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099495*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.16636*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*c4*(s1*s2*s3 + c2*c3*s1) - 0.16636*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*c2*s1*s3 + 0.245*c3*s1*s2) - 0.5*(0.046806*c2_10_3*c1*c5 - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 - 0.047228*s1*s4 + 0.0098041*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.080565*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.047228*c2_10_3*c1*c4 - 0.080565*s2_10_3*c1*c5 + 0.0098041*s2_10_3*c1*s5 + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) - 0.5*(0.057*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.16636*c5*(s1*s2*s3 + c2*c3*s1) + 0.0099495*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*s1*s2*s3 - 0.0099304*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.16636*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.245*c2*c3*s1)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 0.5*(c2*c3 + s2*s3)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + 0.5*(c2*s3 - 1.0*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 0.5*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.095022*s1*s2*s3 + 0.095022*c2*c3*s1 + 0.030156*c2*s1*s3 - 0.030156*c3*s1*s2)*(0.022327*c1 + 0.16735*s1*s2 + 0.018024*s1*s2*s3 + 0.018024*c2*c3*s1 - 0.056793*c2*s1*s3 + 0.056793*c3*s1*s2) + 0.5*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 - 0.057*s1*s4 + 0.16636*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*c4*(c1*c2*c3 + c1*s2*s3) - 0.0099495*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*(0.0099495*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.16636*c5*(c1*c2*c3 + c1*s2*s3) + 0.057*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099304*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.0099304*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.16636*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.245*c1*c2*c3 - 0.245*c1*s2*s3)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*(0.080565*c2_10_3*c1*c5 - 0.080565*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0098041*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.047228*s2_10_3*c1*c4 - 0.0098041*c2_10_3*c1*s5 + 0.245*c1*c2*c3 + 0.245*c1*s2*s3)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*(0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 + 0.010004*c4*s1 - 0.0057515*s1*s4 + 0.0057515*c2_10_3*c1*c4 + 0.010004*c2_10_3*c1*s4 + 0.22719*c1*c2*s3 - 0.22719*c1*c3*s2) + 0.5*(c2*c3 + s2*s3)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) - 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 0.5*(c2*s3 - 1.0*c3*s2)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3) - 0.5*(0.0057515*s2_10_3*c1*c4 + 0.010004*s2_10_3*c1*s4 + 0.22719*c1*c2*c3 + 0.22719*c1*s2*s3)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 0.5*(c2*c3 + s2*s3)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) - 0.5*(0.14234*s1*s2*s3 - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.047228*c1*s4 - 0.01*c1 - 0.28*s1*s2 + 0.0098041*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.080565*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.047228*c2_10_3*c4*s1 - 0.080565*s2_10_3*c5*s1 + 0.0098041*s2_10_3*s1*s5 + 0.245*c2*s1*s3 - 0.245*c3*s1*s2) + 0.5*(0.056793*c1*c2*c3 + 0.018024*c1*c2*s3 - 0.018024*c1*c3*s2 + 0.056793*c1*s2*s3)*(0.28*c1*s2 - 0.037355*s1 + 0.030156*c1*c2*c3 - 0.095022*c1*c2*s3 + 0.095022*c1*c3*s2 + 0.030156*c1*s2*s3) - 0.5*(c2*s3 - 1.0*c3*s2)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) + 0.5*(c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) + 0.5*(0.095022*c1*c2*c3 + 0.030156*c1*c2*s3 - 0.030156*c1*c3*s2 + 0.095022*c1*s2*s3)*(0.16735*c1*s2 - 0.022327*s1 + 0.018024*c1*c2*c3 - 0.056793*c1*c2*s3 + 0.056793*c1*c3*s2 + 0.018024*c1*s2*s3) - 0.5*(0.245*s1*s2*s3 - 0.080565*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0098041*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.080565*c2_10_3*c5*s1 + 0.047228*s2_10_3*c4*s1 - 0.0098041*c2_10_3*s1*s5 + 0.245*c2*c3*s1)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1)*(0.01*c1 + 0.010004*c1*c4 - 0.0057515*c1*s4 + 0.28*s1*s2 - 0.0057515*c2_10_3*c4*s1 - 0.010004*c2_10_3*s1*s4 - 0.22719*c2*s1*s3 + 0.22719*c3*s1*s2) + 0.5*(s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.22719*s1*s2*s3 + 0.0057515*s2_10_3*c4*s1 + 0.010004*s2_10_3*s1*s4 + 0.22719*c2*c3*s1)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) - 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + 0.5*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) - 0.5*s4*(c2*s3 - 1.0*c3*s2)*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - 0.5*c4*(c2*c3 + s2*s3)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - 0.5*s4*(c2*c3 + s2*s3)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)),  qp3*(1.0*s1*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) - 1.0*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - c1*(1.0*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - (c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + 1.0*(s1*s2*s3 + c2*c3*s1)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 1.0*c1*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + c1*((s1*s2*s3 + c2*c3*s1)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) - 1.0*(s1*s2*s3 + c2*c3*s1)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3) + (c2*s1*s3 - 1.0*c3*s1*s2)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + (c2*s1*s3 - 1.0*c3*s1*s2)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) - 1.0*c1*(1.5e-6*c2*c3 + 0.000096*c2*s3 - 0.000096*c3*s2 + 1.5e-6*s2*s3)) + 1.0*c1*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + s1*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + (c1*c2*c3 + c1*s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + 1.0*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + s4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 1.0*s1*((c1*c2*c3 + c1*s2*s3)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) + (c1*c2*s3 - 1.0*c1*c3*s2)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + (c1*c2*s3 - 1.0*c1*c3*s2)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) + s1*(1.5e-6*c2*c3 + 0.000096*c2*s3 - 0.000096*c3*s2 + 1.5e-6*s2*s3) - 1.0*(c1*c2*c3 + c1*s2*s3)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3)) + 0.5*c1*(0.095022*c2_10_3 - 0.030156*s2_10_3)*(0.16735*c2*s1 + 0.056793*s1*s2*s3 + 0.056793*c2*c3*s1 + 0.018024*c2*s1*s3 - 0.018024*c3*s1*s2) + s1*(0.0057515*c2_10_3*c4 - 0.22719*s2_10_3 + 0.010004*c2_10_3*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 0.5*c1*(0.14234*s1*s2*s3 - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) + 0.5*c1*(0.095022*c2_10_3 - 0.030156*s2_10_3 + 0.28*c2)*(0.056793*s1*s2*s3 + 0.056793*c2*c3*s1 + 0.018024*c2*s1*s3 - 0.018024*c3*s1*s2) + c1*(0.0057515*c2_10_3*c4 - 0.22719*s2_10_3 + 0.010004*c2_10_3*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) - 0.5*c1*(0.004902*c2_10_3*s4_5 - 0.040283*c2_10_3*c4_5 - 0.245*s2_10_3 + 0.047228*c2_10_3*c4 - 0.080565*s2_10_3*c5 + 0.0098041*s2_10_3*s5 + 0.040283*c2_10_3*c4_10_5 + 0.004902*c2_10_3*s4_10_5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + c1*(0.245*c2*s3 - 0.245*c3*s2 + 0.057*c2_10_3*c4 + 0.0099495*c6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.16636*s2_10_3*c5 + 0.0099304*s6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.0099304*c2_10_3*c4*c6 + 0.0099495*c2_10_3*c4*s6 + 0.16636*c2_10_3*s4*s5)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*s1*(0.004902*c2_10_3*s4_5 - 0.040283*c2_10_3*c4_5 - 0.245*s2_10_3 + 0.047228*c2_10_3*c4 - 0.080565*s2_10_3*c5 + 0.0098041*s2_10_3*s5 + 0.040283*c2_10_3*c4_10_5 + 0.004902*c2_10_3*s4_10_5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 6.9389e-21*c1*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5)*(0.16267*c2*s1 - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.14234*s1*s2*s3 + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1) + s1*(0.030156*c2_10_3 + 0.095022*s2_10_3)*(0.16735*c1*s2 - 0.022327*s1 + 0.018024*c1*c2*c3 - 0.056793*c1*c2*s3 + 0.056793*c1*c3*s2 + 0.018024*c1*s2*s3) - 0.5*s1*(0.046806*c2_10_3*c1*c5 - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) - 0.5*s1*(0.22719*c2_10_3 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.14754*c1*c2 + 0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3) - 1.0*c1*(0.030156*c2_10_3 + 0.095022*s2_10_3)*(0.022327*c1 + 0.16735*s1*s2 + 0.018024*s1*s2*s3 + 0.018024*c2*c3*s1 - 0.056793*c2*s1*s3 + 0.056793*c3*s1*s2) + 0.5*s1*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 0.5*s1*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5)*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3) + s1*(0.245*c2*s3 - 0.245*c3*s2 + 0.057*c2_10_3*c4 + 0.0099495*c6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.16636*s2_10_3*c5 + 0.0099304*s6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.0099304*c2_10_3*c4*c6 + 0.0099495*c2_10_3*c4*s6 + 0.16636*c2_10_3*s4*s5)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*s1*(0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3)*(0.22719*c2_10_3 + 0.28*c2 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4) - 6.9389e-21*c1*(3.4031e+18*c2_10_3*c4 - 1.7654e+19*s2_10_3 - 5.8053e+18*s2_10_3*c5 + 7.0646e+17*s2_10_3*s5 + 7.0646e+17*c2_10_3*c5*s4 + 5.8053e+18*c2_10_3*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - 0.5*s1*(0.095022*c2_10_3 - 0.030156*s2_10_3)*(0.16735*c1*c2 + 0.056793*c1*c2*c3 + 0.018024*c1*c2*s3 - 0.018024*c1*c3*s2 + 0.056793*c1*s2*s3) + 0.5*c1*(0.22719*c2_10_3 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.14754*c2*s1 + 0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1) - 0.5*c1*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 6.9389e-21*s1*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5)*(0.16267*c1*c2 - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.046806*c2_10_3*c1*c5 + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3) - 0.5*s1*(0.095022*c2_10_3 - 0.030156*s2_10_3 + 0.28*c2)*(0.056793*c1*c2*c3 + 0.018024*c1*c2*s3 - 0.018024*c1*c3*s2 + 0.056793*c1*s2*s3) + 0.5*c1*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5)*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1) + 0.5*c1*(0.22719*c2_10_3 + 0.28*c2 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1) + 6.9389e-21*s1*(3.4031e+18*c2_10_3*c4 - 1.7654e+19*s2_10_3 - 5.8053e+18*s2_10_3*c5 + 7.0646e+17*s2_10_3*s5 + 7.0646e+17*c2_10_3*c5*s4 + 5.8053e+18*c2_10_3*s4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2)) + 1.0*qp5*(0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0098041*c1*c2*c3*c5 + 0.080565*c1*c2*c3*s5 + 0.0098041*c1*c5*s2*s3 + 0.080565*c1*s2*s3*s5 + 0.080565*c1*c2*c5*s3*s4 - 0.080565*c1*c3*c5*s2*s4 - 0.0098041*c1*c2*s3*s4*s5 + 0.0098041*c1*c3*s2*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(0.16267*c1*c2 - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.046806*c2_10_3*c1*c5 + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.080565*c1*c4*c5 - 0.0098041*c1*c4*s5 + 0.0098041*c2*c5*s1*s3 - 0.0098041*c3*c5*s1*s2 + 0.080565*c2*s1*s3*s5 - 0.080565*c3*s1*s2*s5 - 0.080565*c2*c3*c5*s1*s4 + 0.0098041*c2*c3*s1*s4*s5 - 0.080565*c5*s1*s2*s3*s4 + 0.0098041*s1*s2*s3*s4*s5) + 0.5*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.16636*c1*c2*c3*s5 + 0.16636*c1*s2*s3*s5 + 0.0099495*c1*c2*c3*c5*c6 + 0.0099304*c1*c2*c3*c5*s6 + 0.16636*c1*c2*c5*s3*s4 - 0.16636*c1*c3*c5*s2*s4 + 0.0099495*c1*c5*c6*s2*s3 + 0.0099304*c1*c5*s2*s3*s6 - 0.0099495*c1*c2*c6*s3*s4*s5 + 0.0099495*c1*c3*c6*s2*s4*s5 - 0.0099304*c1*c2*s3*s4*s5*s6 + 0.0099304*c1*c3*s2*s4*s5*s6) + 0.5*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.16636*c2*c3*s1*s5 + 0.16636*s1*s2*s3*s5 + 0.0099495*c2*c3*c5*c6*s1 + 0.0099304*c2*c3*c5*s1*s6 + 0.16636*c2*c5*s1*s3*s4 - 0.16636*c3*c5*s1*s2*s4 + 0.0099495*c5*c6*s1*s2*s3 + 0.0099304*c5*s1*s2*s3*s6 - 0.0099495*c2*c6*s1*s3*s4*s5 + 0.0099495*c3*c6*s1*s2*s4*s5 - 0.0099304*c2*s1*s3*s4*s5*s6 + 0.0099304*c3*s1*s2*s4*s5*s6) + 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(0.16267*c2*s1 - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.14234*s1*s2*s3 + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.080565*c4*c5*s1 - 0.0098041*c4*s1*s5 - 0.0098041*c1*c2*c5*s3 + 0.0098041*c1*c3*c5*s2 - 0.080565*c1*c2*s3*s5 + 0.080565*c1*c3*s2*s5 + 0.080565*c1*c2*c3*c5*s4 - 0.0098041*c1*c2*c3*s4*s5 + 0.080565*c1*c5*s2*s3*s4 - 0.0098041*c1*s2*s3*s4*s5) - 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) - 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1)*(0.16636*c1*c2*s3*s5 - 0.16636*c4*c5*s1 - 0.16636*c1*c3*s2*s5 + 0.0099495*c4*c6*s1*s5 + 0.0099304*c4*s1*s5*s6 - 0.16636*c1*c2*c3*c5*s4 + 0.0099495*c1*c2*c5*c6*s3 - 0.0099495*c1*c3*c5*c6*s2 + 0.0099304*c1*c2*c5*s3*s6 - 0.0099304*c1*c3*c5*s2*s6 - 0.16636*c1*c5*s2*s3*s4 + 0.0099495*c1*c2*c3*c6*s4*s5 + 0.0099304*c1*c2*c3*s4*s5*s6 + 0.0099495*c1*c6*s2*s3*s4*s5 + 0.0099304*c1*s2*s3*s4*s5*s6) + 0.5*(0.0098041*c2*c3*c5*s1 + 0.080565*c2*c3*s1*s5 + 0.0098041*c5*s1*s2*s3 + 0.080565*s1*s2*s3*s5 + 0.080565*c2*c5*s1*s3*s4 - 0.080565*c3*c5*s1*s2*s4 - 0.0098041*c2*s1*s3*s4*s5 + 0.0098041*c3*s1*s2*s4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3)*(0.16636*c1*c4*c5 - 0.0099495*c1*c4*c6*s5 - 0.0099304*c1*c4*s5*s6 + 0.16636*c2*s1*s3*s5 - 0.16636*c3*s1*s2*s5 - 0.16636*c2*c3*c5*s1*s4 + 0.0099495*c2*c5*c6*s1*s3 - 0.0099495*c3*c5*c6*s1*s2 + 0.0099304*c2*c5*s1*s3*s6 - 0.0099304*c3*c5*s1*s2*s6 - 0.16636*c5*s1*s2*s3*s4 + 0.0099495*c2*c3*c6*s1*s4*s5 + 0.0099304*c2*c3*s1*s4*s5*s6 + 0.0099495*c6*s1*s2*s3*s4*s5 + 0.0099304*s1*s2*s3*s4*s5*s6) + 0.5*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*(0.046806*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0056959*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0056959*s2_10_3*c1*c5 + 0.046806*s2_10_3*c1*s5)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) + 0.5*c4*(c2*c3 + s2*s3)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*c1*(0.040283*s2_10_3*s4_10_5 + 0.004902*s2_10_3*c4_5 + 0.040283*s2_10_3*s4_5 - 0.0098041*c2_10_3*c5 - 0.080565*c2_10_3*s5 - 0.004902*c4_10_5*s2_10_3)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*s1*(0.0099495*c6*(c2_10_3*c5 + s2_10_3*s4*s5) + 0.16636*c2_10_3*s5 + 0.0099304*s6*(c2_10_3*c5 + s2_10_3*s4*s5) - 0.16636*s2_10_3*c5*s4)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*(0.040283*s2_10_3*s4_10_5 + 0.004902*s2_10_3*c4_5 + 0.040283*s2_10_3*s4_5 - 0.0098041*c2_10_3*c5 - 0.080565*c2_10_3*s5 - 0.004902*c4_10_5*s2_10_3)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 0.5*c1*(0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*c3 + s2*s3)*((c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*c1*(0.046806*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0056959*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0056959*s2_10_3*c5*s1 + 0.046806*s2_10_3*s1*s5)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) + 0.5*c1*(0.0099495*c6*(c2_10_3*c5 + s2_10_3*s4*s5) + 0.16636*c2_10_3*s5 + 0.0099304*s6*(c2_10_3*c5 + s2_10_3*s4*s5) - 0.16636*s2_10_3*c5*s4)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*s1*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*c4*(c2*s3 - 1.0*c3*s2)*((c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3))) - 0.5*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)))) + qp1*(0.5*(0.16735*c1*c2 + 0.056793*c1*c2*c3 + 0.018024*c1*c2*s3 - 0.018024*c1*c3*s2 + 0.056793*c1*s2*s3)*(0.28*c1*s2 - 0.037355*s1 + 0.030156*c1*c2*c3 - 0.095022*c1*c2*s3 + 0.095022*c1*c3*s2 + 0.030156*c1*s2*s3) - 0.5*(0.16267*c2*s1 - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.14234*s1*s2*s3 + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.047228*c1*s4 - 0.01*c1 - 0.28*s1*s2 + 0.0098041*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.080565*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.047228*c2_10_3*c4*s1 - 0.080565*s2_10_3*c5*s1 + 0.0098041*s2_10_3*s1*s5 + 0.245*c2*s1*s3 - 0.245*c3*s1*s2) + 0.5*(0.28*c1*c2 + 0.095022*c1*c2*c3 + 0.030156*c1*c2*s3 - 0.030156*c1*c3*s2 + 0.095022*c1*s2*s3)*(0.16735*c1*s2 - 0.022327*s1 + 0.018024*c1*c2*c3 - 0.056793*c1*c2*s3 + 0.056793*c1*c3*s2 + 0.018024*c1*s2*s3) + 0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) - 0.5*(0.28*c2*s1 - 0.0098041*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.080565*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.245*s1*s2*s3 + 0.080565*c2_10_3*c5*s1 + 0.047228*s2_10_3*c4*s1 - 0.0098041*c2_10_3*s1*s5 + 0.245*c2*c3*s1)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(0.14754*c2*s1 + 0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1)*(0.01*c1 + 0.010004*c1*c4 - 0.0057515*c1*s4 + 0.28*s1*s2 - 0.0057515*c2_10_3*c4*s1 - 0.010004*c2_10_3*s1*s4 - 0.22719*c2*s1*s3 + 0.22719*c3*s1*s2) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 0.5*(c2*c3 + s2*s3)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + 0.5*(0.28*c2*s1 + 0.22719*s1*s2*s3 + 0.0057515*s2_10_3*c4*s1 + 0.010004*s2_10_3*s1*s4 + 0.22719*c2*c3*s1)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) + 0.5*(c2*s3 - 1.0*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 0.5*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(c2*c3 + s2*s3)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) - 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 0.5*(c2*s3 - 1.0*c3*s2)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3) + 0.5*(c2*c3 + s2*s3)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) - 0.5*(0.16267*c1*c2 - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.046806*c2_10_3*c1*c5 + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 - 0.047228*s1*s4 + 0.0098041*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.080565*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.047228*c2_10_3*c1*c4 - 0.080565*s2_10_3*c1*c5 + 0.0098041*s2_10_3*c1*s5 + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*c2*(1.0e-6*c2 - 0.0010285*s2) + 0.5*(0.28*c2*s1 + 0.16636*c5*(s1*s2*s3 + c2*c3*s1) - 0.057*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0099495*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.0099304*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.245*s1*s2*s3 + 0.0099304*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0099495*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.16636*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.245*c2*c3*s1)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(0.16735*c2*s1 + 0.056793*s1*s2*s3 + 0.056793*c2*c3*s1 + 0.018024*c2*s1*s3 - 0.018024*c3*s1*s2)*(0.037355*c1 + 0.28*s1*s2 + 0.030156*s1*s2*s3 + 0.030156*c2*c3*s1 - 0.095022*c2*s1*s3 + 0.095022*c3*s1*s2) + 0.5*(0.24912*c1*c2 - 0.035308*c1*s2)*(0.029983*c1*c2 - 0.075303*s1 + 0.21155*c1*s2) - 0.5*s2*(0.0010285*c2 + 1.0e-6*s2) - 0.5*(0.28*c1*c2 + 0.16636*c5*(c1*c2*c3 + c1*s2*s3) - 0.0099495*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.057*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0099304*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0099495*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.16636*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.245*c1*c2*c3 + 0.245*c1*s2*s3)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*(0.28*c2*s1 + 0.095022*s1*s2*s3 + 0.095022*c2*c3*s1 + 0.030156*c2*s1*s3 - 0.030156*c3*s1*s2)*(0.022327*c1 + 0.16735*s1*s2 + 0.018024*s1*s2*s3 + 0.018024*c2*c3*s1 - 0.056793*c2*s1*s3 + 0.056793*c3*s1*s2) - 0.5*(0.28*c1*c2 - 0.0098041*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.080565*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.080565*c2_10_3*c1*c5 + 0.047228*s2_10_3*c1*c4 - 0.0098041*c2_10_3*c1*s5 + 0.245*c1*c2*c3 + 0.245*c1*s2*s3)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 0.5*c2*(1.0e-6*c2 + 0.011493*s2) + 0.5*(0.21155*c1*c2 - 0.029983*c1*s2)*(0.035308*c1*c2 - 0.088677*s1 + 0.24912*c1*s2) - 0.5*(c2*s3 - 1.0*c3*s2)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) + 0.5*(c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1)*(0.01*c1 - 0.057*c1*s4 + 0.28*s1*s2 + 0.0099304*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099495*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.16636*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*c4*(s1*s2*s3 + c2*c3*s1) - 0.16636*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*c2*s1*s3 + 0.245*c3*s1*s2) + 0.5*s2*(0.011493*c2 - 1.0e-6*s2) - 0.5*(0.14754*c1*c2 + 0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 + 0.010004*c4*s1 - 0.0057515*s1*s4 + 0.0057515*c2_10_3*c1*c4 + 0.010004*c2_10_3*c1*s4 + 0.22719*c1*c2*s3 - 0.22719*c1*c3*s2) - 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 0.5*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 - 0.057*s1*s4 + 0.16636*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*c4*(c1*c2*c3 + c1*s2*s3) - 0.0099495*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*(0.24912*c2*s1 - 0.035308*s1*s2)*(0.075303*c1 + 0.029983*c2*s1 + 0.21155*s1*s2) - 0.5*(0.28*c1*c2 + 0.0057515*s2_10_3*c1*c4 + 0.010004*s2_10_3*c1*s4 + 0.22719*c1*c2*c3 + 0.22719*c1*s2*s3)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 0.5*(s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.21155*c2*s1 - 0.029983*s1*s2)*(0.088677*c1 + 0.035308*c2*s1 + 0.24912*s1*s2) - 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + 0.5*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) - 0.5*s4*(c2*s3 - 1.0*c3*s2)*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - 0.5*c4*(c2*c3 + s2*s3)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - 0.5*s4*(c2*c3 + s2*s3)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + qp6*(0.5*(c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*c1*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1)*(0.0099304*c4*c5*c6*s1 - 0.0099495*c6*s1*s4 - 0.0099304*s1*s4*s6 - 0.0099495*c4*c5*s1*s6 + 0.0099495*c1*c2*c3*c4*c6 + 0.0099304*c1*c2*c3*c4*s6 + 0.0099495*c1*c4*c6*s2*s3 - 0.0099304*c1*c2*c6*s3*s5 + 0.0099304*c1*c3*c6*s2*s5 + 0.0099304*c1*c4*s2*s3*s6 + 0.0099495*c1*c2*s3*s5*s6 - 0.0099495*c1*c3*s2*s5*s6 + 0.0099304*c1*c2*c3*c5*c6*s4 - 0.0099495*c1*c2*c3*c5*s4*s6 + 0.0099304*c1*c5*c6*s2*s3*s4 - 0.0099495*c1*c5*s2*s3*s4*s6) + 0.5*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(0.0099495*c1*c2*c4*c6*s3 - 0.0099495*c1*c3*c4*c6*s2 + 0.0099304*c1*c2*c3*c6*s5 + 0.0099304*c1*c2*c4*s3*s6 - 0.0099304*c1*c3*c4*s2*s6 - 0.0099495*c1*c2*c3*s5*s6 + 0.0099304*c1*c6*s2*s3*s5 - 0.0099495*c1*s2*s3*s5*s6 + 0.0099304*c1*c2*c5*c6*s3*s4 - 0.0099304*c1*c3*c5*c6*s2*s4 - 0.0099495*c1*c2*c5*s3*s4*s6 + 0.0099495*c1*c3*c5*s2*s4*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3)*(0.0099495*c1*c6*s4 + 0.0099304*c1*s4*s6 - 0.0099304*c1*c4*c5*c6 + 0.0099495*c1*c4*c5*s6 + 0.0099495*c2*c3*c4*c6*s1 + 0.0099304*c2*c3*c4*s1*s6 + 0.0099495*c4*c6*s1*s2*s3 - 0.0099304*c2*c6*s1*s3*s5 + 0.0099304*c3*c6*s1*s2*s5 + 0.0099304*c4*s1*s2*s3*s6 + 0.0099495*c2*s1*s3*s5*s6 - 0.0099495*c3*s1*s2*s5*s6 + 0.0099304*c2*c3*c5*c6*s1*s4 - 0.0099495*c2*c3*c5*s1*s4*s6 + 0.0099304*c5*c6*s1*s2*s3*s4 - 0.0099495*c5*s1*s2*s3*s4*s6) + 0.5*(0.0099495*c2*c4*c6*s1*s3 - 0.0099495*c3*c4*c6*s1*s2 + 0.0099304*c2*c3*c6*s1*s5 + 0.0099304*c2*c4*s1*s3*s6 - 0.0099304*c3*c4*s1*s2*s6 - 0.0099495*c2*c3*s1*s5*s6 + 0.0099304*c6*s1*s2*s3*s5 - 0.0099495*s1*s2*s3*s5*s6 + 0.0099304*c2*c5*c6*s1*s3*s4 - 0.0099304*c3*c5*c6*s1*s2*s4 - 0.0099495*c2*c5*s1*s3*s4*s6 + 0.0099495*c3*c5*s1*s2*s4*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*s1*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c1*(0.0099495*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) + 0.0099495*s2_10_3*c4*c6 + 0.0099304*s2_10_3*c4*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*c1*(0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*s1*(0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 0.5*s1*(0.0099495*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) + 0.0099495*s2_10_3*c4*c6 + 0.0099304*s2_10_3*c4*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)) - 1.0*qp4*(0.5*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3)*(0.057*c1*c4 - 0.0099304*c1*c4*c6 + 0.0099495*c1*c4*s6 + 0.16636*c1*s4*s5 + 0.0099495*c1*c5*c6*s4 - 0.057*c2*c3*s1*s4 + 0.0099304*c1*c5*s4*s6 - 0.057*s1*s2*s3*s4 + 0.16636*c2*c3*c4*s1*s5 + 0.0099304*c2*c3*c6*s1*s4 - 0.0099495*c2*c3*s1*s4*s6 + 0.16636*c4*s1*s2*s3*s5 + 0.0099304*c6*s1*s2*s3*s4 - 0.0099495*s1*s2*s3*s4*s6 + 0.0099495*c2*c3*c4*c5*c6*s1 + 0.0099304*c2*c3*c4*c5*s1*s6 + 0.0099495*c4*c5*c6*s1*s2*s3 + 0.0099304*c4*c5*s1*s2*s3*s6) + 0.5*(s1*s2*s3 + c2*c3*s1)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.010004*c2*c4*s1*s3 - 0.010004*c3*c4*s1*s2 - 0.0057515*c2*s1*s3*s4 + 0.0057515*c3*s1*s2*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 0.5*(c1*c2*c3 + c1*s2*s3)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3)) + 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.057*c1*c2*s3*s4 - 0.057*c1*c3*s2*s4 - 0.16636*c1*c2*c4*s3*s5 + 0.16636*c1*c3*c4*s2*s5 - 0.0099304*c1*c2*c6*s3*s4 + 0.0099304*c1*c3*c6*s2*s4 + 0.0099495*c1*c2*s3*s4*s6 - 0.0099495*c1*c3*s2*s4*s6 - 0.0099495*c1*c2*c4*c5*c6*s3 + 0.0099495*c1*c3*c4*c5*c6*s2 - 0.0099304*c1*c2*c4*c5*s3*s6 + 0.0099304*c1*c3*c4*c5*s2*s6) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*(1.0*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - (c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + 1.0*(s1*s2*s3 + c2*c3*s1)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*(s1*s2*s3 + c2*c3*s1)*((c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(c2*s3 - 1.0*c3*s2)*((c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c1*c2*s3 - 1.0*c1*c3*s2) + 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - 1.0*(0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + (2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2))*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 0.5*(c1*c2*c3 + c1*s2*s3)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.16267*c1*c2 - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.046806*c2_10_3*c1*c5 + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.047228*c1*c4 + 0.0098041*c1*c5*s4 + 0.080565*c1*s4*s5 - 0.047228*c2*c3*s1*s4 - 0.047228*s1*s2*s3*s4 + 0.0098041*c2*c3*c4*c5*s1 + 0.080565*c2*c3*c4*s1*s5 + 0.0098041*c4*c5*s1*s2*s3 + 0.080565*c4*s1*s2*s3*s5) + 0.5*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.057*c2*s1*s3*s4 - 0.057*c3*s1*s2*s4 - 0.16636*c2*c4*s1*s3*s5 + 0.16636*c3*c4*s1*s2*s5 - 0.0099304*c2*c6*s1*s3*s4 + 0.0099304*c3*c6*s1*s2*s4 + 0.0099495*c2*s1*s3*s4*s6 - 0.0099495*c3*s1*s2*s4*s6 - 0.0099495*c2*c4*c5*c6*s1*s3 + 0.0099495*c3*c4*c5*c6*s1*s2 - 0.0099304*c2*c4*c5*s1*s3*s6 + 0.0099304*c3*c4*c5*s1*s2*s6) + 0.5*(0.14754*c1*c2 + 0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3)*(0.0057515*c1*c4 + 0.010004*c1*s4 + 0.010004*c2*c3*c4*s1 - 0.0057515*c2*c3*s1*s4 + 0.010004*c4*s1*s2*s3 - 0.0057515*s1*s2*s3*s4) + 0.5*(0.047228*c2*s1*s3*s4 - 0.047228*c3*s1*s2*s4 - 0.0098041*c2*c4*c5*s1*s3 + 0.0098041*c3*c4*c5*s1*s2 - 0.080565*c2*c4*s1*s3*s5 + 0.080565*c3*c4*s1*s2*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + (c1*c2*c3 + c1*s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + 1.0*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + s4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 0.5*c1*(1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2)) - (7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c2*s1*s3 - 1.0*c3*s1*s2) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - (0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))) + 0.5*(0.14754*c2*s1 + 0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1)*(0.0057515*c4*s1 + 0.010004*s1*s4 - 0.010004*c1*c2*c3*c4 + 0.0057515*c1*c2*c3*s4 - 0.010004*c1*c4*s2*s3 + 0.0057515*c1*s2*s3*s4) - 0.5*c1*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*c3 + s2*s3)*(1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - (c2*c3 + s2*s3)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) - (c2*s3 - 1.0*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) + 1.0*s4*(c2*s3 - 1.0*c3*s2)*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + c4*(c2*c3 + s2*s3)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*c3 + s2*s3)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(s1*s2*s3 + c2*c3*s1)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.16267*c2*s1 - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.14234*s1*s2*s3 + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.047228*c4*s1 + 0.080565*s1*s4*s5 + 0.0098041*c5*s1*s4 + 0.047228*c1*c2*c3*s4 + 0.047228*c1*s2*s3*s4 - 0.0098041*c1*c2*c3*c4*c5 - 0.080565*c1*c2*c3*c4*s5 - 0.0098041*c1*c4*c5*s2*s3 - 0.080565*c1*c4*s2*s3*s5) + 0.5*(c2*c3 + s2*s3)*((c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3))) - 0.5*(c1*c2*c3 + c1*s2*s3)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*c3 + s2*s3)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.047228*c1*c2*s3*s4 - 0.047228*c1*c3*s2*s4 - 0.0098041*c1*c2*c4*c5*s3 + 0.0098041*c1*c3*c4*c5*s2 - 0.080565*c1*c2*c4*s3*s5 + 0.080565*c1*c3*c4*s2*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.010004*c1*c2*c4*s3 - 0.010004*c1*c3*c4*s2 - 0.0057515*c1*c2*s3*s4 + 0.0057515*c1*c3*s2*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) - 0.5*(c2*s3 - 1.0*c3*s2)*((c2*c3 + s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) + s4*(c2*s3 - 1.0*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + 1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3)) + 0.5*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1)*(0.057*c4*s1 + 0.16636*s1*s4*s5 - 0.0099304*c4*c6*s1 + 0.0099495*c4*s1*s6 + 0.057*c1*c2*c3*s4 + 0.0099495*c5*c6*s1*s4 + 0.057*c1*s2*s3*s4 + 0.0099304*c5*s1*s4*s6 - 0.16636*c1*c2*c3*c4*s5 - 0.0099304*c1*c2*c3*c6*s4 + 0.0099495*c1*c2*c3*s4*s6 - 0.16636*c1*c4*s2*s3*s5 - 0.0099304*c1*c6*s2*s3*s4 + 0.0099495*c1*s2*s3*s4*s6 - 0.0099495*c1*c2*c3*c4*c5*c6 - 0.0099304*c1*c2*c3*c4*c5*s6 - 0.0099495*c1*c4*c5*c6*s2*s3 - 0.0099304*c1*c4*c5*s2*s3*s6) - 0.5*c1*(0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*c1*(0.010004*s2_10_3*c4 - 0.0057515*s2_10_3*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) + 0.5*s1*(0.16636*s2_10_3*c4*s5 - 0.057*s2_10_3*s4 + 0.0099304*s2_10_3*c6*s4 - 0.0099495*s2_10_3*s4*s6 + 0.0099495*s2_10_3*c4*c5*c6 + 0.0099304*s2_10_3*c4*c5*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*s1*(0.027438*c4*s1 - 0.0056959*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.046806*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.027438*c2_10_3*c1*s4)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) - 0.5*c1*(0.004902*s2_10_3*c4_5 - 0.040283*s2_10_3*s4_10_5 + 0.040283*s2_10_3*s4_5 - 0.047228*s2_10_3*s4 + 0.004902*c4_10_5*s2_10_3)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - 0.5*c1*(0.0030306*c1*c4 + 0.0052716*c1*s4 + 0.0052716*c2_10_3*c4*s1 - 0.0030306*c2_10_3*s1*s4)*(0.22719*c2_10_3 + 0.28*c2 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4) + 0.5*s1*(0.004902*s2_10_3*c4_5 - 0.040283*s2_10_3*s4_10_5 + 0.040283*s2_10_3*s4_5 - 0.047228*s2_10_3*s4 + 0.004902*c4_10_5*s2_10_3)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*s1*(0.0030306*c4*s1 + 0.0052716*s1*s4 - 0.0052716*c2_10_3*c1*c4 + 0.0030306*c2_10_3*c1*s4)*(0.22719*c2_10_3 + 0.28*c2 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4) + 0.5*c1*(0.16636*s2_10_3*c4*s5 - 0.057*s2_10_3*s4 + 0.0099304*s2_10_3*c6*s4 - 0.0099495*s2_10_3*s4*s6 + 0.0099495*s2_10_3*c4*c5*c6 + 0.0099304*s2_10_3*c4*c5*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) - 0.5*c1*(0.027438*c1*c4 + 0.0056959*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.046806*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.027438*c2_10_3*s1*s4)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) - 0.5*s1*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*s1*(0.010004*s2_10_3*c4 - 0.0057515*s2_10_3*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2)) - qp2*(s1*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) - s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 1.0*c1*(1.0*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - (c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + 1.0*(s1*s2*s3 + c2*c3*s1)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - c1*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 1.0*c1*((s1*s2*s3 + c2*c3*s1)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) - 1.0*(s1*s2*s3 + c2*c3*s1)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3) + (c2*s1*s3 - 1.0*c3*s1*s2)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + (c2*s1*s3 - 1.0*c3*s1*s2)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) - 1.0*c1*(1.5e-6*c2*c3 + 0.000096*c2*s3 - 0.000096*c3*s2 + 1.5e-6*s2*s3)) + 1.0*s1*(s1*(1.6e-7*c2 - 0.0014077*s2) + c1*c2*(0.0010285*c2 + 1.0e-6*s2) + c1*s2*(1.0e-6*c2 - 0.0010285*s2) - 1.0*c1*c2*(0.011493*c2 - 1.0e-6*s2) + c1*s2*(1.0e-6*c2 + 0.011493*s2)) + c1*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 1.0*s1*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + (c1*c2*c3 + c1*s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + 1.0*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + s4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - s1*((c1*c2*c3 + c1*s2*s3)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) + (c1*c2*s3 - 1.0*c1*c3*s2)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + (c1*c2*s3 - 1.0*c1*c3*s2)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) + s1*(1.5e-6*c2*c3 + 0.000096*c2*s3 - 0.000096*c3*s2 + 1.5e-6*s2*s3) - 1.0*(c1*c2*c3 + c1*s2*s3)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3)) - c1*(c2*s1*(0.0010285*c2 + 1.0e-6*s2) - 1.0*c1*(1.6e-7*c2 - 0.0014077*s2) + s1*s2*(1.0e-6*c2 - 0.0010285*s2) - 1.0*c2*s1*(0.011493*c2 - 1.0e-6*s2) + s1*s2*(1.0e-6*c2 + 0.011493*s2)) - s1*(0.22719*c2_10_3 + 0.28*c2 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.14754*c1*c2 + 0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3) - s1*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5)*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3) - c1*(0.030156*c2_10_3 + 0.095022*s2_10_3 + 0.28*s2)*(0.022327*c1 + 0.16735*s1*s2 + 0.018024*s1*s2*s3 + 0.018024*c2*c3*s1 - 0.056793*c2*s1*s3 + 0.056793*c3*s1*s2) + 1.0*s1*(0.245*c2*s3 - 0.28*s2 - 0.245*c3*s2 + 0.057*c2_10_3*c4 + 0.0099495*c6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.16636*s2_10_3*c5 + 0.0099304*s6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.0099304*c2_10_3*c4*c6 + 0.0099495*c2_10_3*c4*s6 + 0.16636*c2_10_3*s4*s5)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - s1*(0.095022*c2_10_3 - 0.030156*s2_10_3 + 0.28*c2)*(0.16735*c1*c2 + 0.056793*c1*c2*c3 + 0.018024*c1*c2*s3 - 0.018024*c1*c3*s2 + 0.056793*c1*s2*s3) - 1.3878e-17*s1*(0.24912*c1*c2 - 0.035308*c1*s2)*(1.5244e+16*c2 - 2.1605e+15*s2) + 1.0*c1*(0.22719*c2_10_3 + 0.28*c2 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.14754*c2*s1 + 0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1) + 1.3878e-17*c1*(0.24912*c2*s1 - 0.035308*s1*s2)*(1.5244e+16*c2 - 2.1605e+15*s2) + 1.0*c1*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5)*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1) + 1.0*c1*(0.16267*c2*s1 - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.14234*s1*s2*s3 + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) + 1.0*c1*(0.095022*c2_10_3 - 0.030156*s2_10_3 + 0.28*c2)*(0.16735*c2*s1 + 0.056793*s1*s2*s3 + 0.056793*c2*c3*s1 + 0.018024*c2*s1*s3 - 0.018024*c3*s1*s2) - s1*(0.22719*s2_10_3 + 0.28*s2 - 0.0057515*c2_10_3*c4 - 0.010004*c2_10_3*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) - c1*(0.22719*s2_10_3 + 0.28*s2 - 0.0057515*c2_10_3*c4 - 0.010004*c2_10_3*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) - c1*(0.004902*c2_10_3*s4_5 - 0.28*s2 - 0.040283*c2_10_3*c4_5 - 0.245*s2_10_3 + 0.047228*c2_10_3*c4 - 0.080565*s2_10_3*c5 + 0.0098041*s2_10_3*s5 + 0.040283*c2_10_3*c4_10_5 + 0.004902*c2_10_3*s4_10_5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 1.0*c1*(0.245*c2*s3 - 0.28*s2 - 0.245*c3*s2 + 0.057*c2_10_3*c4 + 0.0099495*c6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.16636*s2_10_3*c5 + 0.0099304*s6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.0099304*c2_10_3*c4*c6 + 0.0099495*c2_10_3*c4*s6 + 0.16636*c2_10_3*s4*s5)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 1.0*s1*(0.004902*c2_10_3*s4_5 - 0.28*s2 - 0.040283*c2_10_3*c4_5 - 0.245*s2_10_3 + 0.047228*c2_10_3*c4 - 0.080565*s2_10_3*c5 + 0.0098041*s2_10_3*s5 + 0.040283*c2_10_3*c4_10_5 + 0.004902*c2_10_3*s4_10_5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - s1*(0.16267*c1*c2 - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.046806*c2_10_3*c1*c5 + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) - 1.3878e-17*c1*(2.1605e+15*c2 + 1.5244e+16*s2)*(0.088677*c1 + 0.035308*c2*s1 + 0.24912*s1*s2) + 1.3878e-17*s1*(2.1605e+15*c2 + 1.5244e+16*s2)*(0.035308*c1*c2 - 0.088677*s1 + 0.24912*c1*s2) + 1.0*s1*(0.030156*c2_10_3 + 0.095022*s2_10_3 + 0.28*s2)*(0.16735*c1*s2 - 0.022327*s1 + 0.018024*c1*c2*c3 - 0.056793*c1*c2*s3 + 0.056793*c1*c3*s2 + 0.018024*c1*s2*s3)),  qp2*(1.0*s1*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) - 1.0*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - c1*(1.0*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - (c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + 1.0*(s1*s2*s3 + c2*c3*s1)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 1.0*c1*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + c1*((s1*s2*s3 + c2*c3*s1)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) - 1.0*(s1*s2*s3 + c2*c3*s1)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3) + (c2*s1*s3 - 1.0*c3*s1*s2)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + (c2*s1*s3 - 1.0*c3*s1*s2)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) - 1.0*c1*(1.5e-6*c2*c3 + 0.000096*c2*s3 - 0.000096*c3*s2 + 1.5e-6*s2*s3)) + 1.0*c1*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + s1*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + (c1*c2*c3 + c1*s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + 1.0*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + s4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 1.0*s1*((c1*c2*c3 + c1*s2*s3)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) + (c1*c2*s3 - 1.0*c1*c3*s2)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + (c1*c2*s3 - 1.0*c1*c3*s2)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) + s1*(1.5e-6*c2*c3 + 0.000096*c2*s3 - 0.000096*c3*s2 + 1.5e-6*s2*s3) - 1.0*(c1*c2*c3 + c1*s2*s3)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3)) + 0.5*c1*(0.095022*c2_10_3 - 0.030156*s2_10_3)*(0.16735*c2*s1 + 0.056793*s1*s2*s3 + 0.056793*c2*c3*s1 + 0.018024*c2*s1*s3 - 0.018024*c3*s1*s2) + s1*(0.0057515*c2_10_3*c4 - 0.22719*s2_10_3 + 0.010004*c2_10_3*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 0.5*c1*(0.14234*s1*s2*s3 - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) + 0.5*c1*(0.095022*c2_10_3 - 0.030156*s2_10_3 + 0.28*c2)*(0.056793*s1*s2*s3 + 0.056793*c2*c3*s1 + 0.018024*c2*s1*s3 - 0.018024*c3*s1*s2) + c1*(0.0057515*c2_10_3*c4 - 0.22719*s2_10_3 + 0.010004*c2_10_3*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) - 0.5*c1*(0.004902*c2_10_3*s4_5 - 0.040283*c2_10_3*c4_5 - 0.245*s2_10_3 + 0.047228*c2_10_3*c4 - 0.080565*s2_10_3*c5 + 0.0098041*s2_10_3*s5 + 0.040283*c2_10_3*c4_10_5 + 0.004902*c2_10_3*s4_10_5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + c1*(0.245*c2*s3 - 0.245*c3*s2 + 0.057*c2_10_3*c4 + 0.0099495*c6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.16636*s2_10_3*c5 + 0.0099304*s6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.0099304*c2_10_3*c4*c6 + 0.0099495*c2_10_3*c4*s6 + 0.16636*c2_10_3*s4*s5)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*s1*(0.004902*c2_10_3*s4_5 - 0.040283*c2_10_3*c4_5 - 0.245*s2_10_3 + 0.047228*c2_10_3*c4 - 0.080565*s2_10_3*c5 + 0.0098041*s2_10_3*s5 + 0.040283*c2_10_3*c4_10_5 + 0.004902*c2_10_3*s4_10_5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 6.9389e-21*c1*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5)*(0.16267*c2*s1 - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.14234*s1*s2*s3 + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1) + s1*(0.030156*c2_10_3 + 0.095022*s2_10_3)*(0.16735*c1*s2 - 0.022327*s1 + 0.018024*c1*c2*c3 - 0.056793*c1*c2*s3 + 0.056793*c1*c3*s2 + 0.018024*c1*s2*s3) - 0.5*s1*(0.046806*c2_10_3*c1*c5 - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) - 0.5*s1*(0.22719*c2_10_3 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.14754*c1*c2 + 0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3) - 1.0*c1*(0.030156*c2_10_3 + 0.095022*s2_10_3)*(0.022327*c1 + 0.16735*s1*s2 + 0.018024*s1*s2*s3 + 0.018024*c2*c3*s1 - 0.056793*c2*s1*s3 + 0.056793*c3*s1*s2) + 0.5*s1*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 0.5*s1*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5)*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3) + s1*(0.245*c2*s3 - 0.245*c3*s2 + 0.057*c2_10_3*c4 + 0.0099495*c6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.16636*s2_10_3*c5 + 0.0099304*s6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.0099304*c2_10_3*c4*c6 + 0.0099495*c2_10_3*c4*s6 + 0.16636*c2_10_3*s4*s5)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*s1*(0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3)*(0.22719*c2_10_3 + 0.28*c2 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4) - 6.9389e-21*c1*(3.4031e+18*c2_10_3*c4 - 1.7654e+19*s2_10_3 - 5.8053e+18*s2_10_3*c5 + 7.0646e+17*s2_10_3*s5 + 7.0646e+17*c2_10_3*c5*s4 + 5.8053e+18*c2_10_3*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - 0.5*s1*(0.095022*c2_10_3 - 0.030156*s2_10_3)*(0.16735*c1*c2 + 0.056793*c1*c2*c3 + 0.018024*c1*c2*s3 - 0.018024*c1*c3*s2 + 0.056793*c1*s2*s3) + 0.5*c1*(0.22719*c2_10_3 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.14754*c2*s1 + 0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1) - 0.5*c1*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 6.9389e-21*s1*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5)*(0.16267*c1*c2 - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.046806*c2_10_3*c1*c5 + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3) - 0.5*s1*(0.095022*c2_10_3 - 0.030156*s2_10_3 + 0.28*c2)*(0.056793*c1*c2*c3 + 0.018024*c1*c2*s3 - 0.018024*c1*c3*s2 + 0.056793*c1*s2*s3) + 0.5*c1*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5)*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1) + 0.5*c1*(0.22719*c2_10_3 + 0.28*c2 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1) + 6.9389e-21*s1*(3.4031e+18*c2_10_3*c4 - 1.7654e+19*s2_10_3 - 5.8053e+18*s2_10_3*c5 + 7.0646e+17*s2_10_3*s5 + 7.0646e+17*c2_10_3*c5*s4 + 5.8053e+18*c2_10_3*s4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2)) - qp5*(0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0098041*c1*c2*c3*c5 + 0.080565*c1*c2*c3*s5 + 0.0098041*c1*c5*s2*s3 + 0.080565*c1*s2*s3*s5 + 0.080565*c1*c2*c5*s3*s4 - 0.080565*c1*c3*c5*s2*s4 - 0.0098041*c1*c2*s3*s4*s5 + 0.0098041*c1*c3*s2*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.16636*c1*c2*c3*s5 + 0.16636*c1*s2*s3*s5 + 0.0099495*c1*c2*c3*c5*c6 + 0.0099304*c1*c2*c3*c5*s6 + 0.16636*c1*c2*c5*s3*s4 - 0.16636*c1*c3*c5*s2*s4 + 0.0099495*c1*c5*c6*s2*s3 + 0.0099304*c1*c5*s2*s3*s6 - 0.0099495*c1*c2*c6*s3*s4*s5 + 0.0099495*c1*c3*c6*s2*s4*s5 - 0.0099304*c1*c2*s3*s4*s5*s6 + 0.0099304*c1*c3*s2*s4*s5*s6) + 0.5*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.16636*c1*c2*s3*s5 - 0.16636*c4*c5*s1 - 0.16636*c1*c3*s2*s5 + 0.0099495*c4*c6*s1*s5 + 0.0099304*c4*s1*s5*s6 - 0.16636*c1*c2*c3*c5*s4 + 0.0099495*c1*c2*c5*c6*s3 - 0.0099495*c1*c3*c5*c6*s2 + 0.0099304*c1*c2*c5*s3*s6 - 0.0099304*c1*c3*c5*s2*s6 - 0.16636*c1*c5*s2*s3*s4 + 0.0099495*c1*c2*c3*c6*s4*s5 + 0.0099304*c1*c2*c3*s4*s5*s6 + 0.0099495*c1*c6*s2*s3*s4*s5 + 0.0099304*c1*s2*s3*s4*s5*s6) + 0.5*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.16636*c2*c3*s1*s5 + 0.16636*s1*s2*s3*s5 + 0.0099495*c2*c3*c5*c6*s1 + 0.0099304*c2*c3*c5*s1*s6 + 0.16636*c2*c5*s1*s3*s4 - 0.16636*c3*c5*s1*s2*s4 + 0.0099495*c5*c6*s1*s2*s3 + 0.0099304*c5*s1*s2*s3*s6 - 0.0099495*c2*c6*s1*s3*s4*s5 + 0.0099495*c3*c6*s1*s2*s4*s5 - 0.0099304*c2*s1*s3*s4*s5*s6 + 0.0099304*c3*s1*s2*s4*s5*s6) + 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) - 0.5*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.16636*c1*c4*c5 - 0.0099495*c1*c4*c6*s5 - 0.0099304*c1*c4*s5*s6 + 0.16636*c2*s1*s3*s5 - 0.16636*c3*s1*s2*s5 - 0.16636*c2*c3*c5*s1*s4 + 0.0099495*c2*c5*c6*s1*s3 - 0.0099495*c3*c5*c6*s1*s2 + 0.0099304*c2*c5*s1*s3*s6 - 0.0099304*c3*c5*s1*s2*s6 - 0.16636*c5*s1*s2*s3*s4 + 0.0099495*c2*c3*c6*s1*s4*s5 + 0.0099304*c2*c3*s1*s4*s5*s6 + 0.0099495*c6*s1*s2*s3*s4*s5 + 0.0099304*s1*s2*s3*s4*s5*s6) - 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) - 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0098041*c2*c3*c5*s1 + 0.080565*c2*c3*s1*s5 + 0.0098041*c5*s1*s2*s3 + 0.080565*s1*s2*s3*s5 + 0.080565*c2*c5*s1*s3*s4 - 0.080565*c3*c5*s1*s2*s4 - 0.0098041*c2*s1*s3*s4*s5 + 0.0098041*c3*s1*s2*s4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 0.5*(0.046806*c2_10_3*c1*c5 - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.080565*c1*c4*c5 - 0.0098041*c1*c4*s5 + 0.0098041*c2*c5*s1*s3 - 0.0098041*c3*c5*s1*s2 + 0.080565*c2*s1*s3*s5 - 0.080565*c3*s1*s2*s5 - 0.080565*c2*c3*c5*s1*s4 + 0.0098041*c2*c3*s1*s4*s5 - 0.080565*c5*s1*s2*s3*s4 + 0.0098041*s1*s2*s3*s4*s5) - 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.14234*s1*s2*s3 - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.080565*c4*c5*s1 - 0.0098041*c4*s1*s5 - 0.0098041*c1*c2*c5*s3 + 0.0098041*c1*c3*c5*s2 - 0.080565*c1*c2*s3*s5 + 0.080565*c1*c3*s2*s5 + 0.080565*c1*c2*c3*c5*s4 - 0.0098041*c1*c2*c3*s4*s5 + 0.080565*c1*c5*s2*s3*s4 - 0.0098041*c1*s2*s3*s4*s5) - 6.9389e-21*s1*(0.046806*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0056959*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0056959*s2_10_3*c1*c5 + 0.046806*s2_10_3*c1*s5)*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5) + 0.5*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*c4*(c2*c3 + s2*s3)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*c1*(0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*s1*(0.0099495*c6*(c2_10_3*c5 + s2_10_3*s4*s5) + 0.16636*c2_10_3*s5 + 0.0099304*s6*(c2_10_3*c5 + s2_10_3*s4*s5) - 0.16636*s2_10_3*c5*s4)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*s1*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 6.9389e-21*c1*(7.0646e+17*c2_10_3*c5 + 5.8053e+18*c2_10_3*s5 - 5.8053e+18*s2_10_3*c5*s4 + 7.0646e+17*s2_10_3*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*c3 + s2*s3)*((c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 6.9389e-21*c1*(0.046806*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0056959*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0056959*s2_10_3*c5*s1 + 0.046806*s2_10_3*s1*s5)*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5) + 0.5*c1*(0.0099495*c6*(c2_10_3*c5 + s2_10_3*s4*s5) + 0.16636*c2_10_3*s5 + 0.0099304*s6*(c2_10_3*c5 + s2_10_3*s4*s5) - 0.16636*s2_10_3*c5*s4)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*c4*(c2*s3 - 1.0*c3*s2)*((c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3))) + 6.9389e-21*s1*(7.0646e+17*c2_10_3*c5 + 5.8053e+18*c2_10_3*s5 - 5.8053e+18*s2_10_3*c5*s4 + 7.0646e+17*s2_10_3*s4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)))) - 1.0*qp3*(1.0*s1*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) - 1.0*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - c1*(1.0*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - (c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + 1.0*(s1*s2*s3 + c2*c3*s1)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 1.0*c1*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + c1*((s1*s2*s3 + c2*c3*s1)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) - 1.0*(s1*s2*s3 + c2*c3*s1)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3) + (c2*s1*s3 - 1.0*c3*s1*s2)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + (c2*s1*s3 - 1.0*c3*s1*s2)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) - 1.0*c1*(1.5e-6*c2*c3 + 0.000096*c2*s3 - 0.000096*c3*s2 + 1.5e-6*s2*s3)) + 1.0*c1*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + s1*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + (c1*c2*c3 + c1*s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + 1.0*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + s4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 1.0*s1*((c1*c2*c3 + c1*s2*s3)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) + (c1*c2*s3 - 1.0*c1*c3*s2)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + (c1*c2*s3 - 1.0*c1*c3*s2)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) + s1*(1.5e-6*c2*c3 + 0.000096*c2*s3 - 0.000096*c3*s2 + 1.5e-6*s2*s3) - 1.0*(c1*c2*c3 + c1*s2*s3)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3)) + s1*(0.0057515*c2_10_3*c4 - 0.22719*s2_10_3 + 0.010004*c2_10_3*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 1.3878e-20*c1*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5)*(0.14234*s1*s2*s3 - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1) + c1*(0.0057515*c2_10_3*c4 - 0.22719*s2_10_3 + 0.010004*c2_10_3*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) + s1*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + c1*(0.245*c2*s3 - 0.245*c3*s2 + 0.057*c2_10_3*c4 + 0.0099495*c6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.16636*s2_10_3*c5 + 0.0099304*s6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.0099304*c2_10_3*c4*c6 + 0.0099495*c2_10_3*c4*s6 + 0.16636*c2_10_3*s4*s5)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) - 1.0*s1*(0.22719*c2_10_3 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3) - 1.0*c1*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 1.0*s1*(0.095022*c2_10_3 - 0.030156*s2_10_3)*(0.056793*c1*c2*c3 + 0.018024*c1*c2*s3 - 0.018024*c1*c3*s2 + 0.056793*c1*s2*s3) + c1*(0.22719*c2_10_3 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1) + s1*(0.030156*c2_10_3 + 0.095022*s2_10_3)*(0.16735*c1*s2 - 0.022327*s1 + 0.018024*c1*c2*c3 - 0.056793*c1*c2*s3 + 0.056793*c1*c3*s2 + 0.018024*c1*s2*s3) - 1.3878e-20*s1*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5)*(0.046806*c2_10_3*c1*c5 - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3) + c1*(0.095022*c2_10_3 - 0.030156*s2_10_3)*(0.056793*s1*s2*s3 + 0.056793*c2*c3*s1 + 0.018024*c2*s1*s3 - 0.018024*c3*s1*s2) - 1.0*c1*(0.030156*c2_10_3 + 0.095022*s2_10_3)*(0.022327*c1 + 0.16735*s1*s2 + 0.018024*s1*s2*s3 + 0.018024*c2*c3*s1 - 0.056793*c2*s1*s3 + 0.056793*c3*s1*s2) + s1*(0.245*c2*s3 - 0.245*c3*s2 + 0.057*c2_10_3*c4 + 0.0099495*c6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.16636*s2_10_3*c5 + 0.0099304*s6*(s2_10_3*s5 + c2_10_3*c5*s4) - 0.0099304*c2_10_3*c4*c6 + 0.0099495*c2_10_3*c4*s6 + 0.16636*c2_10_3*s4*s5)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 1.3878e-20*c1*(3.4031e+18*c2_10_3*c4 - 1.7654e+19*s2_10_3 - 5.8053e+18*s2_10_3*c5 + 7.0646e+17*s2_10_3*s5 + 7.0646e+17*c2_10_3*c5*s4 + 5.8053e+18*c2_10_3*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 1.3878e-20*s1*(3.4031e+18*c2_10_3*c4 - 1.7654e+19*s2_10_3 - 5.8053e+18*s2_10_3*c5 + 7.0646e+17*s2_10_3*s5 + 7.0646e+17*c2_10_3*c5*s4 + 5.8053e+18*c2_10_3*s4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2)) - 1.0*qp6*(0.5*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.0099495*c1*c6*s4 + 0.0099304*c1*s4*s6 - 0.0099304*c1*c4*c5*c6 + 0.0099495*c1*c4*c5*s6 + 0.0099495*c2*c3*c4*c6*s1 + 0.0099304*c2*c3*c4*s1*s6 + 0.0099495*c4*c6*s1*s2*s3 - 0.0099304*c2*c6*s1*s3*s5 + 0.0099304*c3*c6*s1*s2*s5 + 0.0099304*c4*s1*s2*s3*s6 + 0.0099495*c2*s1*s3*s5*s6 - 0.0099495*c3*s1*s2*s5*s6 + 0.0099304*c2*c3*c5*c6*s1*s4 - 0.0099495*c2*c3*c5*s1*s4*s6 + 0.0099304*c5*c6*s1*s2*s3*s4 - 0.0099495*c5*s1*s2*s3*s4*s6) - 0.5*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*c1*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + 0.5*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(0.0099495*c1*c2*c4*c6*s3 - 0.0099495*c1*c3*c4*c6*s2 + 0.0099304*c1*c2*c3*c6*s5 + 0.0099304*c1*c2*c4*s3*s6 - 0.0099304*c1*c3*c4*s2*s6 - 0.0099495*c1*c2*c3*s5*s6 + 0.0099304*c1*c6*s2*s3*s5 - 0.0099495*c1*s2*s3*s5*s6 + 0.0099304*c1*c2*c5*c6*s3*s4 - 0.0099304*c1*c3*c5*c6*s2*s4 - 0.0099495*c1*c2*c5*s3*s4*s6 + 0.0099495*c1*c3*c5*s2*s4*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0099495*c2*c4*c6*s1*s3 - 0.0099495*c3*c4*c6*s1*s2 + 0.0099304*c2*c3*c6*s1*s5 + 0.0099304*c2*c4*s1*s3*s6 - 0.0099304*c3*c4*s1*s2*s6 - 0.0099495*c2*c3*s1*s5*s6 + 0.0099304*c6*s1*s2*s3*s5 - 0.0099495*s1*s2*s3*s5*s6 + 0.0099304*c2*c5*c6*s1*s3*s4 - 0.0099304*c3*c5*c6*s1*s2*s4 - 0.0099495*c2*c5*s1*s3*s4*s6 + 0.0099495*c3*c5*s1*s2*s4*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*s1*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.0099304*c4*c5*c6*s1 - 0.0099495*c6*s1*s4 - 0.0099304*s1*s4*s6 - 0.0099495*c4*c5*s1*s6 + 0.0099495*c1*c2*c3*c4*c6 + 0.0099304*c1*c2*c3*c4*s6 + 0.0099495*c1*c4*c6*s2*s3 - 0.0099304*c1*c2*c6*s3*s5 + 0.0099304*c1*c3*c6*s2*s5 + 0.0099304*c1*c4*s2*s3*s6 + 0.0099495*c1*c2*s3*s5*s6 - 0.0099495*c1*c3*s2*s5*s6 + 0.0099304*c1*c2*c3*c5*c6*s4 - 0.0099495*c1*c2*c3*c5*s4*s6 + 0.0099304*c1*c5*c6*s2*s3*s4 - 0.0099495*c1*c5*s2*s3*s4*s6) - 0.5*c1*(0.0099495*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) + 0.0099495*s2_10_3*c4*c6 + 0.0099304*s2_10_3*c4*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) - 0.5*s1*(0.0099495*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) + 0.0099495*s2_10_3*c4*c6 + 0.0099304*s2_10_3*c4*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*c1*(0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*s1*(0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5)) - 1.0*qp1*(0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + 0.5*(0.056793*s1*s2*s3 + 0.056793*c2*c3*s1 + 0.018024*c2*s1*s3 - 0.018024*c3*s1*s2)*(0.037355*c1 + 0.28*s1*s2 + 0.030156*s1*s2*s3 + 0.030156*c2*c3*s1 - 0.095022*c2*s1*s3 + 0.095022*c3*s1*s2) - 0.5*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.01*c1 - 0.057*c1*s4 + 0.28*s1*s2 + 0.0099304*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099495*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.16636*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*c4*(s1*s2*s3 + c2*c3*s1) - 0.16636*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*c2*s1*s3 + 0.245*c3*s1*s2) - 0.5*(0.046806*c2_10_3*c1*c5 - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 - 0.047228*s1*s4 + 0.0098041*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.080565*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.047228*c2_10_3*c1*c4 - 0.080565*s2_10_3*c1*c5 + 0.0098041*s2_10_3*c1*s5 + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) - 0.5*(0.057*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.16636*c5*(s1*s2*s3 + c2*c3*s1) + 0.0099495*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*s1*s2*s3 - 0.0099304*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.16636*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.245*c2*c3*s1)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 0.5*(c2*c3 + s2*s3)*(7.1e-6*c2*c3 + 0.000298*c2*s3 - 0.000298*c3*s2 + 7.1e-6*s2*s3) + 0.5*(c2*s3 - 1.0*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 0.5*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.095022*s1*s2*s3 + 0.095022*c2*c3*s1 + 0.030156*c2*s1*s3 - 0.030156*c3*s1*s2)*(0.022327*c1 + 0.16735*s1*s2 + 0.018024*s1*s2*s3 + 0.018024*c2*c3*s1 - 0.056793*c2*s1*s3 + 0.056793*c3*s1*s2) + 0.5*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 - 0.057*s1*s4 + 0.16636*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*c4*(c1*c2*c3 + c1*s2*s3) - 0.0099495*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*(0.0099495*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.16636*c5*(c1*c2*c3 + c1*s2*s3) + 0.057*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099304*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.0099304*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.16636*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.245*c1*c2*c3 - 0.245*c1*s2*s3)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*(0.080565*c2_10_3*c1*c5 - 0.080565*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0098041*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.047228*s2_10_3*c1*c4 - 0.0098041*c2_10_3*c1*s5 + 0.245*c1*c2*c3 + 0.245*c1*s2*s3)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*(0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3)*(0.01*s1 - 0.28*c1*s2 + 0.010004*c4*s1 - 0.0057515*s1*s4 + 0.0057515*c2_10_3*c1*c4 + 0.010004*c2_10_3*c1*s4 + 0.22719*c1*c2*s3 - 0.22719*c1*c3*s2) + 0.5*(c2*c3 + s2*s3)*(7.1e-6*c2*c3 - 0.0016326*c2*s3 + 0.0016326*c3*s2 + 7.1e-6*s2*s3) - 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 0.5*(c2*s3 - 1.0*c3*s2)*(0.000298*c2*c3 - 7.1e-6*c2*s3 + 7.1e-6*c3*s2 + 0.000298*s2*s3) - 0.5*(0.0057515*s2_10_3*c1*c4 + 0.010004*s2_10_3*c1*s4 + 0.22719*c1*c2*c3 + 0.22719*c1*s2*s3)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 0.5*(c2*c3 + s2*s3)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) - 0.5*(0.14234*s1*s2*s3 - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.047228*c1*s4 - 0.01*c1 - 0.28*s1*s2 + 0.0098041*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.080565*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.047228*c2_10_3*c4*s1 - 0.080565*s2_10_3*c5*s1 + 0.0098041*s2_10_3*s1*s5 + 0.245*c2*s1*s3 - 0.245*c3*s1*s2) + 0.5*(0.056793*c1*c2*c3 + 0.018024*c1*c2*s3 - 0.018024*c1*c3*s2 + 0.056793*c1*s2*s3)*(0.28*c1*s2 - 0.037355*s1 + 0.030156*c1*c2*c3 - 0.095022*c1*c2*s3 + 0.095022*c1*c3*s2 + 0.030156*c1*s2*s3) - 0.5*(c2*s3 - 1.0*c3*s2)*(0.0016326*c2*c3 + 7.1e-6*c2*s3 - 7.1e-6*c3*s2 + 0.0016326*s2*s3) + 0.5*(c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) + 0.5*(0.095022*c1*c2*c3 + 0.030156*c1*c2*s3 - 0.030156*c1*c3*s2 + 0.095022*c1*s2*s3)*(0.16735*c1*s2 - 0.022327*s1 + 0.018024*c1*c2*c3 - 0.056793*c1*c2*s3 + 0.056793*c1*c3*s2 + 0.018024*c1*s2*s3) - 0.5*(0.245*s1*s2*s3 - 0.080565*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0098041*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.080565*c2_10_3*c5*s1 + 0.047228*s2_10_3*c4*s1 - 0.0098041*c2_10_3*s1*s5 + 0.245*c2*c3*s1)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1)*(0.01*c1 + 0.010004*c1*c4 - 0.0057515*c1*s4 + 0.28*s1*s2 - 0.0057515*c2_10_3*c4*s1 - 0.010004*c2_10_3*s1*s4 - 0.22719*c2*s1*s3 + 0.22719*c3*s1*s2) + 0.5*(s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.22719*s1*s2*s3 + 0.0057515*s2_10_3*c4*s1 + 0.010004*s2_10_3*s1*s4 + 0.22719*c2*c3*s1)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) - 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + 0.5*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) - 0.5*s4*(c2*s3 - 1.0*c3*s2)*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - 0.5*c4*(c2*c3 + s2*s3)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - 0.5*s4*(c2*c3 + s2*s3)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 1.0*qp4*(0.5*(0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3)*(0.0057515*c1*c4 + 0.010004*c1*s4 + 0.010004*c2*c3*c4*s1 - 0.0057515*c2*c3*s1*s4 + 0.010004*c4*s1*s2*s3 - 0.0057515*s1*s2*s3*s4) + 0.5*(s1*s2*s3 + c2*c3*s1)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.010004*c2*c4*s1*s3 - 0.010004*c3*c4*s1*s2 - 0.0057515*c2*s1*s3*s4 + 0.0057515*c3*s1*s2*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 0.5*(c1*c2*c3 + c1*s2*s3)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3)) + 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1)*(0.0057515*c4*s1 + 0.010004*s1*s4 - 0.010004*c1*c2*c3*c4 + 0.0057515*c1*c2*c3*s4 - 0.010004*c1*c4*s2*s3 + 0.0057515*c1*s2*s3*s4) + 0.5*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.057*c1*c2*s3*s4 - 0.057*c1*c3*s2*s4 - 0.16636*c1*c2*c4*s3*s5 + 0.16636*c1*c3*c4*s2*s5 - 0.0099304*c1*c2*c6*s3*s4 + 0.0099304*c1*c3*c6*s2*s4 + 0.0099495*c1*c2*s3*s4*s6 - 0.0099495*c1*c3*s2*s4*s6 - 0.0099495*c1*c2*c4*c5*c6*s3 + 0.0099495*c1*c3*c4*c5*c6*s2 - 0.0099304*c1*c2*c4*c5*s3*s6 + 0.0099304*c1*c3*c4*c5*s2*s6) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*(1.0*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - (c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + 1.0*(s1*s2*s3 + c2*c3*s1)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*(s1*s2*s3 + c2*c3*s1)*((c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(c2*s3 - 1.0*c3*s2)*((c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c1*c2*s3 - 1.0*c1*c3*s2) + 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - 1.0*(0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + (2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2))*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 0.5*(c1*c2*c3 + c1*s2*s3)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.057*c2*s1*s3*s4 - 0.057*c3*s1*s2*s4 - 0.16636*c2*c4*s1*s3*s5 + 0.16636*c3*c4*s1*s2*s5 - 0.0099304*c2*c6*s1*s3*s4 + 0.0099304*c3*c6*s1*s2*s4 + 0.0099495*c2*s1*s3*s4*s6 - 0.0099495*c3*s1*s2*s4*s6 - 0.0099495*c2*c4*c5*c6*s1*s3 + 0.0099495*c3*c4*c5*c6*s1*s2 - 0.0099304*c2*c4*c5*s1*s3*s6 + 0.0099304*c3*c4*c5*s1*s2*s6) + 0.5*(0.14234*s1*s2*s3 - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.047228*c4*s1 + 0.080565*s1*s4*s5 + 0.0098041*c5*s1*s4 + 0.047228*c1*c2*c3*s4 + 0.047228*c1*s2*s3*s4 - 0.0098041*c1*c2*c3*c4*c5 - 0.080565*c1*c2*c3*c4*s5 - 0.0098041*c1*c4*c5*s2*s3 - 0.080565*c1*c4*s2*s3*s5) + 0.5*(0.047228*c2*s1*s3*s4 - 0.047228*c3*s1*s2*s4 - 0.0098041*c2*c4*c5*s1*s3 + 0.0098041*c3*c4*c5*s1*s2 - 0.080565*c2*c4*s1*s3*s5 + 0.080565*c3*c4*s1*s2*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.057*c4*s1 + 0.16636*s1*s4*s5 - 0.0099304*c4*c6*s1 + 0.0099495*c4*s1*s6 + 0.057*c1*c2*c3*s4 + 0.0099495*c5*c6*s1*s4 + 0.057*c1*s2*s3*s4 + 0.0099304*c5*s1*s4*s6 - 0.16636*c1*c2*c3*c4*s5 - 0.0099304*c1*c2*c3*c6*s4 + 0.0099495*c1*c2*c3*s4*s6 - 0.16636*c1*c4*s2*s3*s5 - 0.0099304*c1*c6*s2*s3*s4 + 0.0099495*c1*s2*s3*s4*s6 - 0.0099495*c1*c2*c3*c4*c5*c6 - 0.0099304*c1*c2*c3*c4*c5*s6 - 0.0099495*c1*c4*c5*c6*s2*s3 - 0.0099304*c1*c4*c5*s2*s3*s6) + 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + (c1*c2*c3 + c1*s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + 1.0*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + s4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 0.5*c1*(1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2)) - (7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c2*s1*s3 - 1.0*c3*s1*s2) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - (0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))) - 0.5*c1*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.057*c1*c4 - 0.0099304*c1*c4*c6 + 0.0099495*c1*c4*s6 + 0.16636*c1*s4*s5 + 0.0099495*c1*c5*c6*s4 - 0.057*c2*c3*s1*s4 + 0.0099304*c1*c5*s4*s6 - 0.057*s1*s2*s3*s4 + 0.16636*c2*c3*c4*s1*s5 + 0.0099304*c2*c3*c6*s1*s4 - 0.0099495*c2*c3*s1*s4*s6 + 0.16636*c4*s1*s2*s3*s5 + 0.0099304*c6*s1*s2*s3*s4 - 0.0099495*s1*s2*s3*s4*s6 + 0.0099495*c2*c3*c4*c5*c6*s1 + 0.0099304*c2*c3*c4*c5*s1*s6 + 0.0099495*c4*c5*c6*s1*s2*s3 + 0.0099304*c4*c5*s1*s2*s3*s6) + 0.5*(c2*c3 + s2*s3)*(1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - (c2*c3 + s2*s3)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) - (c2*s3 - 1.0*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) + 1.0*s4*(c2*s3 - 1.0*c3*s2)*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + c4*(c2*c3 + s2*s3)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*c3 + s2*s3)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(s1*s2*s3 + c2*c3*s1)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*c3 + s2*s3)*((c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3))) - 0.5*(c1*c2*c3 + c1*s2*s3)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*c3 + s2*s3)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.047228*c1*c2*s3*s4 - 0.047228*c1*c3*s2*s4 - 0.0098041*c1*c2*c4*c5*s3 + 0.0098041*c1*c3*c4*c5*s2 - 0.080565*c1*c2*c4*s3*s5 + 0.080565*c1*c3*c4*s2*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(0.046806*c2_10_3*c1*c5 - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.047228*c1*c4 + 0.0098041*c1*c5*s4 + 0.080565*c1*s4*s5 - 0.047228*c2*c3*s1*s4 - 0.047228*s1*s2*s3*s4 + 0.0098041*c2*c3*c4*c5*s1 + 0.080565*c2*c3*c4*s1*s5 + 0.0098041*c4*c5*s1*s2*s3 + 0.080565*c4*s1*s2*s3*s5) - 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.010004*c1*c2*c4*s3 - 0.010004*c1*c3*c4*s2 - 0.0057515*c1*c2*s3*s4 + 0.0057515*c1*c3*s2*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) - 0.5*(c2*s3 - 1.0*c3*s2)*((c2*c3 + s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) + s4*(c2*s3 - 1.0*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + 1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3)) + 0.5*c1*(0.010004*s2_10_3*c4 - 0.0057515*s2_10_3*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) + 6.9389e-21*s1*(7.0646e+17*s2_10_3*c4*c5 - 3.4031e+18*s2_10_3*s4 + 5.8053e+18*s2_10_3*c4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*c1*(0.22719*c2_10_3 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.0030306*c1*c4 + 0.0052716*c1*s4 + 0.0052716*c2_10_3*c4*s1 - 0.0030306*c2_10_3*s1*s4) - 0.5*s1*(0.22719*c2_10_3 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.0030306*c4*s1 + 0.0052716*s1*s4 - 0.0052716*c2_10_3*c1*c4 + 0.0030306*c2_10_3*c1*s4) - 6.9389e-21*s1*(0.027438*c4*s1 - 0.0056959*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.046806*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.027438*c2_10_3*c1*s4)*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5) + 0.5*s1*(0.16636*s2_10_3*c4*s5 - 0.057*s2_10_3*s4 + 0.0099304*s2_10_3*c6*s4 - 0.0099495*s2_10_3*s4*s6 + 0.0099495*s2_10_3*c4*c5*c6 + 0.0099304*s2_10_3*c4*c5*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*s1*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 0.5*c1*(0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 6.9389e-21*c1*(0.027438*c1*c4 + 0.0056959*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.046806*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.027438*c2_10_3*s1*s4)*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5) + 0.5*c1*(0.16636*s2_10_3*c4*s5 - 0.057*s2_10_3*s4 + 0.0099304*s2_10_3*c6*s4 - 0.0099495*s2_10_3*s4*s6 + 0.0099495*s2_10_3*c4*c5*c6 + 0.0099304*s2_10_3*c4*c5*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) - 6.9389e-21*c1*(7.0646e+17*s2_10_3*c4*c5 - 3.4031e+18*s2_10_3*s4 + 5.8053e+18*s2_10_3*c4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*s1*(0.010004*s2_10_3*c4 - 0.0057515*s2_10_3*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2)),  qp1*(0.5*(7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c2*c3 + s2*s3) - 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(s4*s6*(c2*s3 - 1.0*c3*s2) - 1.0*c4*c5*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.057*c1*c4 + 0.16636*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099304*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0099495*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*s4*(s1*s2*s3 + c2*c3*s1) + 0.0099495*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.0099304*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(0.047228*c1*c4 + 0.0098041*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.080565*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.047228*c2_10_3*s1*s4)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(c6*s4*(c2*s3 - 1.0*c3*s2) + c4*c5*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.01*c1 - 0.057*c1*s4 + 0.28*s1*s2 + 0.0099304*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099495*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.16636*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*c4*(s1*s2*s3 + c2*c3*s1) - 0.16636*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*c2*s1*s3 + 0.245*c3*s1*s2) + 0.5*(0.027438*c1*c4 + 0.0056959*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.046806*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.027438*c2_10_3*s1*s4)*(0.047228*c1*s4 - 0.01*c1 - 0.28*s1*s2 + 0.0098041*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.080565*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.047228*c2_10_3*c4*s1 - 0.080565*s2_10_3*c5*s1 + 0.0098041*s2_10_3*s1*s5 + 0.245*c2*s1*s3 - 0.245*c3*s1*s2) - 0.5*(0.057*c4*s1 - 0.0099304*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099495*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*s4*(c1*c2*c3 + c1*s2*s3) + 0.0099495*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*(0.0030306*c4*s1 + 0.0052716*s1*s4 - 0.0052716*c2_10_3*c1*c4 + 0.0030306*c2_10_3*c1*s4)*(0.01*s1 - 0.28*c1*s2 + 0.010004*c4*s1 - 0.0057515*s1*s4 + 0.0057515*c2_10_3*c1*c4 + 0.010004*c2_10_3*c1*s4 + 0.22719*c1*c2*s3 - 0.22719*c1*c3*s2) - 0.5*(0.0057515*c4*s1 + 0.010004*s1*s4 - 0.010004*c2_10_3*c1*c4 + 0.0057515*c2_10_3*c1*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) - 0.5*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.01*s1 - 0.28*c1*s2 - 0.057*s1*s4 + 0.16636*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*c4*(c1*c2*c3 + c1*s2*s3) - 0.0099495*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) - 0.5*(0.0030306*c1*c4 + 0.0052716*c1*s4 + 0.0052716*c2_10_3*c4*s1 - 0.0030306*c2_10_3*s1*s4)*(0.01*c1 + 0.010004*c1*c4 - 0.0057515*c1*s4 + 0.28*s1*s2 - 0.0057515*c2_10_3*c4*s1 - 0.010004*c2_10_3*s1*s4 - 0.22719*c2*s1*s3 + 0.22719*c3*s1*s2) - 0.5*(0.0057515*c1*c4 + 0.010004*c1*s4 + 0.010004*c2_10_3*c4*s1 - 0.0057515*c2_10_3*s1*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) + 0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.047228*c4*s1 - 0.0098041*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.080565*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.047228*c2_10_3*c1*s4)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.027438*c4*s1 - 0.0056959*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.046806*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.027438*c2_10_3*c1*s4)*(0.01*s1 - 0.28*c1*s2 - 0.047228*s1*s4 + 0.0098041*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.080565*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.047228*c2_10_3*c1*c4 - 0.080565*s2_10_3*c1*c5 + 0.0098041*s2_10_3*c1*s5 + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 0.5*s4*(c2*s3 - 1.0*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 0.5*c4*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2))*(c2*s3 - 1.0*c3*s2) + 0.5*c4*(c2*s3 - 1.0*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + 0.5*s4*(0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c2*s3 - 1.0*c3*s2) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*s4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*c4*s5*(c2*s3 - 1.0*c3*s2)*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*c4*c5*(c2*s3 - 1.0*c3*s2)*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*c4*s5*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2))) + qp5*(0.5*(0.046806*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0056959*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0056959*s2_10_3*c5*s1 + 0.046806*s2_10_3*s1*s5)*(0.047228*c4*s1 + 0.080565*s1*s4*s5 + 0.0098041*c5*s1*s4 + 0.047228*c1*c2*c3*s4 + 0.047228*c1*s2*s3*s4 - 0.0098041*c1*c2*c3*c4*c5 - 0.080565*c1*c2*c3*c4*s5 - 0.0098041*c1*c4*c5*s2*s3 - 0.080565*c1*c4*s2*s3*s5) - 1.0*(0.0099495*c1*c6*s4*s5 - 0.16636*c1*c5*s4 + 0.0099304*c1*s4*s5*s6 - 0.16636*c2*c3*c4*c5*s1 - 0.16636*c4*c5*s1*s2*s3 + 0.0099495*c2*c3*c4*c6*s1*s5 + 0.0099304*c2*c3*c4*s1*s5*s6 + 0.0099495*c4*c6*s1*s2*s3*s5 + 0.0099304*c4*s1*s2*s3*s5*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 1.0*(0.0098041*s1*s4*s5 - 0.080565*c5*s1*s4 + 0.080565*c1*c2*c3*c4*c5 - 0.0098041*c1*c2*c3*c4*s5 + 0.080565*c1*c4*c5*s2*s3 - 0.0098041*c1*c4*s2*s3*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.057*c4*s1 + 0.16636*s1*s4*s5 - 0.0099304*c4*c6*s1 + 0.0099495*c4*s1*s6 + 0.057*c1*c2*c3*s4 + 0.0099495*c5*c6*s1*s4 + 0.057*c1*s2*s3*s4 + 0.0099304*c5*s1*s4*s6 - 0.16636*c1*c2*c3*c4*s5 - 0.0099304*c1*c2*c3*c6*s4 + 0.0099495*c1*c2*c3*s4*s6 - 0.16636*c1*c4*s2*s3*s5 - 0.0099304*c1*c6*s2*s3*s4 + 0.0099495*c1*s2*s3*s4*s6 - 0.0099495*c1*c2*c3*c4*c5*c6 - 0.0099304*c1*c2*c3*c4*c5*s6 - 0.0099495*c1*c4*c5*c6*s2*s3 - 0.0099304*c1*c4*c5*s2*s3*s6) + 0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*c3 + s2*s3)*((s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.16636*c1*c2*s3*s5 - 0.16636*c4*c5*s1 - 0.16636*c1*c3*s2*s5 + 0.0099495*c4*c6*s1*s5 + 0.0099304*c4*s1*s5*s6 - 0.16636*c1*c2*c3*c5*s4 + 0.0099495*c1*c2*c5*c6*s3 - 0.0099495*c1*c3*c5*c6*s2 + 0.0099304*c1*c2*c5*s3*s6 - 0.0099304*c1*c3*c5*s2*s6 - 0.16636*c1*c5*s2*s3*s4 + 0.0099495*c1*c2*c3*c6*s4*s5 + 0.0099304*c1*c2*c3*s4*s5*s6 + 0.0099495*c1*c6*s2*s3*s4*s5 + 0.0099304*c1*s2*s3*s4*s5*s6) + (0.080565*c1*c5*s4 - 0.0098041*c1*s4*s5 + 0.080565*c2*c3*c4*c5*s1 - 0.0098041*c2*c3*c4*s1*s5 + 0.080565*c4*c5*s1*s2*s3 - 0.0098041*c4*s1*s2*s3*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.046806*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0056959*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0056959*s2_10_3*c1*c5 + 0.046806*s2_10_3*c1*s5)*(0.047228*c1*c4 + 0.0098041*c1*c5*s4 + 0.080565*c1*s4*s5 - 0.047228*c2*c3*s1*s4 - 0.047228*s1*s2*s3*s4 + 0.0098041*c2*c3*c4*c5*s1 + 0.080565*c2*c3*c4*s1*s5 + 0.0098041*c4*c5*s1*s2*s3 + 0.080565*c4*s1*s2*s3*s5) - 1.0*(0.16636*c5*s1*s4 - 0.0099495*c6*s1*s4*s5 - 0.0099304*s1*s4*s5*s6 - 0.16636*c1*c2*c3*c4*c5 - 0.16636*c1*c4*c5*s2*s3 + 0.0099495*c1*c2*c3*c4*c6*s5 + 0.0099304*c1*c2*c3*c4*s5*s6 + 0.0099495*c1*c4*c6*s2*s3*s5 + 0.0099304*c1*c4*s2*s3*s5*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(0.027438*c4*s1 - 0.0056959*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.046806*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.027438*c2_10_3*c1*s4)*(0.080565*c1*c4*c5 - 0.0098041*c1*c4*s5 + 0.0098041*c2*c5*s1*s3 - 0.0098041*c3*c5*s1*s2 + 0.080565*c2*s1*s3*s5 - 0.080565*c3*s1*s2*s5 - 0.080565*c2*c3*c5*s1*s4 + 0.0098041*c2*c3*s1*s4*s5 - 0.080565*c5*s1*s2*s3*s4 + 0.0098041*s1*s2*s3*s4*s5) + 0.5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.057*c1*c4 - 0.0099304*c1*c4*c6 + 0.0099495*c1*c4*s6 + 0.16636*c1*s4*s5 + 0.0099495*c1*c5*c6*s4 - 0.057*c2*c3*s1*s4 + 0.0099304*c1*c5*s4*s6 - 0.057*s1*s2*s3*s4 + 0.16636*c2*c3*c4*s1*s5 + 0.0099304*c2*c3*c6*s1*s4 - 0.0099495*c2*c3*s1*s4*s6 + 0.16636*c4*s1*s2*s3*s5 + 0.0099304*c6*s1*s2*s3*s4 - 0.0099495*s1*s2*s3*s4*s6 + 0.0099495*c2*c3*c4*c5*c6*s1 + 0.0099304*c2*c3*c4*c5*s1*s6 + 0.0099495*c4*c5*c6*s1*s2*s3 + 0.0099304*c4*c5*s1*s2*s3*s6) - 0.5*(0.027438*c1*c4 + 0.0056959*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.046806*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.027438*c2_10_3*s1*s4)*(0.080565*c4*c5*s1 - 0.0098041*c4*s1*s5 - 0.0098041*c1*c2*c5*s3 + 0.0098041*c1*c3*c5*s2 - 0.080565*c1*c2*s3*s5 + 0.080565*c1*c3*s2*s5 + 0.080565*c1*c2*c3*c5*s4 - 0.0098041*c1*c2*c3*s4*s5 + 0.080565*c1*c5*s2*s3*s4 - 0.0098041*c1*s2*s3*s4*s5) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.16636*c1*c4*c5 - 0.0099495*c1*c4*c6*s5 - 0.0099304*c1*c4*s5*s6 + 0.16636*c2*s1*s3*s5 - 0.16636*c3*s1*s2*s5 - 0.16636*c2*c3*c5*s1*s4 + 0.0099495*c2*c5*c6*s1*s3 - 0.0099495*c3*c5*c6*s1*s2 + 0.0099304*c2*c5*s1*s3*s6 - 0.0099304*c3*c5*s1*s2*s6 - 0.16636*c5*s1*s2*s3*s4 + 0.0099495*c2*c3*c6*s1*s4*s5 + 0.0099304*c2*c3*s1*s4*s5*s6 + 0.0099495*c6*s1*s2*s3*s4*s5 + 0.0099304*s1*s2*s3*s4*s5*s6) + 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*c3 + s2*s3)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*((s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + s4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*c5*(c2*s3 - 1.0*c3*s2)*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*s5*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s4*(c2*s3 - 1.0*c3*s2)*((c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*((c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(s4*s6*(c2*s3 - 1.0*c3*s2) - 1.0*c4*c5*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*s4*(c2*s3 - 1.0*c3*s2) + c4*c5*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + c4*s5*(c2*s3 - 1.0*c3*s2)*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*s4*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)))) - 1.0*qp2*(0.5*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3)*(0.057*c1*c4 - 0.0099304*c1*c4*c6 + 0.0099495*c1*c4*s6 + 0.16636*c1*s4*s5 + 0.0099495*c1*c5*c6*s4 - 0.057*c2*c3*s1*s4 + 0.0099304*c1*c5*s4*s6 - 0.057*s1*s2*s3*s4 + 0.16636*c2*c3*c4*s1*s5 + 0.0099304*c2*c3*c6*s1*s4 - 0.0099495*c2*c3*s1*s4*s6 + 0.16636*c4*s1*s2*s3*s5 + 0.0099304*c6*s1*s2*s3*s4 - 0.0099495*s1*s2*s3*s4*s6 + 0.0099495*c2*c3*c4*c5*c6*s1 + 0.0099304*c2*c3*c4*c5*s1*s6 + 0.0099495*c4*c5*c6*s1*s2*s3 + 0.0099304*c4*c5*s1*s2*s3*s6) + 0.5*(s1*s2*s3 + c2*c3*s1)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.010004*c2*c4*s1*s3 - 0.010004*c3*c4*s1*s2 - 0.0057515*c2*s1*s3*s4 + 0.0057515*c3*s1*s2*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 0.5*(c1*c2*c3 + c1*s2*s3)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3)) + 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.057*c1*c2*s3*s4 - 0.057*c1*c3*s2*s4 - 0.16636*c1*c2*c4*s3*s5 + 0.16636*c1*c3*c4*s2*s5 - 0.0099304*c1*c2*c6*s3*s4 + 0.0099304*c1*c3*c6*s2*s4 + 0.0099495*c1*c2*s3*s4*s6 - 0.0099495*c1*c3*s2*s4*s6 - 0.0099495*c1*c2*c4*c5*c6*s3 + 0.0099495*c1*c3*c4*c5*c6*s2 - 0.0099304*c1*c2*c4*c5*s3*s6 + 0.0099304*c1*c3*c4*c5*s2*s6) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*(1.0*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - (c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + 1.0*(s1*s2*s3 + c2*c3*s1)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*(s1*s2*s3 + c2*c3*s1)*((c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(c2*s3 - 1.0*c3*s2)*((c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c1*c2*s3 - 1.0*c1*c3*s2) + 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - 1.0*(0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + (2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2))*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 0.5*(c1*c2*c3 + c1*s2*s3)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.16267*c1*c2 - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.046806*c2_10_3*c1*c5 + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.047228*c1*c4 + 0.0098041*c1*c5*s4 + 0.080565*c1*s4*s5 - 0.047228*c2*c3*s1*s4 - 0.047228*s1*s2*s3*s4 + 0.0098041*c2*c3*c4*c5*s1 + 0.080565*c2*c3*c4*s1*s5 + 0.0098041*c4*c5*s1*s2*s3 + 0.080565*c4*s1*s2*s3*s5) + 0.5*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.057*c2*s1*s3*s4 - 0.057*c3*s1*s2*s4 - 0.16636*c2*c4*s1*s3*s5 + 0.16636*c3*c4*s1*s2*s5 - 0.0099304*c2*c6*s1*s3*s4 + 0.0099304*c3*c6*s1*s2*s4 + 0.0099495*c2*s1*s3*s4*s6 - 0.0099495*c3*s1*s2*s4*s6 - 0.0099495*c2*c4*c5*c6*s1*s3 + 0.0099495*c3*c4*c5*c6*s1*s2 - 0.0099304*c2*c4*c5*s1*s3*s6 + 0.0099304*c3*c4*c5*s1*s2*s6) + 0.5*(0.14754*c1*c2 + 0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3)*(0.0057515*c1*c4 + 0.010004*c1*s4 + 0.010004*c2*c3*c4*s1 - 0.0057515*c2*c3*s1*s4 + 0.010004*c4*s1*s2*s3 - 0.0057515*s1*s2*s3*s4) + 0.5*(0.047228*c2*s1*s3*s4 - 0.047228*c3*s1*s2*s4 - 0.0098041*c2*c4*c5*s1*s3 + 0.0098041*c3*c4*c5*s1*s2 - 0.080565*c2*c4*s1*s3*s5 + 0.080565*c3*c4*s1*s2*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + (c1*c2*c3 + c1*s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + 1.0*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + s4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 0.5*c1*(1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2)) - (7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c2*s1*s3 - 1.0*c3*s1*s2) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - (0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))) + 0.5*(0.14754*c2*s1 + 0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1)*(0.0057515*c4*s1 + 0.010004*s1*s4 - 0.010004*c1*c2*c3*c4 + 0.0057515*c1*c2*c3*s4 - 0.010004*c1*c4*s2*s3 + 0.0057515*c1*s2*s3*s4) - 0.5*c1*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*c3 + s2*s3)*(1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - (c2*c3 + s2*s3)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) - (c2*s3 - 1.0*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) + 1.0*s4*(c2*s3 - 1.0*c3*s2)*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + c4*(c2*c3 + s2*s3)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*c3 + s2*s3)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(s1*s2*s3 + c2*c3*s1)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.16267*c2*s1 - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.14234*s1*s2*s3 + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.047228*c4*s1 + 0.080565*s1*s4*s5 + 0.0098041*c5*s1*s4 + 0.047228*c1*c2*c3*s4 + 0.047228*c1*s2*s3*s4 - 0.0098041*c1*c2*c3*c4*c5 - 0.080565*c1*c2*c3*c4*s5 - 0.0098041*c1*c4*c5*s2*s3 - 0.080565*c1*c4*s2*s3*s5) + 0.5*(c2*c3 + s2*s3)*((c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3))) - 0.5*(c1*c2*c3 + c1*s2*s3)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*c3 + s2*s3)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.047228*c1*c2*s3*s4 - 0.047228*c1*c3*s2*s4 - 0.0098041*c1*c2*c4*c5*s3 + 0.0098041*c1*c3*c4*c5*s2 - 0.080565*c1*c2*c4*s3*s5 + 0.080565*c1*c3*c4*s2*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.010004*c1*c2*c4*s3 - 0.010004*c1*c3*c4*s2 - 0.0057515*c1*c2*s3*s4 + 0.0057515*c1*c3*s2*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) - 0.5*(c2*s3 - 1.0*c3*s2)*((c2*c3 + s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) + s4*(c2*s3 - 1.0*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + 1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3)) + 0.5*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1)*(0.057*c4*s1 + 0.16636*s1*s4*s5 - 0.0099304*c4*c6*s1 + 0.0099495*c4*s1*s6 + 0.057*c1*c2*c3*s4 + 0.0099495*c5*c6*s1*s4 + 0.057*c1*s2*s3*s4 + 0.0099304*c5*s1*s4*s6 - 0.16636*c1*c2*c3*c4*s5 - 0.0099304*c1*c2*c3*c6*s4 + 0.0099495*c1*c2*c3*s4*s6 - 0.16636*c1*c4*s2*s3*s5 - 0.0099304*c1*c6*s2*s3*s4 + 0.0099495*c1*s2*s3*s4*s6 - 0.0099495*c1*c2*c3*c4*c5*c6 - 0.0099304*c1*c2*c3*c4*c5*s6 - 0.0099495*c1*c4*c5*c6*s2*s3 - 0.0099304*c1*c4*c5*s2*s3*s6) - 0.5*c1*(0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*c1*(0.010004*s2_10_3*c4 - 0.0057515*s2_10_3*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) + 0.5*s1*(0.16636*s2_10_3*c4*s5 - 0.057*s2_10_3*s4 + 0.0099304*s2_10_3*c6*s4 - 0.0099495*s2_10_3*s4*s6 + 0.0099495*s2_10_3*c4*c5*c6 + 0.0099304*s2_10_3*c4*c5*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*s1*(0.027438*c4*s1 - 0.0056959*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.046806*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.027438*c2_10_3*c1*s4)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) - 0.5*c1*(0.004902*s2_10_3*c4_5 - 0.040283*s2_10_3*s4_10_5 + 0.040283*s2_10_3*s4_5 - 0.047228*s2_10_3*s4 + 0.004902*c4_10_5*s2_10_3)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - 0.5*c1*(0.0030306*c1*c4 + 0.0052716*c1*s4 + 0.0052716*c2_10_3*c4*s1 - 0.0030306*c2_10_3*s1*s4)*(0.22719*c2_10_3 + 0.28*c2 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4) + 0.5*s1*(0.004902*s2_10_3*c4_5 - 0.040283*s2_10_3*s4_10_5 + 0.040283*s2_10_3*s4_5 - 0.047228*s2_10_3*s4 + 0.004902*c4_10_5*s2_10_3)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*s1*(0.0030306*c4*s1 + 0.0052716*s1*s4 - 0.0052716*c2_10_3*c1*c4 + 0.0030306*c2_10_3*c1*s4)*(0.22719*c2_10_3 + 0.28*c2 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4) + 0.5*c1*(0.16636*s2_10_3*c4*s5 - 0.057*s2_10_3*s4 + 0.0099304*s2_10_3*c6*s4 - 0.0099495*s2_10_3*s4*s6 + 0.0099495*s2_10_3*c4*c5*c6 + 0.0099304*s2_10_3*c4*c5*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) - 0.5*c1*(0.027438*c1*c4 + 0.0056959*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.046806*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.027438*c2_10_3*s1*s4)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) - 0.5*s1*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*s1*(0.010004*s2_10_3*c4 - 0.0057515*s2_10_3*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2)) - qp4*(1.0*(0.0030306*c4*s1 + 0.0052716*s1*s4 - 0.0052716*c2_10_3*c1*c4 + 0.0030306*c2_10_3*c1*s4)*(0.0057515*c1*c4 + 0.010004*c1*s4 + 0.010004*c2*c3*c4*s1 - 0.0057515*c2*c3*s1*s4 + 0.010004*c4*s1*s2*s3 - 0.0057515*s1*s2*s3*s4) - (0.0098041*c4*c5*s1 - 0.047228*s1*s4 + 0.080565*c4*s1*s5 + 0.047228*c1*c2*c3*c4 + 0.047228*c1*c4*s2*s3 + 0.0098041*c1*c2*c3*c5*s4 + 0.080565*c1*c2*c3*s4*s5 + 0.0098041*c1*c5*s2*s3*s4 + 0.080565*c1*s2*s3*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - (0.027438*c1*c4 + 0.0056959*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.046806*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.027438*c2_10_3*s1*s4)*(0.047228*c4*s1 + 0.080565*s1*s4*s5 + 0.0098041*c5*s1*s4 + 0.047228*c1*c2*c3*s4 + 0.047228*c1*s2*s3*s4 - 0.0098041*c1*c2*c3*c4*c5 - 0.080565*c1*c2*c3*c4*s5 - 0.0098041*c1*c4*c5*s2*s3 - 0.080565*c1*c4*s2*s3*s5) + 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - (c2*c3 + s2*s3)*((c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(s4*s6*(c2*s3 - 1.0*c3*s2) - 1.0*c4*c5*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*s4*(c2*s3 - 1.0*c3*s2) + c4*c5*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + c4*s5*(c2*s3 - 1.0*c3*s2)*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + (c2*s1*s3 - 1.0*c3*s1*s2)*(1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2)) - (7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c2*s1*s3 - 1.0*c3*s1*s2) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - (0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))) + 1.0*(c2*s1*s3 - 1.0*c3*s1*s2)*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 1.0*(0.047228*c1*s4 - 0.0098041*c1*c4*c5 - 0.080565*c1*c4*s5 + 0.047228*c2*c3*c4*s1 + 0.047228*c4*s1*s2*s3 + 0.0098041*c2*c3*c5*s1*s4 + 0.080565*c2*c3*s1*s4*s5 + 0.0098041*c5*s1*s2*s3*s4 + 0.080565*s1*s2*s3*s4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - (c1*c2*s3 - 1.0*c1*c3*s2)*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - (0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.057*c4*s1 + 0.16636*s1*s4*s5 - 0.0099304*c4*c6*s1 + 0.0099495*c4*s1*s6 + 0.057*c1*c2*c3*s4 + 0.0099495*c5*c6*s1*s4 + 0.057*c1*s2*s3*s4 + 0.0099304*c5*s1*s4*s6 - 0.16636*c1*c2*c3*c4*s5 - 0.0099304*c1*c2*c3*c6*s4 + 0.0099495*c1*c2*c3*s4*s6 - 0.16636*c1*c4*s2*s3*s5 - 0.0099304*c1*c6*s2*s3*s4 + 0.0099495*c1*s2*s3*s4*s6 - 0.0099495*c1*c2*c3*c4*c5*c6 - 0.0099304*c1*c2*c3*c4*c5*s6 - 0.0099495*c1*c4*c5*c6*s2*s3 - 0.0099304*c1*c4*c5*s2*s3*s6) + 1.0*(0.027438*c4*s1 - 0.0056959*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.046806*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.027438*c2_10_3*c1*s4)*(0.047228*c1*c4 + 0.0098041*c1*c5*s4 + 0.080565*c1*s4*s5 - 0.047228*c2*c3*s1*s4 - 0.047228*s1*s2*s3*s4 + 0.0098041*c2*c3*c4*c5*s1 + 0.080565*c2*c3*c4*s1*s5 + 0.0098041*c4*c5*s1*s2*s3 + 0.080565*c4*s1*s2*s3*s5) + 1.0*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.057*c1*s4 - 0.16636*c1*c4*s5 - 0.0099304*c1*c6*s4 + 0.0099495*c1*s4*s6 - 0.0099495*c1*c4*c5*c6 + 0.057*c2*c3*c4*s1 - 0.0099304*c1*c4*c5*s6 + 0.057*c4*s1*s2*s3 - 0.0099304*c2*c3*c4*c6*s1 + 0.0099495*c2*c3*c4*s1*s6 + 0.16636*c2*c3*s1*s4*s5 - 0.0099304*c4*c6*s1*s2*s3 + 0.0099495*c4*s1*s2*s3*s6 + 0.16636*s1*s2*s3*s4*s5 + 0.0099495*c2*c3*c5*c6*s1*s4 + 0.0099304*c2*c3*c5*s1*s4*s6 + 0.0099495*c5*c6*s1*s2*s3*s4 + 0.0099304*c5*s1*s2*s3*s4*s6) - (c1*c2*s3 - 1.0*c1*c3*s2)*((7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c1*c2*s3 - 1.0*c1*c3*s2) + 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - 1.0*(0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + (2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2))*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - (c2*c3 + s2*s3)*((7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c2*c3 + s2*s3) - s4*(c2*s3 - 1.0*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + c4*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2))*(c2*s3 - 1.0*c3*s2) + c4*(c2*s3 - 1.0*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + s4*(0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c2*s3 - 1.0*c3*s2)) + 1.0*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.057*c1*c4 - 0.0099304*c1*c4*c6 + 0.0099495*c1*c4*s6 + 0.16636*c1*s4*s5 + 0.0099495*c1*c5*c6*s4 - 0.057*c2*c3*s1*s4 + 0.0099304*c1*c5*s4*s6 - 0.057*s1*s2*s3*s4 + 0.16636*c2*c3*c4*s1*s5 + 0.0099304*c2*c3*c6*s1*s4 - 0.0099495*c2*c3*s1*s4*s6 + 0.16636*c4*s1*s2*s3*s5 + 0.0099304*c6*s1*s2*s3*s4 - 0.0099495*s1*s2*s3*s4*s6 + 0.0099495*c2*c3*c4*c5*c6*s1 + 0.0099304*c2*c3*c4*c5*s1*s6 + 0.0099495*c4*c5*c6*s1*s2*s3 + 0.0099304*c4*c5*s1*s2*s3*s6) + 1.0*(0.010004*c4*s1 - 0.0057515*s1*s4 + 0.0057515*c1*c2*c3*c4 + 0.010004*c1*c2*c3*s4 + 0.0057515*c1*c4*s2*s3 + 0.010004*c1*s2*s3*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) + 1.0*(0.0057515*c1*s4 - 0.010004*c1*c4 + 0.0057515*c2*c3*c4*s1 + 0.010004*c2*c3*s1*s4 + 0.0057515*c4*s1*s2*s3 + 0.010004*s1*s2*s3*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) - (c2*s1*s3 - 1.0*c3*s1*s2)*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - (0.0030306*c1*c4 + 0.0052716*c1*s4 + 0.0052716*c2_10_3*c4*s1 - 0.0030306*c2_10_3*s1*s4)*(0.0057515*c4*s1 + 0.010004*s1*s4 - 0.010004*c1*c2*c3*c4 + 0.0057515*c1*c2*c3*s4 - 0.010004*c1*c4*s2*s3 + 0.0057515*c1*s2*s3*s4) + 1.0*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.16636*c4*s1*s5 - 0.0099495*s1*s4*s6 - 0.057*s1*s4 + 0.0099304*c6*s1*s4 + 0.057*c1*c2*c3*c4 + 0.0099495*c4*c5*c6*s1 + 0.057*c1*c4*s2*s3 + 0.0099304*c4*c5*s1*s6 - 0.0099304*c1*c2*c3*c4*c6 + 0.0099495*c1*c2*c3*c4*s6 + 0.16636*c1*c2*c3*s4*s5 - 0.0099304*c1*c4*c6*s2*s3 + 0.0099495*c1*c4*s2*s3*s6 + 0.16636*c1*s2*s3*s4*s5 + 0.0099495*c1*c2*c3*c5*c6*s4 + 0.0099304*c1*c2*c3*c5*s4*s6 + 0.0099495*c1*c5*c6*s2*s3*s4 + 0.0099304*c1*c5*s2*s3*s4*s6) - (c2*c3 + s2*s3)*((s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + s4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*c5*(c2*s3 - 1.0*c3*s2)*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*s5*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)))) + 1.0*qp3*(0.5*(0.0030306*s2_10_3*c1*c4 + 0.0052716*s2_10_3*c1*s4 + 0.11971*c1*c2*c3 + 0.11971*c1*s2*s3)*(0.0057515*c1*c4 + 0.010004*c1*s4 + 0.010004*c2*c3*c4*s1 - 0.0057515*c2*c3*s1*s4 + 0.010004*c4*s1*s2*s3 - 0.0057515*s1*s2*s3*s4) + 0.5*(s1*s2*s3 + c2*c3*s1)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.010004*c2*c4*s1*s3 - 0.010004*c3*c4*s1*s2 - 0.0057515*c2*s1*s3*s4 + 0.0057515*c3*s1*s2*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2) + 0.5*(c1*c2*c3 + c1*s2*s3)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3)) + 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.11971*s1*s2*s3 + 0.0030306*s2_10_3*c4*s1 + 0.0052716*s2_10_3*s1*s4 + 0.11971*c2*c3*s1)*(0.0057515*c4*s1 + 0.010004*s1*s4 - 0.010004*c1*c2*c3*c4 + 0.0057515*c1*c2*c3*s4 - 0.010004*c1*c4*s2*s3 + 0.0057515*c1*s2*s3*s4) + 0.5*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.057*c1*c2*s3*s4 - 0.057*c1*c3*s2*s4 - 0.16636*c1*c2*c4*s3*s5 + 0.16636*c1*c3*c4*s2*s5 - 0.0099304*c1*c2*c6*s3*s4 + 0.0099304*c1*c3*c6*s2*s4 + 0.0099495*c1*c2*s3*s4*s6 - 0.0099495*c1*c3*s2*s4*s6 - 0.0099495*c1*c2*c4*c5*c6*s3 + 0.0099495*c1*c3*c4*c5*c6*s2 - 0.0099304*c1*c2*c4*c5*s3*s6 + 0.0099304*c1*c3*c4*c5*s2*s6) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*(1.0*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) - (c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + 1.0*(s1*s2*s3 + c2*c3*s1)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) + c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*(s1*s2*s3 + c2*c3*s1)*((c2*s1*s3 - 1.0*c3*s1*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(c2*s3 - 1.0*c3*s2)*((c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c1*c2*s3 - 1.0*c1*c3*s2) + 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - 1.0*(0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + (2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2))*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 0.5*(c1*c2*c3 + c1*s2*s3)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.057*c2*s1*s3*s4 - 0.057*c3*s1*s2*s4 - 0.16636*c2*c4*s1*s3*s5 + 0.16636*c3*c4*s1*s2*s5 - 0.0099304*c2*c6*s1*s3*s4 + 0.0099304*c3*c6*s1*s2*s4 + 0.0099495*c2*s1*s3*s4*s6 - 0.0099495*c3*s1*s2*s4*s6 - 0.0099495*c2*c4*c5*c6*s1*s3 + 0.0099495*c3*c4*c5*c6*s1*s2 - 0.0099304*c2*c4*c5*s1*s3*s6 + 0.0099304*c3*c4*c5*s1*s2*s6) + 0.5*(0.14234*s1*s2*s3 - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.047228*c4*s1 + 0.080565*s1*s4*s5 + 0.0098041*c5*s1*s4 + 0.047228*c1*c2*c3*s4 + 0.047228*c1*s2*s3*s4 - 0.0098041*c1*c2*c3*c4*c5 - 0.080565*c1*c2*c3*c4*s5 - 0.0098041*c1*c4*c5*s2*s3 - 0.080565*c1*c4*s2*s3*s5) + 0.5*(0.047228*c2*s1*s3*s4 - 0.047228*c3*s1*s2*s4 - 0.0098041*c2*c4*c5*s1*s3 + 0.0098041*c3*c4*c5*s1*s2 - 0.080565*c2*c4*s1*s3*s5 + 0.080565*c3*c4*s1*s2*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.057*c4*s1 + 0.16636*s1*s4*s5 - 0.0099304*c4*c6*s1 + 0.0099495*c4*s1*s6 + 0.057*c1*c2*c3*s4 + 0.0099495*c5*c6*s1*s4 + 0.057*c1*s2*s3*s4 + 0.0099304*c5*s1*s4*s6 - 0.16636*c1*c2*c3*c4*s5 - 0.0099304*c1*c2*c3*c6*s4 + 0.0099495*c1*c2*c3*s4*s6 - 0.16636*c1*c4*s2*s3*s5 - 0.0099304*c1*c6*s2*s3*s4 + 0.0099495*c1*s2*s3*s4*s6 - 0.0099495*c1*c2*c3*c4*c5*c6 - 0.0099304*c1*c2*c3*c4*c5*s6 - 0.0099495*c1*c4*c5*c6*s2*s3 - 0.0099304*c1*c4*c5*s2*s3*s6) + 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - 1.0*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) + (c1*c2*c3 + c1*s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) - 1.0*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + 1.0*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + s4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) - 0.5*c1*(1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) - 0.00069098*s4*(c2*s3 - 1.0*c3*s2)) - (7.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.00016483*s4*(c2*s3 - 1.0*c3*s2))*(c2*s1*s3 - 1.0*c3*s1*s2) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) - (0.00078519*c4*(c2*s3 - 1.0*c3*s2) - 2.4e-7*s4*(c2*s3 - 1.0*c3*s2))*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))) - 0.5*c1*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.057*c1*c4 - 0.0099304*c1*c4*c6 + 0.0099495*c1*c4*s6 + 0.16636*c1*s4*s5 + 0.0099495*c1*c5*c6*s4 - 0.057*c2*c3*s1*s4 + 0.0099304*c1*c5*s4*s6 - 0.057*s1*s2*s3*s4 + 0.16636*c2*c3*c4*s1*s5 + 0.0099304*c2*c3*c6*s1*s4 - 0.0099495*c2*c3*s1*s4*s6 + 0.16636*c4*s1*s2*s3*s5 + 0.0099304*c6*s1*s2*s3*s4 - 0.0099495*s1*s2*s3*s4*s6 + 0.0099495*c2*c3*c4*c5*c6*s1 + 0.0099304*c2*c3*c4*c5*s1*s6 + 0.0099495*c4*c5*c6*s1*s2*s3 + 0.0099304*c4*c5*s1*s2*s3*s6) + 0.5*(c2*c3 + s2*s3)*(1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.00016483*c2*s3 - 0.00016483*c3*s2 + 0.00069098*c4*(c2*c3 + s2*s3) + 2.4e-7*s4*(c2*c3 + s2*s3)) - (c2*c3 + s2*s3)*(0.00034115*c2*s3 - 0.00034115*c3*s2 + 0.00016483*c4*(c2*c3 + s2*s3) - 7.4e-7*s4*(c2*c3 + s2*s3)) - (c2*s3 - 1.0*c3*s2)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) + 1.0*s4*(c2*s3 - 1.0*c3*s2)*(7.4e-7*c3*s2 - 7.4e-7*c2*s3 + 2.4e-7*c4*(c2*c3 + s2*s3) + 0.00078519*s4*(c2*c3 + s2*s3)) + c4*(c2*c3 + s2*s3)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3) + 1.0*s4*(c2*c3 + s2*s3)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3)) + 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(s1*s2*s3 + c2*c3*s1)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*c3 + s2*s3)*((c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3))) - 0.5*(c1*c2*c3 + c1*s2*s3)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*c3 + s2*s3)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.047228*c1*c2*s3*s4 - 0.047228*c1*c3*s2*s4 - 0.0098041*c1*c2*c4*c5*s3 + 0.0098041*c1*c3*c4*c5*s2 - 0.080565*c1*c2*c4*s3*s5 + 0.080565*c1*c3*c4*s2*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(0.046806*c2_10_3*c1*c5 - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.047228*c1*c4 + 0.0098041*c1*c5*s4 + 0.080565*c1*s4*s5 - 0.047228*c2*c3*s1*s4 - 0.047228*s1*s2*s3*s4 + 0.0098041*c2*c3*c4*c5*s1 + 0.080565*c2*c3*c4*s1*s5 + 0.0098041*c4*c5*s1*s2*s3 + 0.080565*c4*s1*s2*s3*s5) - 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.010004*c1*c2*c4*s3 - 0.010004*c1*c3*c4*s2 - 0.0057515*c1*c2*s3*s4 + 0.0057515*c1*c3*s2*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) - 0.5*(c2*s3 - 1.0*c3*s2)*((c2*c3 + s2*s3)*(0.00034115*c2*c3 - 0.00016483*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.00034115*s2*s3) + s4*(c2*s3 - 1.0*c3*s2)*(2.4e-7*c4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*c2*c3 + 0.00078519*s4*(c2*s3 - 1.0*c3*s2) + 7.4e-7*s2*s3) + 1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.00069098*c4*(c2*s3 - 1.0*c3*s2) - 0.00016483*c2*c3 + 2.4e-7*s4*(c2*s3 - 1.0*c3*s2) - 0.00016483*s2*s3)) + 0.5*c1*(0.010004*s2_10_3*c4 - 0.0057515*s2_10_3*s4)*(0.0052693*c1 + 0.0052716*c1*c4 - 0.0030306*c1*s4 + 0.14754*s1*s2 - 0.0030306*c2_10_3*c4*s1 - 0.0052716*c2_10_3*s1*s4 - 0.11971*c2*s1*s3 + 0.11971*c3*s1*s2) + 6.9389e-21*s1*(7.0646e+17*s2_10_3*c4*c5 - 3.4031e+18*s2_10_3*s4 + 5.8053e+18*s2_10_3*c4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*c1*(0.22719*c2_10_3 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.0030306*c1*c4 + 0.0052716*c1*s4 + 0.0052716*c2_10_3*c4*s1 - 0.0030306*c2_10_3*s1*s4) - 0.5*s1*(0.22719*c2_10_3 + 0.0057515*s2_10_3*c4 + 0.010004*s2_10_3*s4)*(0.0030306*c4*s1 + 0.0052716*s1*s4 - 0.0052716*c2_10_3*c1*c4 + 0.0030306*c2_10_3*c1*s4) - 6.9389e-21*s1*(0.027438*c4*s1 - 0.0056959*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.046806*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.027438*c2_10_3*c1*s4)*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5) + 0.5*s1*(0.16636*s2_10_3*c4*s5 - 0.057*s2_10_3*s4 + 0.0099304*s2_10_3*c6*s4 - 0.0099495*s2_10_3*s4*s6 + 0.0099495*s2_10_3*c4*c5*c6 + 0.0099304*s2_10_3*c4*c5*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*s1*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 0.5*c1*(0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 6.9389e-21*c1*(0.027438*c1*c4 + 0.0056959*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.046806*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.027438*c2_10_3*s1*s4)*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5) + 0.5*c1*(0.16636*s2_10_3*c4*s5 - 0.057*s2_10_3*s4 + 0.0099304*s2_10_3*c6*s4 - 0.0099495*s2_10_3*s4*s6 + 0.0099495*s2_10_3*c4*c5*c6 + 0.0099304*s2_10_3*c4*c5*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) - 6.9389e-21*c1*(7.0646e+17*s2_10_3*c4*c5 - 3.4031e+18*s2_10_3*s4 + 5.8053e+18*s2_10_3*c4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*s1*(0.010004*s2_10_3*c4 - 0.0057515*s2_10_3*s4)*(0.0052693*s1 - 0.14754*c1*s2 + 0.0052716*c4*s1 - 0.0030306*s1*s4 + 0.0030306*c2_10_3*c1*c4 + 0.0052716*c2_10_3*c1*s4 + 0.11971*c1*c2*s3 - 0.11971*c1*c3*s2)) + qp6*(0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*((c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(s4*s6*(c2*s3 - 1.0*c3*s2) - 1.0*c4*c5*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*s4*(c2*s3 - 1.0*c3*s2) + c4*c5*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + c4*s5*(c2*s3 - 1.0*c3*s2)*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.0099304*c4*c5*c6*s1 - 0.0099495*c6*s1*s4 - 0.0099304*s1*s4*s6 - 0.0099495*c4*c5*s1*s6 + 0.0099495*c1*c2*c3*c4*c6 + 0.0099304*c1*c2*c3*c4*s6 + 0.0099495*c1*c4*c6*s2*s3 - 0.0099304*c1*c2*c6*s3*s5 + 0.0099304*c1*c3*c6*s2*s5 + 0.0099304*c1*c4*s2*s3*s6 + 0.0099495*c1*c2*s3*s5*s6 - 0.0099495*c1*c3*s2*s5*s6 + 0.0099304*c1*c2*c3*c5*c6*s4 - 0.0099495*c1*c2*c3*c5*s4*s6 + 0.0099304*c1*c5*c6*s2*s3*s4 - 0.0099495*c1*c5*s2*s3*s4*s6) - 1.0*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.0099495*c1*c5*s4*s6 - 0.0099304*c1*c4*s6 - 0.0099304*c1*c5*c6*s4 - 0.0099495*c1*c4*c6 + 0.0099495*c2*c3*c6*s1*s4 + 0.0099304*c2*c3*s1*s4*s6 + 0.0099495*c6*s1*s2*s3*s4 + 0.0099304*s1*s2*s3*s4*s6 - 0.0099304*c2*c3*c4*c5*c6*s1 + 0.0099495*c2*c3*c4*c5*s1*s6 - 0.0099304*c4*c5*c6*s1*s2*s3 + 0.0099495*c4*c5*s1*s2*s3*s6) + 0.5*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.057*c1*c4 - 0.0099304*c1*c4*c6 + 0.0099495*c1*c4*s6 + 0.16636*c1*s4*s5 + 0.0099495*c1*c5*c6*s4 - 0.057*c2*c3*s1*s4 + 0.0099304*c1*c5*s4*s6 - 0.057*s1*s2*s3*s4 + 0.16636*c2*c3*c4*s1*s5 + 0.0099304*c2*c3*c6*s1*s4 - 0.0099495*c2*c3*s1*s4*s6 + 0.16636*c4*s1*s2*s3*s5 + 0.0099304*c6*s1*s2*s3*s4 - 0.0099495*s1*s2*s3*s4*s6 + 0.0099495*c2*c3*c4*c5*c6*s1 + 0.0099304*c2*c3*c4*c5*s1*s6 + 0.0099495*c4*c5*c6*s1*s2*s3 + 0.0099304*c4*c5*s1*s2*s3*s6) + 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.0099495*c1*c6*s4 + 0.0099304*c1*s4*s6 - 0.0099304*c1*c4*c5*c6 + 0.0099495*c1*c4*c5*s6 + 0.0099495*c2*c3*c4*c6*s1 + 0.0099304*c2*c3*c4*s1*s6 + 0.0099495*c4*c6*s1*s2*s3 - 0.0099304*c2*c6*s1*s3*s5 + 0.0099304*c3*c6*s1*s2*s5 + 0.0099304*c4*s1*s2*s3*s6 + 0.0099495*c2*s1*s3*s5*s6 - 0.0099495*c3*s1*s2*s5*s6 + 0.0099304*c2*c3*c5*c6*s1*s4 - 0.0099495*c2*c3*c5*s1*s4*s6 + 0.0099304*c5*c6*s1*s2*s3*s4 - 0.0099495*c5*s1*s2*s3*s4*s6) - 1.0*(0.0099495*c4*c6*s1 + 0.0099304*c4*s1*s6 + 0.0099304*c5*c6*s1*s4 - 0.0099495*c5*s1*s4*s6 + 0.0099495*c1*c2*c3*c6*s4 + 0.0099304*c1*c2*c3*s4*s6 + 0.0099495*c1*c6*s2*s3*s4 + 0.0099304*c1*s2*s3*s4*s6 - 0.0099304*c1*c2*c3*c4*c5*c6 + 0.0099495*c1*c2*c3*c4*c5*s6 - 0.0099304*c1*c4*c5*c6*s2*s3 + 0.0099495*c1*c4*c5*s2*s3*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*c3 + s2*s3)*((c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.057*c4*s1 + 0.16636*s1*s4*s5 - 0.0099304*c4*c6*s1 + 0.0099495*c4*s1*s6 + 0.057*c1*c2*c3*s4 + 0.0099495*c5*c6*s1*s4 + 0.057*c1*s2*s3*s4 + 0.0099304*c5*s1*s4*s6 - 0.16636*c1*c2*c3*c4*s5 - 0.0099304*c1*c2*c3*c6*s4 + 0.0099495*c1*c2*c3*s4*s6 - 0.16636*c1*c4*s2*s3*s5 - 0.0099304*c1*c6*s2*s3*s4 + 0.0099495*c1*s2*s3*s4*s6 - 0.0099495*c1*c2*c3*c4*c5*c6 - 0.0099304*c1*c2*c3*c4*c5*s6 - 0.0099495*c1*c4*c5*c6*s2*s3 - 0.0099304*c1*c4*c5*s2*s3*s6) + 0.5*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*s5*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)))),  1.0*qp2*(0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0098041*c1*c2*c3*c5 + 0.080565*c1*c2*c3*s5 + 0.0098041*c1*c5*s2*s3 + 0.080565*c1*s2*s3*s5 + 0.080565*c1*c2*c5*s3*s4 - 0.080565*c1*c3*c5*s2*s4 - 0.0098041*c1*c2*s3*s4*s5 + 0.0098041*c1*c3*s2*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(0.16267*c1*c2 - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.046806*c2_10_3*c1*c5 + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.080565*c1*c4*c5 - 0.0098041*c1*c4*s5 + 0.0098041*c2*c5*s1*s3 - 0.0098041*c3*c5*s1*s2 + 0.080565*c2*s1*s3*s5 - 0.080565*c3*s1*s2*s5 - 0.080565*c2*c3*c5*s1*s4 + 0.0098041*c2*c3*s1*s4*s5 - 0.080565*c5*s1*s2*s3*s4 + 0.0098041*s1*s2*s3*s4*s5) + 0.5*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.16636*c1*c2*c3*s5 + 0.16636*c1*s2*s3*s5 + 0.0099495*c1*c2*c3*c5*c6 + 0.0099304*c1*c2*c3*c5*s6 + 0.16636*c1*c2*c5*s3*s4 - 0.16636*c1*c3*c5*s2*s4 + 0.0099495*c1*c5*c6*s2*s3 + 0.0099304*c1*c5*s2*s3*s6 - 0.0099495*c1*c2*c6*s3*s4*s5 + 0.0099495*c1*c3*c6*s2*s4*s5 - 0.0099304*c1*c2*s3*s4*s5*s6 + 0.0099304*c1*c3*s2*s4*s5*s6) + 0.5*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.16636*c2*c3*s1*s5 + 0.16636*s1*s2*s3*s5 + 0.0099495*c2*c3*c5*c6*s1 + 0.0099304*c2*c3*c5*s1*s6 + 0.16636*c2*c5*s1*s3*s4 - 0.16636*c3*c5*s1*s2*s4 + 0.0099495*c5*c6*s1*s2*s3 + 0.0099304*c5*s1*s2*s3*s6 - 0.0099495*c2*c6*s1*s3*s4*s5 + 0.0099495*c3*c6*s1*s2*s4*s5 - 0.0099304*c2*s1*s3*s4*s5*s6 + 0.0099304*c3*s1*s2*s4*s5*s6) + 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(0.16267*c2*s1 - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.14234*s1*s2*s3 + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.080565*c4*c5*s1 - 0.0098041*c4*s1*s5 - 0.0098041*c1*c2*c5*s3 + 0.0098041*c1*c3*c5*s2 - 0.080565*c1*c2*s3*s5 + 0.080565*c1*c3*s2*s5 + 0.080565*c1*c2*c3*c5*s4 - 0.0098041*c1*c2*c3*s4*s5 + 0.080565*c1*c5*s2*s3*s4 - 0.0098041*c1*s2*s3*s4*s5) - 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) - 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1)*(0.16636*c1*c2*s3*s5 - 0.16636*c4*c5*s1 - 0.16636*c1*c3*s2*s5 + 0.0099495*c4*c6*s1*s5 + 0.0099304*c4*s1*s5*s6 - 0.16636*c1*c2*c3*c5*s4 + 0.0099495*c1*c2*c5*c6*s3 - 0.0099495*c1*c3*c5*c6*s2 + 0.0099304*c1*c2*c5*s3*s6 - 0.0099304*c1*c3*c5*s2*s6 - 0.16636*c1*c5*s2*s3*s4 + 0.0099495*c1*c2*c3*c6*s4*s5 + 0.0099304*c1*c2*c3*s4*s5*s6 + 0.0099495*c1*c6*s2*s3*s4*s5 + 0.0099304*c1*s2*s3*s4*s5*s6) + 0.5*(0.0098041*c2*c3*c5*s1 + 0.080565*c2*c3*s1*s5 + 0.0098041*c5*s1*s2*s3 + 0.080565*s1*s2*s3*s5 + 0.080565*c2*c5*s1*s3*s4 - 0.080565*c3*c5*s1*s2*s4 - 0.0098041*c2*s1*s3*s4*s5 + 0.0098041*c3*s1*s2*s4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3)*(0.16636*c1*c4*c5 - 0.0099495*c1*c4*c6*s5 - 0.0099304*c1*c4*s5*s6 + 0.16636*c2*s1*s3*s5 - 0.16636*c3*s1*s2*s5 - 0.16636*c2*c3*c5*s1*s4 + 0.0099495*c2*c5*c6*s1*s3 - 0.0099495*c3*c5*c6*s1*s2 + 0.0099304*c2*c5*s1*s3*s6 - 0.0099304*c3*c5*s1*s2*s6 - 0.16636*c5*s1*s2*s3*s4 + 0.0099495*c2*c3*c6*s1*s4*s5 + 0.0099304*c2*c3*s1*s4*s5*s6 + 0.0099495*c6*s1*s2*s3*s4*s5 + 0.0099304*s1*s2*s3*s4*s5*s6) + 0.5*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*(0.046806*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0056959*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0056959*s2_10_3*c1*c5 + 0.046806*s2_10_3*c1*s5)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) + 0.5*c4*(c2*c3 + s2*s3)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*c1*(0.040283*s2_10_3*s4_10_5 + 0.004902*s2_10_3*c4_5 + 0.040283*s2_10_3*s4_5 - 0.0098041*c2_10_3*c5 - 0.080565*c2_10_3*s5 - 0.004902*c4_10_5*s2_10_3)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*s1*(0.0099495*c6*(c2_10_3*c5 + s2_10_3*s4*s5) + 0.16636*c2_10_3*s5 + 0.0099304*s6*(c2_10_3*c5 + s2_10_3*s4*s5) - 0.16636*s2_10_3*c5*s4)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*(0.040283*s2_10_3*s4_10_5 + 0.004902*s2_10_3*c4_5 + 0.040283*s2_10_3*s4_5 - 0.0098041*c2_10_3*c5 - 0.080565*c2_10_3*s5 - 0.004902*c4_10_5*s2_10_3)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 0.5*c1*(0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*c3 + s2*s3)*((c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*c1*(0.046806*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0056959*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0056959*s2_10_3*c5*s1 + 0.046806*s2_10_3*s1*s5)*(0.245*c2_10_3 + 0.28*c2 + 0.004902*s2_10_3*s4_10_5 - 0.040283*s2_10_3*c4_5 + 0.004902*s2_10_3*s4_5 + 0.080565*c2_10_3*c5 + 0.047228*s2_10_3*c4 - 0.0098041*c2_10_3*s5 + 0.040283*c4_10_5*s2_10_3) + 0.5*c1*(0.0099495*c6*(c2_10_3*c5 + s2_10_3*s4*s5) + 0.16636*c2_10_3*s5 + 0.0099304*s6*(c2_10_3*c5 + s2_10_3*s4*s5) - 0.16636*s2_10_3*c5*s4)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*s1*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*c4*(c2*s3 - 1.0*c3*s2)*((c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3))) - 0.5*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)))) - qp3*(0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - (s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + 1.0*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0098041*c1*c2*c3*c5 + 0.080565*c1*c2*c3*s5 + 0.0098041*c1*c5*s2*s3 + 0.080565*c1*s2*s3*s5 + 0.080565*c1*c2*c5*s3*s4 - 0.080565*c1*c3*c5*s2*s4 - 0.0098041*c1*c2*s3*s4*s5 + 0.0098041*c1*c3*s2*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.16636*c1*c2*c3*s5 + 0.16636*c1*s2*s3*s5 + 0.0099495*c1*c2*c3*c5*c6 + 0.0099304*c1*c2*c3*c5*s6 + 0.16636*c1*c2*c5*s3*s4 - 0.16636*c1*c3*c5*s2*s4 + 0.0099495*c1*c5*c6*s2*s3 + 0.0099304*c1*c5*s2*s3*s6 - 0.0099495*c1*c2*c6*s3*s4*s5 + 0.0099495*c1*c3*c6*s2*s4*s5 - 0.0099304*c1*c2*s3*s4*s5*s6 + 0.0099304*c1*c3*s2*s4*s5*s6) + 0.5*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.16636*c1*c2*s3*s5 - 0.16636*c4*c5*s1 - 0.16636*c1*c3*s2*s5 + 0.0099495*c4*c6*s1*s5 + 0.0099304*c4*s1*s5*s6 - 0.16636*c1*c2*c3*c5*s4 + 0.0099495*c1*c2*c5*c6*s3 - 0.0099495*c1*c3*c5*c6*s2 + 0.0099304*c1*c2*c5*s3*s6 - 0.0099304*c1*c3*c5*s2*s6 - 0.16636*c1*c5*s2*s3*s4 + 0.0099495*c1*c2*c3*c6*s4*s5 + 0.0099304*c1*c2*c3*s4*s5*s6 + 0.0099495*c1*c6*s2*s3*s4*s5 + 0.0099304*c1*s2*s3*s4*s5*s6) + 0.5*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.16636*c2*c3*s1*s5 + 0.16636*s1*s2*s3*s5 + 0.0099495*c2*c3*c5*c6*s1 + 0.0099304*c2*c3*c5*s1*s6 + 0.16636*c2*c5*s1*s3*s4 - 0.16636*c3*c5*s1*s2*s4 + 0.0099495*c5*c6*s1*s2*s3 + 0.0099304*c5*s1*s2*s3*s6 - 0.0099495*c2*c6*s1*s3*s4*s5 + 0.0099495*c3*c6*s1*s2*s4*s5 - 0.0099304*c2*s1*s3*s4*s5*s6 + 0.0099304*c3*s1*s2*s4*s5*s6) + 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) - 0.5*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.16636*c1*c4*c5 - 0.0099495*c1*c4*c6*s5 - 0.0099304*c1*c4*s5*s6 + 0.16636*c2*s1*s3*s5 - 0.16636*c3*s1*s2*s5 - 0.16636*c2*c3*c5*s1*s4 + 0.0099495*c2*c5*c6*s1*s3 - 0.0099495*c3*c5*c6*s1*s2 + 0.0099304*c2*c5*s1*s3*s6 - 0.0099304*c3*c5*s1*s2*s6 - 0.16636*c5*s1*s2*s3*s4 + 0.0099495*c2*c3*c6*s1*s4*s5 + 0.0099304*c2*c3*s1*s4*s5*s6 + 0.0099495*c6*s1*s2*s3*s4*s5 + 0.0099304*s1*s2*s3*s4*s5*s6) - 0.5*c1*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) - 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c1*c2*s3 - 1.0*c1*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0098041*c2*c3*c5*s1 + 0.080565*c2*c3*s1*s5 + 0.0098041*c5*s1*s2*s3 + 0.080565*s1*s2*s3*s5 + 0.080565*c2*c5*s1*s3*s4 - 0.080565*c3*c5*s1*s2*s4 - 0.0098041*c2*s1*s3*s4*s5 + 0.0098041*c3*s1*s2*s4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) + 0.5*(0.046806*c2_10_3*c1*c5 - 0.046806*s5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) - 0.0056959*c5*(c1*c2*s3*s4 - 1.0*c1*c3*s2*s4) + 0.027438*s2_10_3*c1*c4 - 0.0056959*c2_10_3*c1*s5 + 0.14234*c1*c2*c3 + 0.14234*c1*s2*s3)*(0.080565*c1*c4*c5 - 0.0098041*c1*c4*s5 + 0.0098041*c2*c5*s1*s3 - 0.0098041*c3*c5*s1*s2 + 0.080565*c2*s1*s3*s5 - 0.080565*c3*s1*s2*s5 - 0.080565*c2*c3*c5*s1*s4 + 0.0098041*c2*c3*s1*s4*s5 - 0.080565*c5*s1*s2*s3*s4 + 0.0098041*s1*s2*s3*s4*s5) - 0.5*s1*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.14234*s1*s2*s3 - 0.046806*s5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) - 0.0056959*c5*(c2*s1*s3*s4 - 1.0*c3*s1*s2*s4) + 0.046806*c2_10_3*c5*s1 + 0.027438*s2_10_3*c4*s1 - 0.0056959*c2_10_3*s1*s5 + 0.14234*c2*c3*s1)*(0.080565*c4*c5*s1 - 0.0098041*c4*s1*s5 - 0.0098041*c1*c2*c5*s3 + 0.0098041*c1*c3*c5*s2 - 0.080565*c1*c2*s3*s5 + 0.080565*c1*c3*s2*s5 + 0.080565*c1*c2*c3*c5*s4 - 0.0098041*c1*c2*c3*s4*s5 + 0.080565*c1*c5*s2*s3*s4 - 0.0098041*c1*s2*s3*s4*s5) - 6.9389e-21*s1*(0.046806*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0056959*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0056959*s2_10_3*c1*c5 + 0.046806*s2_10_3*c1*s5)*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5) + 0.5*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*c4*(c2*c3 + s2*s3)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*c1*(0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*s1*(0.0099495*c6*(c2_10_3*c5 + s2_10_3*s4*s5) + 0.16636*c2_10_3*s5 + 0.0099304*s6*(c2_10_3*c5 + s2_10_3*s4*s5) - 0.16636*s2_10_3*c5*s4)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 0.5*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*s1*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 6.9389e-21*c1*(7.0646e+17*c2_10_3*c5 + 5.8053e+18*c2_10_3*s5 - 5.8053e+18*s2_10_3*c5*s4 + 7.0646e+17*s2_10_3*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) + 0.5*c4*(c2*s1*s3 - 1.0*c3*s1*s2)*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*c3 + s2*s3)*((c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 6.9389e-21*c1*(0.046806*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0056959*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0056959*s2_10_3*c5*s1 + 0.046806*s2_10_3*s1*s5)*(1.7654e+19*c2_10_3 + 5.8053e+18*c2_10_3*c5 + 3.4031e+18*s2_10_3*c4 - 7.0646e+17*c2_10_3*s5 + 7.0646e+17*s2_10_3*c5*s4 + 5.8053e+18*s2_10_3*s4*s5) + 0.5*c1*(0.0099495*c6*(c2_10_3*c5 + s2_10_3*s4*s5) + 0.16636*c2_10_3*s5 + 0.0099304*s6*(c2_10_3*c5 + s2_10_3*s4*s5) - 0.16636*s2_10_3*c5*s4)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*c4*(c2*s3 - 1.0*c3*s2)*((c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00021268*c5*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s5*(c2*s3 - 1.0*c3*s2) + 2.91e-6*c4*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*c3 + s2*s3) + 0.00021268*s4*s5*(c2*c3 + s2*s3)) - 1.0*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*s3 - 1.0*c3*s2) - 0.0010637*s5*(c2*s3 - 1.0*c3*s2) - 1.1e-7*c4*(c2*c3 + s2*s3) + 0.0010637*c5*s4*(c2*c3 + s2*s3) + 5.21e-6*s4*s5*(c2*c3 + s2*s3)) - 1.0*c4*(c2*c3 + s2*s3)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*s3 - 1.0*c3*s2) + 0.0010847*c4*(c2*c3 + s2*s3) - 1.1e-7*c5*s4*(c2*c3 + s2*s3) + 2.91e-6*s4*s5*(c2*c3 + s2*s3))) + 6.9389e-21*s1*(7.0646e+17*c2_10_3*c5 + 5.8053e+18*c2_10_3*s5 - 5.8053e+18*s2_10_3*c5*s4 + 7.0646e+17*s2_10_3*s4*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*c4*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)))) + qp4*(0.5*(0.046806*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0056959*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0056959*s2_10_3*c5*s1 + 0.046806*s2_10_3*s1*s5)*(0.047228*c4*s1 + 0.080565*s1*s4*s5 + 0.0098041*c5*s1*s4 + 0.047228*c1*c2*c3*s4 + 0.047228*c1*s2*s3*s4 - 0.0098041*c1*c2*c3*c4*c5 - 0.080565*c1*c2*c3*c4*s5 - 0.0098041*c1*c4*c5*s2*s3 - 0.080565*c1*c4*s2*s3*s5) - 1.0*(0.0099495*c1*c6*s4*s5 - 0.16636*c1*c5*s4 + 0.0099304*c1*s4*s5*s6 - 0.16636*c2*c3*c4*c5*s1 - 0.16636*c4*c5*s1*s2*s3 + 0.0099495*c2*c3*c4*c6*s1*s5 + 0.0099304*c2*c3*c4*s1*s5*s6 + 0.0099495*c4*c6*s1*s2*s3*s5 + 0.0099304*c4*s1*s2*s3*s5*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) - 1.0*(0.0098041*s1*s4*s5 - 0.080565*c5*s1*s4 + 0.080565*c1*c2*c3*c4*c5 - 0.0098041*c1*c2*c3*c4*s5 + 0.080565*c1*c4*c5*s2*s3 - 0.0098041*c1*c4*s2*s3*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.057*c4*s1 + 0.16636*s1*s4*s5 - 0.0099304*c4*c6*s1 + 0.0099495*c4*s1*s6 + 0.057*c1*c2*c3*s4 + 0.0099495*c5*c6*s1*s4 + 0.057*c1*s2*s3*s4 + 0.0099304*c5*s1*s4*s6 - 0.16636*c1*c2*c3*c4*s5 - 0.0099304*c1*c2*c3*c6*s4 + 0.0099495*c1*c2*c3*s4*s6 - 0.16636*c1*c4*s2*s3*s5 - 0.0099304*c1*c6*s2*s3*s4 + 0.0099495*c1*s2*s3*s4*s6 - 0.0099495*c1*c2*c3*c4*c5*c6 - 0.0099304*c1*c2*c3*c4*c5*s6 - 0.0099495*c1*c4*c5*c6*s2*s3 - 0.0099304*c1*c4*c5*s2*s3*s6) + 0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*c3 + s2*s3)*((s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.16636*c1*c2*s3*s5 - 0.16636*c4*c5*s1 - 0.16636*c1*c3*s2*s5 + 0.0099495*c4*c6*s1*s5 + 0.0099304*c4*s1*s5*s6 - 0.16636*c1*c2*c3*c5*s4 + 0.0099495*c1*c2*c5*c6*s3 - 0.0099495*c1*c3*c5*c6*s2 + 0.0099304*c1*c2*c5*s3*s6 - 0.0099304*c1*c3*c5*s2*s6 - 0.16636*c1*c5*s2*s3*s4 + 0.0099495*c1*c2*c3*c6*s4*s5 + 0.0099304*c1*c2*c3*s4*s5*s6 + 0.0099495*c1*c6*s2*s3*s4*s5 + 0.0099304*c1*s2*s3*s4*s5*s6) + (0.080565*c1*c5*s4 - 0.0098041*c1*s4*s5 + 0.080565*c2*c3*c4*c5*s1 - 0.0098041*c2*c3*c4*s1*s5 + 0.080565*c4*c5*s1*s2*s3 - 0.0098041*c4*s1*s2*s3*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.046806*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0056959*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0056959*s2_10_3*c1*c5 + 0.046806*s2_10_3*c1*s5)*(0.047228*c1*c4 + 0.0098041*c1*c5*s4 + 0.080565*c1*s4*s5 - 0.047228*c2*c3*s1*s4 - 0.047228*s1*s2*s3*s4 + 0.0098041*c2*c3*c4*c5*s1 + 0.080565*c2*c3*c4*s1*s5 + 0.0098041*c4*c5*s1*s2*s3 + 0.080565*c4*s1*s2*s3*s5) - 1.0*(0.16636*c5*s1*s4 - 0.0099495*c6*s1*s4*s5 - 0.0099304*s1*s4*s5*s6 - 0.16636*c1*c2*c3*c4*c5 - 0.16636*c1*c4*c5*s2*s3 + 0.0099495*c1*c2*c3*c4*c6*s5 + 0.0099304*c1*c2*c3*c4*s5*s6 + 0.0099495*c1*c4*c6*s2*s3*s5 + 0.0099304*c1*c4*s2*s3*s5*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(0.027438*c4*s1 - 0.0056959*c5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) - 0.046806*s5*(c1*c2*c3*c4 - 1.0*s1*s4 + c1*c4*s2*s3) + 0.027438*c2_10_3*c1*s4)*(0.080565*c1*c4*c5 - 0.0098041*c1*c4*s5 + 0.0098041*c2*c5*s1*s3 - 0.0098041*c3*c5*s1*s2 + 0.080565*c2*s1*s3*s5 - 0.080565*c3*s1*s2*s5 - 0.080565*c2*c3*c5*s1*s4 + 0.0098041*c2*c3*s1*s4*s5 - 0.080565*c5*s1*s2*s3*s4 + 0.0098041*s1*s2*s3*s4*s5) + 0.5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.057*c1*c4 - 0.0099304*c1*c4*c6 + 0.0099495*c1*c4*s6 + 0.16636*c1*s4*s5 + 0.0099495*c1*c5*c6*s4 - 0.057*c2*c3*s1*s4 + 0.0099304*c1*c5*s4*s6 - 0.057*s1*s2*s3*s4 + 0.16636*c2*c3*c4*s1*s5 + 0.0099304*c2*c3*c6*s1*s4 - 0.0099495*c2*c3*s1*s4*s6 + 0.16636*c4*s1*s2*s3*s5 + 0.0099304*c6*s1*s2*s3*s4 - 0.0099495*s1*s2*s3*s4*s6 + 0.0099495*c2*c3*c4*c5*c6*s1 + 0.0099304*c2*c3*c4*c5*s1*s6 + 0.0099495*c4*c5*c6*s1*s2*s3 + 0.0099304*c4*c5*s1*s2*s3*s6) - 0.5*(0.027438*c1*c4 + 0.0056959*c5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) + 0.046806*s5*(c1*s4 + c2*c3*c4*s1 + c4*s1*s2*s3) - 0.027438*c2_10_3*s1*s4)*(0.080565*c4*c5*s1 - 0.0098041*c4*s1*s5 - 0.0098041*c1*c2*c5*s3 + 0.0098041*c1*c3*c5*s2 - 0.080565*c1*c2*s3*s5 + 0.080565*c1*c3*s2*s5 + 0.080565*c1*c2*c3*c5*s4 - 0.0098041*c1*c2*c3*s4*s5 + 0.080565*c1*c5*s2*s3*s4 - 0.0098041*c1*s2*s3*s4*s5) + 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.16636*c1*c4*c5 - 0.0099495*c1*c4*c6*s5 - 0.0099304*c1*c4*s5*s6 + 0.16636*c2*s1*s3*s5 - 0.16636*c3*s1*s2*s5 - 0.16636*c2*c3*c5*s1*s4 + 0.0099495*c2*c5*c6*s1*s3 - 0.0099495*c3*c5*c6*s1*s2 + 0.0099304*c2*c5*s1*s3*s6 - 0.0099304*c3*c5*s1*s2*s6 - 0.16636*c5*s1*s2*s3*s4 + 0.0099495*c2*c3*c6*s1*s4*s5 + 0.0099304*c2*c3*s1*s4*s5*s6 + 0.0099495*c6*s1*s2*s3*s4*s5 + 0.0099304*s1*s2*s3*s4*s5*s6) + 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) + (c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + (s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*c3 + s2*s3)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3))*((s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*((s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(1.1e-7*s4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c4*c5*(c2*s3 - 1.0*c3*s2) + 5.21e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*c4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*(c2*s3 - 1.0*c3*s2)*(0.0010847*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*c4*c5*(c2*s3 - 1.0*c3*s2) - 2.91e-6*c4*s5*(c2*s3 - 1.0*c3*s2)) + s4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*c5*(c2*s3 - 1.0*c3*s2)*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*s5*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*s4*(c2*s3 - 1.0*c3*s2)*((c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*c5*(c2*c3 + s2*s3) - 0.0010847*c4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s5*(c2*c3 + s2*s3) + 1.1e-7*c5*s4*(c2*s3 - 1.0*c3*s2) - 2.91e-6*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*(c2*s3 - 1.0*c3*s2)*((c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(s4*s6*(c2*s3 - 1.0*c3*s2) - 1.0*c4*c5*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*s4*(c2*s3 - 1.0*c3*s2) + c4*c5*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + c4*s5*(c2*s3 - 1.0*c3*s2)*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*s4*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)))) + qp1*(0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.0099495*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.16636*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.16636*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*(0.080565*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0098041*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0098041*s2_10_3*c5*s1 + 0.080565*s2_10_3*s1*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.046806*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0056959*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0056959*s2_10_3*c5*s1 + 0.046806*s2_10_3*s1*s5)*(0.047228*c1*s4 - 0.01*c1 - 0.28*s1*s2 + 0.0098041*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.080565*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.047228*c2_10_3*c4*s1 - 0.080565*s2_10_3*c5*s1 + 0.0098041*s2_10_3*s1*s5 + 0.245*c2*s1*s3 - 0.245*c3*s1*s2) + 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.01*s1 - 0.28*c1*s2 - 0.057*s1*s4 + 0.16636*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*c4*(c1*c2*c3 + c1*s2*s3) - 0.0099495*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*(0.080565*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0098041*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0098041*s2_10_3*c1*c5 + 0.080565*s2_10_3*c1*s5)*(0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2) - 0.5*(0.0099495*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.16636*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.16636*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(0.046806*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0056959*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0056959*s2_10_3*c1*c5 + 0.046806*s2_10_3*c1*s5)*(0.01*s1 - 0.28*c1*s2 - 0.047228*s1*s4 + 0.0098041*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.080565*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.047228*c2_10_3*c1*c4 - 0.080565*s2_10_3*c1*c5 + 0.0098041*s2_10_3*c1*s5 + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) + 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.01*c1 - 0.057*c1*s4 + 0.28*s1*s2 + 0.0099304*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099495*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.16636*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*c4*(s1*s2*s3 + c2*c3*s1) - 0.16636*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*c2*s1*s3 + 0.245*c3*s1*s2) + 0.5*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.5*c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 1.0*qp5*((0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.16636*c1*c2*s3*s5 - 0.16636*c4*c5*s1 - 0.16636*c1*c3*s2*s5 + 0.0099495*c4*c6*s1*s5 + 0.0099304*c4*s1*s5*s6 - 0.16636*c1*c2*c3*c5*s4 + 0.0099495*c1*c2*c5*c6*s3 - 0.0099495*c1*c3*c5*c6*s2 + 0.0099304*c1*c2*c5*s3*s6 - 0.0099304*c1*c3*c5*s2*s6 - 0.16636*c1*c5*s2*s3*s4 + 0.0099495*c1*c2*c3*c6*s4*s5 + 0.0099304*c1*c2*c3*s4*s5*s6 + 0.0099495*c1*c6*s2*s3*s4*s5 + 0.0099304*c1*s2*s3*s4*s5*s6) + 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - (0.0058097*s1 - 0.16267*c1*s2 - 0.027438*s1*s4 + 0.0056959*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.046806*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.027438*c2_10_3*c1*c4 - 0.046806*s2_10_3*c1*c5 + 0.0056959*s2_10_3*c1*s5 + 0.14234*c1*c2*s3 - 0.14234*c1*c3*s2)*(0.080565*c2*c5*s1*s3 - 0.080565*c1*c4*s5 - 0.0098041*c1*c4*c5 - 0.080565*c3*c5*s1*s2 - 0.0098041*c2*s1*s3*s5 + 0.0098041*c3*s1*s2*s5 + 0.0098041*c2*c3*c5*s1*s4 + 0.080565*c2*c3*s1*s4*s5 + 0.0098041*c5*s1*s2*s3*s4 + 0.080565*s1*s2*s3*s4*s5) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - (0.046806*c5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) - 0.0056959*s5*(c4*s1 + c1*c2*c3*s4 + c1*s2*s3*s4) + 0.0056959*s2_10_3*c1*c5 + 0.046806*s2_10_3*c1*s5)*(0.080565*c1*c4*c5 - 0.0098041*c1*c4*s5 + 0.0098041*c2*c5*s1*s3 - 0.0098041*c3*c5*s1*s2 + 0.080565*c2*s1*s3*s5 - 0.080565*c3*s1*s2*s5 - 0.080565*c2*c3*c5*s1*s4 + 0.0098041*c2*c3*s1*s4*s5 - 0.080565*c5*s1*s2*s3*s4 + 0.0098041*s1*s2*s3*s4*s5) + 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2))) - (0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.16636*c2*c5*s1*s3 - 0.0099495*c1*c4*c5*c6 - 0.0099304*c1*c4*c5*s6 - 0.16636*c1*c4*s5 - 0.16636*c3*c5*s1*s2 + 0.16636*c2*c3*s1*s4*s5 - 0.0099495*c2*c6*s1*s3*s5 + 0.0099495*c3*c6*s1*s2*s5 - 0.0099304*c2*s1*s3*s5*s6 + 0.0099304*c3*s1*s2*s5*s6 + 0.16636*s1*s2*s3*s4*s5 + 0.0099495*c2*c3*c5*c6*s1*s4 + 0.0099304*c2*c3*c5*s1*s4*s6 + 0.0099495*c5*c6*s1*s2*s3*s4 + 0.0099304*c5*s1*s2*s3*s4*s6) - (0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.16636*c4*s1*s5 + 0.16636*c1*c2*c5*s3 - 0.16636*c1*c3*c5*s2 + 0.0099495*c4*c5*c6*s1 + 0.0099304*c4*c5*s1*s6 + 0.16636*c1*c2*c3*s4*s5 - 0.0099495*c1*c2*c6*s3*s5 + 0.0099495*c1*c3*c6*s2*s5 - 0.0099304*c1*c2*s3*s5*s6 + 0.0099304*c1*c3*s2*s5*s6 + 0.16636*c1*s2*s3*s4*s5 + 0.0099495*c1*c2*c3*c5*c6*s4 + 0.0099304*c1*c2*c3*c5*s4*s6 + 0.0099495*c1*c5*c6*s2*s3*s4 + 0.0099304*c1*c5*s2*s3*s4*s6) + 1.0*(0.0098041*c4*c5*s1 + 0.080565*c4*s1*s5 + 0.080565*c1*c2*c5*s3 - 0.080565*c1*c3*c5*s2 - 0.0098041*c1*c2*s3*s5 + 0.0098041*c1*c3*s2*s5 + 0.0098041*c1*c2*c3*c5*s4 + 0.080565*c1*c2*c3*s4*s5 + 0.0098041*c1*c5*s2*s3*s4 + 0.080565*c1*s2*s3*s4*s5)*(0.027438*c1*s4 - 0.0058097*c1 - 0.16267*s1*s2 + 0.0056959*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.046806*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.027438*c2_10_3*c4*s1 - 0.046806*s2_10_3*c5*s1 + 0.0056959*s2_10_3*s1*s5 + 0.14234*c2*s1*s3 - 0.14234*c3*s1*s2) - (0.046806*c5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) - 0.0056959*s5*(c2*c3*s1*s4 - 1.0*c1*c4 + s1*s2*s3*s4) + 0.0056959*s2_10_3*c5*s1 + 0.046806*s2_10_3*s1*s5)*(0.080565*c4*c5*s1 - 0.0098041*c4*s1*s5 - 0.0098041*c1*c2*c5*s3 + 0.0098041*c1*c3*c5*s2 - 0.080565*c1*c2*s3*s5 + 0.080565*c1*c3*s2*s5 + 0.080565*c1*c2*c3*c5*s4 - 0.0098041*c1*c2*c3*s4*s5 + 0.080565*c1*c5*s2*s3*s4 - 0.0098041*c1*s2*s3*s4*s5) + 1.0*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.16636*c1*c4*c5 - 0.0099495*c1*c4*c6*s5 - 0.0099304*c1*c4*s5*s6 + 0.16636*c2*s1*s3*s5 - 0.16636*c3*s1*s2*s5 - 0.16636*c2*c3*c5*s1*s4 + 0.0099495*c2*c5*c6*s1*s3 - 0.0099495*c3*c5*c6*s1*s2 + 0.0099304*c2*c5*s1*s3*s6 - 0.0099304*c3*c5*s1*s2*s6 - 0.16636*c5*s1*s2*s3*s4 + 0.0099495*c2*c3*c6*s1*s4*s5 + 0.0099304*c2*c3*s1*s4*s5*s6 + 0.0099495*c6*s1*s2*s3*s4*s5 + 0.0099304*s1*s2*s3*s4*s5*s6) + 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - c4*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - c4*(c2*s3 - 1.0*c3*s2)*((s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.91e-6*c4*(c2*s3 - 1.0*c3*s2) - 0.00021268*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 0.00021268*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(5.21e-6*c5*(c2*c3 + s2*s3) + 0.00021268*s5*(c2*c3 + s2*s3) + 0.00021268*c5*s4*(c2*s3 - 1.0*c3*s2) - 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.0010637*s5*(c2*c3 + s2*s3) - 5.21e-6*c5*(c2*c3 + s2*s3) - 1.1e-7*c4*(c2*s3 - 1.0*c3*s2) + 0.0010637*c5*s4*(c2*s3 - 1.0*c3*s2) + 5.21e-6*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(0.0010637*c5*(c2*c3 + s2*s3) + 5.21e-6*s5*(c2*c3 + s2*s3) + 5.21e-6*c5*s4*(c2*s3 - 1.0*c3*s2) - 0.0010637*s4*s5*(c2*s3 - 1.0*c3*s2)) + c4*(c2*s3 - 1.0*c3*s2)*(2.91e-6*s5*(c2*c3 + s2*s3) - 1.1e-7*c5*(c2*c3 + s2*s3) + 2.91e-6*c5*s4*(c2*s3 - 1.0*c3*s2) + 1.1e-7*s4*s5*(c2*s3 - 1.0*c3*s2)))) - 1.0*qp6*(0.5*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.16636*c1*c4*c5 - 0.0099495*c1*c4*c6*s5 - 0.0099304*c1*c4*s5*s6 + 0.16636*c2*s1*s3*s5 - 0.16636*c3*s1*s2*s5 - 0.16636*c2*c3*c5*s1*s4 + 0.0099495*c2*c5*c6*s1*s3 - 0.0099495*c3*c5*c6*s1*s2 + 0.0099304*c2*c5*s1*s3*s6 - 0.0099304*c3*c5*s1*s2*s6 - 0.16636*c5*s1*s2*s3*s4 + 0.0099495*c2*c3*c6*s1*s4*s5 + 0.0099304*c2*c3*s1*s4*s5*s6 + 0.0099495*c6*s1*s2*s3*s4*s5 + 0.0099304*s1*s2*s3*s4*s5*s6) + (0.0099304*c4*c6*s1*s5 - 0.0099495*c4*s1*s5*s6 + 0.0099304*c1*c2*c5*c6*s3 - 0.0099304*c1*c3*c5*c6*s2 - 0.0099495*c1*c2*c5*s3*s6 + 0.0099495*c1*c3*c5*s2*s6 + 0.0099304*c1*c2*c3*c6*s4*s5 - 0.0099495*c1*c2*c3*s4*s5*s6 + 0.0099304*c1*c6*s2*s3*s4*s5 - 0.0099495*c1*s2*s3*s4*s5*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0099495*c1*c6*s4 + 0.0099304*c1*s4*s6 - 0.0099304*c1*c4*c5*c6 + 0.0099495*c1*c4*c5*s6 + 0.0099495*c2*c3*c4*c6*s1 + 0.0099304*c2*c3*c4*s1*s6 + 0.0099495*c4*c6*s1*s2*s3 - 0.0099304*c2*c6*s1*s3*s5 + 0.0099304*c3*c6*s1*s2*s5 + 0.0099304*c4*s1*s2*s3*s6 + 0.0099495*c2*s1*s3*s5*s6 - 0.0099495*c3*s1*s2*s5*s6 + 0.0099304*c2*c3*c5*c6*s1*s4 - 0.0099495*c2*c3*c5*s1*s4*s6 + 0.0099304*c5*c6*s1*s2*s3*s4 - 0.0099495*c5*s1*s2*s3*s4*s6) - 0.5*(0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.16636*c1*c2*s3*s5 - 0.16636*c4*c5*s1 - 0.16636*c1*c3*s2*s5 + 0.0099495*c4*c6*s1*s5 + 0.0099304*c4*s1*s5*s6 - 0.16636*c1*c2*c3*c5*s4 + 0.0099495*c1*c2*c5*c6*s3 - 0.0099495*c1*c3*c5*c6*s2 + 0.0099304*c1*c2*c5*s3*s6 - 0.0099304*c1*c3*c5*s2*s6 - 0.16636*c1*c5*s2*s3*s4 + 0.0099495*c1*c2*c3*c6*s4*s5 + 0.0099304*c1*c2*c3*s4*s5*s6 + 0.0099495*c1*c6*s2*s3*s4*s5 + 0.0099304*c1*s2*s3*s4*s5*s6) + 0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) - 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0099304*c4*c5*c6*s1 - 0.0099495*c6*s1*s4 - 0.0099304*s1*s4*s6 - 0.0099495*c4*c5*s1*s6 + 0.0099495*c1*c2*c3*c4*c6 + 0.0099304*c1*c2*c3*c4*s6 + 0.0099495*c1*c4*c6*s2*s3 - 0.0099304*c1*c2*c6*s3*s5 + 0.0099304*c1*c3*c6*s2*s5 + 0.0099304*c1*c4*s2*s3*s6 + 0.0099495*c1*c2*s3*s5*s6 - 0.0099495*c1*c3*s2*s5*s6 + 0.0099304*c1*c2*c3*c5*c6*s4 - 0.0099495*c1*c2*c3*c5*s4*s6 + 0.0099304*c1*c5*c6*s2*s3*s4 - 0.0099495*c1*c5*s2*s3*s4*s6) + 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 1.0*(0.0099304*c1*c4*c6*s5 - 0.0099495*c1*c4*s5*s6 - 0.0099304*c2*c5*c6*s1*s3 + 0.0099304*c3*c5*c6*s1*s2 + 0.0099495*c2*c5*s1*s3*s6 - 0.0099495*c3*c5*s1*s2*s6 - 0.0099304*c2*c3*c6*s1*s4*s5 + 0.0099495*c2*c3*s1*s4*s5*s6 - 0.0099304*c6*s1*s2*s3*s4*s5 + 0.0099495*s1*s2*s3*s4*s5*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*c4*(c2*s3 - 1.0*c3*s2)*((c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2)))),  - 1.0*qp1*(0.5*(0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.01*s1 - 0.28*c1*s2 - 0.057*s1*s4 + 0.16636*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.057*c4*(c1*c2*c3 + c1*s2*s3) - 0.0099495*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.16636*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0099495*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099304*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.245*c1*c2*s3 - 0.245*c1*c3*s2) - 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.0099495*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.0099304*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099304*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099495*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) - 0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + 0.5*(0.0099495*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.0099304*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.0099304*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0099495*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*(0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.01*c1 - 0.057*c1*s4 + 0.28*s1*s2 + 0.0099304*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0099495*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.16636*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.057*c4*(s1*s2*s3 + c2*c3*s1) - 0.16636*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0099495*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0099304*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.245*c2*s1*s3 + 0.245*c3*s1*s2) + 0.5*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + qp2*(0.5*(c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*c1*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.056504*c2*s1 + 0.033571*c5*(s1*s2*s3 + c2*c3*s1) - 0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.049441*s1*s2*s3 + 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.049441*c2*c3*s1)*(0.0099304*c4*c5*c6*s1 - 0.0099495*c6*s1*s4 - 0.0099304*s1*s4*s6 - 0.0099495*c4*c5*s1*s6 + 0.0099495*c1*c2*c3*c4*c6 + 0.0099304*c1*c2*c3*c4*s6 + 0.0099495*c1*c4*c6*s2*s3 - 0.0099304*c1*c2*c6*s3*s5 + 0.0099304*c1*c3*c6*s2*s5 + 0.0099304*c1*c4*s2*s3*s6 + 0.0099495*c1*c2*s3*s5*s6 - 0.0099495*c1*c3*s2*s5*s6 + 0.0099304*c1*c2*c3*c5*c6*s4 - 0.0099495*c1*c2*c3*c5*s4*s6 + 0.0099304*c1*c5*c6*s2*s3*s4 - 0.0099495*c1*c5*s2*s3*s4*s6) + 0.5*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(0.0099495*c1*c2*c4*c6*s3 - 0.0099495*c1*c3*c4*c6*s2 + 0.0099304*c1*c2*c3*c6*s5 + 0.0099304*c1*c2*c4*s3*s6 - 0.0099304*c1*c3*c4*s2*s6 - 0.0099495*c1*c2*c3*s5*s6 + 0.0099304*c1*c6*s2*s3*s5 - 0.0099495*c1*s2*s3*s5*s6 + 0.0099304*c1*c2*c5*c6*s3*s4 - 0.0099304*c1*c3*c5*c6*s2*s4 - 0.0099495*c1*c2*c5*s3*s4*s6 + 0.0099495*c1*c3*c5*s2*s4*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.056504*c1*c2 + 0.033571*c5*(c1*c2*c3 + c1*s2*s3) - 0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.049441*c1*c2*c3 + 0.049441*c1*s2*s3)*(0.0099495*c1*c6*s4 + 0.0099304*c1*s4*s6 - 0.0099304*c1*c4*c5*c6 + 0.0099495*c1*c4*c5*s6 + 0.0099495*c2*c3*c4*c6*s1 + 0.0099304*c2*c3*c4*s1*s6 + 0.0099495*c4*c6*s1*s2*s3 - 0.0099304*c2*c6*s1*s3*s5 + 0.0099304*c3*c6*s1*s2*s5 + 0.0099304*c4*s1*s2*s3*s6 + 0.0099495*c2*s1*s3*s5*s6 - 0.0099495*c3*s1*s2*s5*s6 + 0.0099304*c2*c3*c5*c6*s1*s4 - 0.0099495*c2*c3*c5*s1*s4*s6 + 0.0099304*c5*c6*s1*s2*s3*s4 - 0.0099495*c5*s1*s2*s3*s4*s6) + 0.5*(0.0099495*c2*c4*c6*s1*s3 - 0.0099495*c3*c4*c6*s1*s2 + 0.0099304*c2*c3*c6*s1*s5 + 0.0099304*c2*c4*s1*s3*s6 - 0.0099304*c3*c4*s1*s2*s6 - 0.0099495*c2*c3*s1*s5*s6 + 0.0099304*c6*s1*s2*s3*s5 - 0.0099495*s1*s2*s3*s5*s6 + 0.0099304*c2*c5*c6*s1*s3*s4 - 0.0099304*c3*c5*c6*s1*s2*s4 - 0.0099495*c2*c5*s1*s3*s4*s6 + 0.0099495*c3*c5*s1*s2*s4*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*s1*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c1*(0.0099495*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) + 0.0099495*s2_10_3*c4*c6 + 0.0099304*s2_10_3*c4*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*c1*(0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*s1*(0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.28*c2 + 0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) - 0.5*s1*(0.0099495*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) + 0.0099495*s2_10_3*c4*c6 + 0.0099304*s2_10_3*c4*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)) - 1.0*qp3*(0.5*(0.0020078*c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c1*c2*c3 + c1*s2*s3) + 0.011503*c4*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.002004*s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.002004*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.033571*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2) - 0.049441*c1*c2*c3 - 0.049441*c1*s2*s3)*(0.0099495*c1*c6*s4 + 0.0099304*c1*s4*s6 - 0.0099304*c1*c4*c5*c6 + 0.0099495*c1*c4*c5*s6 + 0.0099495*c2*c3*c4*c6*s1 + 0.0099304*c2*c3*c4*s1*s6 + 0.0099495*c4*c6*s1*s2*s3 - 0.0099304*c2*c6*s1*s3*s5 + 0.0099304*c3*c6*s1*s2*s5 + 0.0099304*c4*s1*s2*s3*s6 + 0.0099495*c2*s1*s3*s5*s6 - 0.0099495*c3*s1*s2*s5*s6 + 0.0099304*c2*c3*c5*c6*s1*s4 - 0.0099495*c2*c3*c5*s1*s4*s6 + 0.0099304*c5*c6*s1*s2*s3*s4 - 0.0099495*c5*s1*s2*s3*s4*s6) - 0.5*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0*(s6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) - 1.0*c4*c6*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s5*(c1*c2*c3 + c1*s2*s3) + c5*s4*(c1*c2*s3 - 1.0*c1*c3*s2)) + c4*s6*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (c5*(c1*c2*c3 + c1*s2*s3) - 1.0*s4*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*c1*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + 0.5*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 1.0*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) - 1.0*(c5*(s1*s2*s3 + c2*c3*s1) - 1.0*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3))) + 0.5*(0.0099495*c1*c2*c4*c6*s3 - 0.0099495*c1*c3*c4*c6*s2 + 0.0099304*c1*c2*c3*c6*s5 + 0.0099304*c1*c2*c4*s3*s6 - 0.0099304*c1*c3*c4*s2*s6 - 0.0099495*c1*c2*c3*s5*s6 + 0.0099304*c1*c6*s2*s3*s5 - 0.0099495*c1*s2*s3*s5*s6 + 0.0099304*c1*c2*c5*c6*s3*s4 - 0.0099304*c1*c3*c5*c6*s2*s4 - 0.0099495*c1*c2*c5*s3*s4*s6 + 0.0099495*c1*c3*c5*s2*s4*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 0.0003428*c4*c6*(c2*c3 + s2*s3) - 1.9e-7*c4*s6*(c2*c3 + s2*s3) + 1.0e-7*s4*s5*(c2*c3 + s2*s3)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00013076*c5*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + 1.0e-7*c4*c6*(c2*c3 + s2*s3) + 2.7e-7*c4*s6*(c2*c3 + s2*s3) + 0.00013076*s4*s5*(c2*c3 + s2*s3)) - (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c5*(c2*s3 - 1.0*c3*s2) - 0.00028915*c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.9e-7*c4*c6*(c2*c3 + s2*s3) + 0.00028915*c4*s6*(c2*c3 + s2*s3) + 2.7e-7*s4*s5*(c2*c3 + s2*s3)) + (c6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) - 1.0*c4*s6*(c2*c3 + s2*s3))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*s3 - 1.0*c3*s2) - 1.0*c5*s4*(c2*c3 + s2*s3)) + c4*c6*(c2*c3 + s2*s3))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0099495*c2*c4*c6*s1*s3 - 0.0099495*c3*c4*c6*s1*s2 + 0.0099304*c2*c3*c6*s1*s5 + 0.0099304*c2*c4*s1*s3*s6 - 0.0099304*c3*c4*s1*s2*s6 - 0.0099495*c2*c3*s1*s5*s6 + 0.0099304*c6*s1*s2*s3*s5 - 0.0099495*s1*s2*s3*s5*s6 + 0.0099304*c2*c5*c6*s1*s3*s4 - 0.0099304*c3*c5*c6*s1*s2*s4 - 0.0099495*c2*c5*s1*s3*s4*s6 + 0.0099495*c3*c5*s1*s2*s4*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*s1*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + 0.5*(c5*(c2*s3 - 1.0*c3*s2) + s4*s5*(c2*c3 + s2*s3))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.011503*c4*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.033571*c5*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(s1*s2*s3 + c2*c3*s1) + c5*s4*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*s1*s2*s3 - 0.002004*c4*c6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c4*s6*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.033571*s4*s5*(c2*s1*s3 - 1.0*c3*s1*s2) - 0.049441*c2*c3*s1)*(0.0099304*c4*c5*c6*s1 - 0.0099495*c6*s1*s4 - 0.0099304*s1*s4*s6 - 0.0099495*c4*c5*s1*s6 + 0.0099495*c1*c2*c3*c4*c6 + 0.0099304*c1*c2*c3*c4*s6 + 0.0099495*c1*c4*c6*s2*s3 - 0.0099304*c1*c2*c6*s3*s5 + 0.0099304*c1*c3*c6*s2*s5 + 0.0099304*c1*c4*s2*s3*s6 + 0.0099495*c1*c2*s3*s5*s6 - 0.0099495*c1*c3*s2*s5*s6 + 0.0099304*c1*c2*c3*c5*c6*s4 - 0.0099495*c1*c2*c3*c5*s4*s6 + 0.0099304*c1*c5*c6*s2*s3*s4 - 0.0099495*c1*c5*s2*s3*s4*s6) - 0.5*c1*(0.0099495*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) + 0.0099495*s2_10_3*c4*c6 + 0.0099304*s2_10_3*c4*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) - 0.5*s1*(0.0099495*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) + 0.0099495*s2_10_3*c4*c6 + 0.0099304*s2_10_3*c4*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*c1*(0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5) + 0.5*s1*(0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.245*c2*c3 + 0.245*s2*s3 + 0.16636*c2_10_3*c5 + 0.057*s2_10_3*c4 - 0.0099495*c6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s6*(c2_10_3*s5 - 1.0*s2_10_3*c5*s4) - 0.0099304*s2_10_3*c4*c6 + 0.0099495*s2_10_3*c4*s6 + 0.16636*s2_10_3*s4*s5)) - 1.0*qp6*((0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.0099495*c1*s4*s6 - 0.0099304*c1*c6*s4 - 0.0099495*c1*c4*c5*c6 - 0.0099304*c1*c4*c5*s6 - 0.0099304*c2*c3*c4*c6*s1 + 0.0099495*c2*c3*c4*s1*s6 - 0.0099304*c4*c6*s1*s2*s3 - 0.0099495*c2*c6*s1*s3*s5 + 0.0099495*c3*c6*s1*s2*s5 + 0.0099495*c4*s1*s2*s3*s6 - 0.0099304*c2*s1*s3*s5*s6 + 0.0099304*c3*s1*s2*s5*s6 + 0.0099495*c2*c3*c5*c6*s1*s4 + 0.0099304*c2*c3*c5*s1*s4*s6 + 0.0099495*c5*c6*s1*s2*s3*s4 + 0.0099304*c5*s1*s2*s3*s4*s6) - 1.0*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + (0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0099304*c4*c5*c6*s1 - 0.0099495*c6*s1*s4 - 0.0099304*s1*s4*s6 - 0.0099495*c4*c5*s1*s6 + 0.0099495*c1*c2*c3*c4*c6 + 0.0099304*c1*c2*c3*c4*s6 + 0.0099495*c1*c4*c6*s2*s3 - 0.0099304*c1*c2*c6*s3*s5 + 0.0099304*c1*c3*c6*s2*s5 + 0.0099304*c1*c4*s2*s3*s6 + 0.0099495*c1*c2*s3*s5*s6 - 0.0099495*c1*c3*s2*s5*s6 + 0.0099304*c1*c2*c3*c5*c6*s4 - 0.0099495*c1*c2*c3*c5*s4*s6 + 0.0099304*c1*c5*c6*s2*s3*s4 - 0.0099495*c1*c5*s2*s3*s4*s6) + (0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2)*(0.0099304*c6*s1*s4 - 0.0099495*s1*s4*s6 + 0.0099495*c4*c5*c6*s1 + 0.0099304*c4*c5*s1*s6 - 0.0099304*c1*c2*c3*c4*c6 + 0.0099495*c1*c2*c3*c4*s6 - 0.0099304*c1*c4*c6*s2*s3 - 0.0099495*c1*c2*c6*s3*s5 + 0.0099495*c1*c3*c6*s2*s5 + 0.0099495*c1*c4*s2*s3*s6 - 0.0099304*c1*c2*s3*s5*s6 + 0.0099304*c1*c3*s2*s5*s6 + 0.0099495*c1*c2*c3*c5*c6*s4 + 0.0099304*c1*c2*c3*c5*s4*s6 + 0.0099495*c1*c5*c6*s2*s3*s4 + 0.0099304*c1*c5*s2*s3*s4*s6) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*((c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) - 1.0*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + (0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0099495*c1*c6*s4 + 0.0099304*c1*s4*s6 - 0.0099304*c1*c4*c5*c6 + 0.0099495*c1*c4*c5*s6 + 0.0099495*c2*c3*c4*c6*s1 + 0.0099304*c2*c3*c4*s1*s6 + 0.0099495*c4*c6*s1*s2*s3 - 0.0099304*c2*c6*s1*s3*s5 + 0.0099304*c3*c6*s1*s2*s5 + 0.0099304*c4*s1*s2*s3*s6 + 0.0099495*c2*s1*s3*s5*s6 - 0.0099495*c3*s1*s2*s5*s6 + 0.0099304*c2*c3*c5*c6*s1*s4 - 0.0099495*c2*c3*c5*s1*s4*s6 + 0.0099304*c5*c6*s1*s2*s3*s4 - 0.0099495*c5*s1*s2*s3*s4*s6)) - 1.0*qp5*(0.5*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*((c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.16636*c1*c4*c5 - 0.0099495*c1*c4*c6*s5 - 0.0099304*c1*c4*s5*s6 + 0.16636*c2*s1*s3*s5 - 0.16636*c3*s1*s2*s5 - 0.16636*c2*c3*c5*s1*s4 + 0.0099495*c2*c5*c6*s1*s3 - 0.0099495*c3*c5*c6*s1*s2 + 0.0099304*c2*c5*s1*s3*s6 - 0.0099304*c3*c5*s1*s2*s6 - 0.16636*c5*s1*s2*s3*s4 + 0.0099495*c2*c3*c6*s1*s4*s5 + 0.0099304*c2*c3*s1*s4*s5*s6 + 0.0099495*c6*s1*s2*s3*s4*s5 + 0.0099304*s1*s2*s3*s4*s5*s6) + (0.0099304*c4*c6*s1*s5 - 0.0099495*c4*s1*s5*s6 + 0.0099304*c1*c2*c5*c6*s3 - 0.0099304*c1*c3*c5*c6*s2 - 0.0099495*c1*c2*c5*s3*s6 + 0.0099495*c1*c3*c5*s2*s6 + 0.0099304*c1*c2*c3*c6*s4*s5 - 0.0099495*c1*c2*c3*s4*s5*s6 + 0.0099304*c1*c6*s2*s3*s4*s5 - 0.0099495*c1*s2*s3*s4*s5*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0020078*c6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) - 0.033571*c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.033571*s5*(c1*c2*s3 - 1.0*c1*c3*s2))*(0.0099495*c1*c6*s4 + 0.0099304*c1*s4*s6 - 0.0099304*c1*c4*c5*c6 + 0.0099495*c1*c4*c5*s6 + 0.0099495*c2*c3*c4*c6*s1 + 0.0099304*c2*c3*c4*s1*s6 + 0.0099495*c4*c6*s1*s2*s3 - 0.0099304*c2*c6*s1*s3*s5 + 0.0099304*c3*c6*s1*s2*s5 + 0.0099304*c4*s1*s2*s3*s6 + 0.0099495*c2*s1*s3*s5*s6 - 0.0099495*c3*s1*s2*s5*s6 + 0.0099304*c2*c3*c5*c6*s1*s4 - 0.0099495*c2*c3*c5*s1*s4*s6 + 0.0099304*c5*c6*s1*s2*s3*s4 - 0.0099495*c5*s1*s2*s3*s4*s6) - 0.5*(0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.16636*c1*c2*s3*s5 - 0.16636*c4*c5*s1 - 0.16636*c1*c3*s2*s5 + 0.0099495*c4*c6*s1*s5 + 0.0099304*c4*s1*s5*s6 - 0.16636*c1*c2*c3*c5*s4 + 0.0099495*c1*c2*c5*c6*s3 - 0.0099495*c1*c3*c5*c6*s2 + 0.0099304*c1*c2*c5*s3*s6 - 0.0099304*c1*c3*c5*s2*s6 - 0.16636*c1*c5*s2*s3*s4 + 0.0099495*c1*c2*c3*c6*s4*s5 + 0.0099304*c1*c2*c3*s4*s5*s6 + 0.0099495*c1*c6*s2*s3*s4*s5 + 0.0099304*c1*s2*s3*s4*s5*s6) + 0.5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) - 0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s5*(c2*c3 + s2*s3) - 1.0e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 0.00013076*s5*(c2*c3 + s2*s3) + 0.00013076*c5*s4*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s5*(c2*c3 + s2*s3) + 2.7e-7*c5*s4*(c2*s3 - 1.0*c3*s2)) + (s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + c6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s6*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0020078*c6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.033571*c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.033571*s5*(c2*s1*s3 - 1.0*c3*s1*s2))*(0.0099304*c4*c5*c6*s1 - 0.0099495*c6*s1*s4 - 0.0099304*s1*s4*s6 - 0.0099495*c4*c5*s1*s6 + 0.0099495*c1*c2*c3*c4*c6 + 0.0099304*c1*c2*c3*c4*s6 + 0.0099495*c1*c4*c6*s2*s3 - 0.0099304*c1*c2*c6*s3*s5 + 0.0099304*c1*c3*c6*s2*s5 + 0.0099304*c1*c4*s2*s3*s6 + 0.0099495*c1*c2*s3*s5*s6 - 0.0099495*c1*c3*s2*s5*s6 + 0.0099304*c1*c2*c3*c5*c6*s4 - 0.0099495*c1*c2*c3*c5*s4*s6 + 0.0099304*c1*c5*c6*s2*s3*s4 - 0.0099495*c1*c5*s2*s3*s4*s6) + 0.5*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2))*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 1.0*(0.0099304*c1*c4*c6*s5 - 0.0099495*c1*c4*s5*s6 - 0.0099304*c2*c5*c6*s1*s3 + 0.0099304*c3*c5*c6*s1*s2 + 0.0099495*c2*c5*s1*s3*s6 - 0.0099495*c3*c5*s1*s2*s6 - 0.0099304*c2*c3*c6*s1*s4*s5 + 0.0099495*c2*c3*s1*s4*s5*s6 - 0.0099304*c6*s1*s2*s3*s4*s5 + 0.0099495*s1*s2*s3*s4*s5*s6)*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2) + 0.5*c4*(c2*s3 - 1.0*c3*s2)*((c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2)))) + qp4*(0.5*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*((c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(s4*s6*(c2*s3 - 1.0*c3*s2) - 1.0*c4*c5*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*s4*(c2*s3 - 1.0*c3*s2) + c4*c5*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + c4*s5*(c2*s3 - 1.0*c3*s2)*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.011503*c1*c4 + 0.033571*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + 0.0020078*s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*s4*(s1*s2*s3 + c2*c3*s1) + 0.0020078*c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.0099304*c4*c5*c6*s1 - 0.0099495*c6*s1*s4 - 0.0099304*s1*s4*s6 - 0.0099495*c4*c5*s1*s6 + 0.0099495*c1*c2*c3*c4*c6 + 0.0099304*c1*c2*c3*c4*s6 + 0.0099495*c1*c4*c6*s2*s3 - 0.0099304*c1*c2*c6*s3*s5 + 0.0099304*c1*c3*c6*s2*s5 + 0.0099304*c1*c4*s2*s3*s6 + 0.0099495*c1*c2*s3*s5*s6 - 0.0099495*c1*c3*s2*s5*s6 + 0.0099304*c1*c2*c3*c5*c6*s4 - 0.0099495*c1*c2*c3*c5*s4*s6 + 0.0099304*c1*c5*c6*s2*s3*s4 - 0.0099495*c1*c5*s2*s3*s4*s6) - 1.0*(0.002018*s1 - 0.056504*c1*s2 - 0.011503*s1*s4 + 0.033571*s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*c4*(c1*c2*c3 + c1*s2*s3) - 0.0020078*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*c5*(c1*c2*s3 - 1.0*c1*c3*s2) + 0.0020078*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.002004*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.049441*c1*c2*s3 - 0.049441*c1*c3*s2)*(0.0099495*c1*c5*s4*s6 - 0.0099304*c1*c4*s6 - 0.0099304*c1*c5*c6*s4 - 0.0099495*c1*c4*c6 + 0.0099495*c2*c3*c6*s1*s4 + 0.0099304*c2*c3*s1*s4*s6 + 0.0099495*c6*s1*s2*s3*s4 + 0.0099304*s1*s2*s3*s4*s6 - 0.0099304*c2*c3*c4*c5*c6*s1 + 0.0099495*c2*c3*c4*c5*s1*s6 - 0.0099304*c4*c5*c6*s1*s2*s3 + 0.0099495*c4*c5*s1*s2*s3*s6) + 0.5*(s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*(c2*s1*s3 - 1.0*c3*s1*s2)*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.0020078*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 0.002004*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)) + 0.0020078*s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.057*c1*c4 - 0.0099304*c1*c4*c6 + 0.0099495*c1*c4*s6 + 0.16636*c1*s4*s5 + 0.0099495*c1*c5*c6*s4 - 0.057*c2*c3*s1*s4 + 0.0099304*c1*c5*s4*s6 - 0.057*s1*s2*s3*s4 + 0.16636*c2*c3*c4*s1*s5 + 0.0099304*c2*c3*c6*s1*s4 - 0.0099495*c2*c3*s1*s4*s6 + 0.16636*c4*s1*s2*s3*s5 + 0.0099304*c6*s1*s2*s3*s4 - 0.0099495*s1*s2*s3*s4*s6 + 0.0099495*c2*c3*c4*c5*c6*s1 + 0.0099304*c2*c3*c4*c5*s1*s6 + 0.0099495*c4*c5*c6*s1*s2*s3 + 0.0099304*c4*c5*s1*s2*s3*s6) + 0.5*(c1*c2*s3 - 1.0*c1*c3*s2)*((s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) - 0.5*(0.011503*c4*s1 - 0.002004*c6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.0020078*s6*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + 0.033571*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.011503*s4*(c1*c2*c3 + c1*s2*s3) + 0.0020078*c5*c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + 0.002004*c5*s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)))*(0.0099495*c1*c6*s4 + 0.0099304*c1*s4*s6 - 0.0099304*c1*c4*c5*c6 + 0.0099495*c1*c4*c5*s6 + 0.0099495*c2*c3*c4*c6*s1 + 0.0099304*c2*c3*c4*s1*s6 + 0.0099495*c4*c6*s1*s2*s3 - 0.0099304*c2*c6*s1*s3*s5 + 0.0099304*c3*c6*s1*s2*s5 + 0.0099304*c4*s1*s2*s3*s6 + 0.0099495*c2*s1*s3*s5*s6 - 0.0099495*c3*s1*s2*s5*s6 + 0.0099304*c2*c3*c5*c6*s1*s4 - 0.0099495*c2*c3*c5*s1*s4*s6 + 0.0099304*c5*c6*s1*s2*s3*s4 - 0.0099495*c5*s1*s2*s3*s4*s6) - 1.0*(0.0099495*c4*c6*s1 + 0.0099304*c4*s1*s6 + 0.0099304*c5*c6*s1*s4 - 0.0099495*c5*s1*s4*s6 + 0.0099495*c1*c2*c3*c6*s4 + 0.0099304*c1*c2*c3*s4*s6 + 0.0099495*c1*c6*s2*s3*s4 + 0.0099304*c1*s2*s3*s4*s6 - 0.0099304*c1*c2*c3*c4*c5*c6 + 0.0099495*c1*c2*c3*c4*c5*s6 - 0.0099304*c1*c4*c5*c6*s2*s3 + 0.0099495*c1*c4*c5*s2*s3*s6)*(0.002018*c1 - 0.011503*c1*s4 + 0.056504*s1*s2 + 0.002004*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.0020078*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.033571*s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 0.011503*c4*(s1*s2*s3 + c2*c3*s1) - 0.033571*c5*(c2*s1*s3 - 1.0*c3*s1*s2) + 0.0020078*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.002004*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) - 0.049441*c2*s1*s3 + 0.049441*c3*s1*s2) + 0.5*(s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*((s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(2.7e-7*c4*s5*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00028915*s4*s6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.0003428*c6*s4*(c2*s3 - 1.0*c3*s2) - 1.0e-7*c4*s5*(c2*s3 - 1.0*c3*s2) - 1.9e-7*s4*s6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(c6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(1.0e-7*c6*s4*(c2*s3 - 1.0*c3*s2) - 0.00013076*c4*s5*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s6*(c2*s3 - 1.0*c3*s2) - 2.7e-7*c4*c5*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*c5*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + c5*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*(c2*c3 + s2*s3)*((c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00028915*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.00028915*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2)) + (s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.0003428*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.0003428*c4*s6*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(1.0e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 2.7e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.0e-7*c4*s6*(c2*s3 - 1.0*c3*s2))) + 0.5*(0.0020078*c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + 0.002004*s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 0.002004*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)) + 0.0020078*s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.057*c4*s1 + 0.16636*s1*s4*s5 - 0.0099304*c4*c6*s1 + 0.0099495*c4*s1*s6 + 0.057*c1*c2*c3*s4 + 0.0099495*c5*c6*s1*s4 + 0.057*c1*s2*s3*s4 + 0.0099304*c5*s1*s4*s6 - 0.16636*c1*c2*c3*c4*s5 - 0.0099304*c1*c2*c3*c6*s4 + 0.0099495*c1*c2*c3*s4*s6 - 0.16636*c1*c4*s2*s3*s5 - 0.0099304*c1*c6*s2*s3*s4 + 0.0099495*c1*s2*s3*s4*s6 - 0.0099495*c1*c2*c3*c4*c5*c6 - 0.0099304*c1*c2*c3*c4*c5*s6 - 0.0099495*c1*c4*c5*c6*s2*s3 - 0.0099304*c1*c4*c5*s2*s3*s6) + 0.5*s5*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1))*((s5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c5*(c2*s1*s3 - 1.0*c3*s1*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) - 1.0*c6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(c1*s4 + c4*(s1*s2*s3 + c2*c3*s1)) + s6*(c5*(c1*c4 - 1.0*s4*(s1*s2*s3 + c2*c3*s1)) + s5*(c2*s1*s3 - 1.0*c3*s1*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) + 0.5*s5*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3))*((s5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) + c5*(c1*c2*s3 - 1.0*c1*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(s6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) - 1.0*c6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s1*s4 - 1.0*c4*(c1*c2*c3 + c1*s2*s3)) + s6*(c5*(c4*s1 + s4*(c1*c2*c3 + c1*s2*s3)) - 1.0*s5*(c1*c2*s3 - 1.0*c1*c3*s2)))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2))) - 0.5*c4*s5*(c2*s3 - 1.0*c3*s2)*((s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0*c4*c6*(c2*s3 - 1.0*c3*s2))*(1.9e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 0.0003428*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.0e-7*c5*(c2*c3 + s2*s3) - 0.0003428*c4*c6*(c2*s3 - 1.0*c3*s2) + 1.9e-7*c4*s6*(c2*s3 - 1.0*c3*s2) - 1.0e-7*s4*s5*(c2*s3 - 1.0*c3*s2)) - 1.0*(c5*(c2*c3 + s2*s3) - 1.0*s4*s5*(c2*s3 - 1.0*c3*s2))*(2.7e-7*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 1.0e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 0.00013076*c5*(c2*c3 + s2*s3) + 1.0e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*c4*s6*(c2*s3 - 1.0*c3*s2) + 0.00013076*s4*s5*(c2*s3 - 1.0*c3*s2)) + (c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + c4*s6*(c2*s3 - 1.0*c3*s2))*(0.00028915*c6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) + 1.9e-7*s6*(s5*(c2*c3 + s2*s3) + c5*s4*(c2*s3 - 1.0*c3*s2)) - 2.7e-7*c5*(c2*c3 + s2*s3) - 1.9e-7*c4*c6*(c2*s3 - 1.0*c3*s2) + 0.00028915*c4*s6*(c2*s3 - 1.0*c3*s2) + 2.7e-7*s4*s5*(c2*s3 - 1.0*c3*s2)))) };
 C[1] = std::array<double,6>{ 0 , 0 , 0 , 0 , 0 , 0};
 C[2] = std::array<double,6>{ 0 , 0 , 0 , 0 , 0 , 0};
 C[3] = std::array<double,6>{ 0 , 0 , 0 , 0 , 0 , 0};
 C[4] = std::array<double,6>{ 0 , 0 , 0 , 0 , 0 , 0};
 C[5] = std::array<double,6>{ 0 , 0 , 0 , 0 , 0 , 0};
  
 return C;
}

